package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.msteventextractor.BioNLP11FilesystemInformation;
import edu.stanford.nlp.ie.machinereading.msteventextractor.CoNLLToBioNLP;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DataSet;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DependencyEvaluator;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DependencyParser;
import edu.stanford.nlp.ie.machinereading.msteventextractor.ParserOptions;
import edu.stanford.nlp.ie.machinereading.msteventextractor.ParserOptions.conjoiningMethod;
import edu.stanford.nlp.io.FileSystem;
import edu.stanford.nlp.util.StringUtils;

/**
 * This is the wrapper program that runs everything needed for msteventextractor,
 *   that is, converting to CoNLL format, training/evaluating MST, and generating BioNLP format
 */
public class EventParser {
  
  private static final String TRAINING_CONLL = "training.conll";
  private static final String TESTING_CONLL = "testing.conll";
  private static String mstOutputFilename = "testing.out";
  private static final String OUTPUT_DIR_BIONLP = "output";
  
  private static Level LOG_LEVEL = Level.INFO;
  private static Level READER_LOG_LEVEL = Level.SEVERE;
  private static Logger logger = Logger.getLogger(EventParser.class.getName());
  
  static class CorpusMeta {
    boolean hasGold;
  }
  
  static CorpusMeta readMetaInfo(String output) throws IOException {
    BufferedReader is = new BufferedReader(new FileReader(output + ".meta"));
    CorpusMeta meta = new CorpusMeta();
    String line;
    while((line = is.readLine()) != null){
      if(line.startsWith("hasGold")){
        String [] bits = line.split("\\s+");
        meta.hasGold = Boolean.parseBoolean(bits[1]);
      }
    }
    is.close();
    return meta;
  }
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    System.err.println("Using the following properties:\n" + props);
    logger.setLevel(LOG_LEVEL);
    
    int testKBest = 1;
    if (props.containsKey("mst.kbest.test")) {
      testKBest = Integer.valueOf(props.getProperty("mst.kbest.test"));
    }
    
    mstOutputFilename = props.getProperty("mst.outputfilename", mstOutputFilename);
    
    //
    // Sanity checks
    //
    // make sure the working directory exists
    File wdir = new File(props.getProperty("working.dir"));
    if (!wdir.exists() || !wdir.isDirectory() || !wdir.canWrite()) {
      // TODO why is this commented out?
//      throw new RuntimeException("Working directory " + wdir.getAbsolutePath() + " must exist and be a writable directory!");
    }
    
    String workingDirectory = wdir.getAbsolutePath() + File.separator;
    String modelFileName = workingDirectory + props.getProperty("mst.model");
    File modelFile = new File(modelFileName);

    String outputDirectory = props.getProperty("mst.outputdir", wdir.getAbsolutePath());
    String outputFileName = outputDirectory + File.separator + mstOutputFilename;
    File outputFile = new File(outputFileName);
    
    String trainingConll = workingDirectory + TRAINING_CONLL;
    String testingConll = workingDirectory + TESTING_CONLL;
    String stackingTrainingDirectoryUnsplit = props.getProperty("stacking.train.dir", "");
    String[] stackingTrainingDirectories = stackingTrainingDirectoryUnsplit.trim().length() > 0 ?
        stackingTrainingDirectoryUnsplit.split(",") : new String[0];
    String stackingTestingDirectoryUnsplit = props.getProperty("stacking.test.dir", "");
    String[] stackingTestingDirectories = stackingTestingDirectoryUnsplit.trim().length() > 0 ?
        stackingTestingDirectoryUnsplit.split(",") : new String[0];
    
    String[] stackingTrainingConllFilenames = new String[stackingTrainingDirectories.length];
    for (int index = 0; index < stackingTrainingDirectories.length; index++) {
      stackingTrainingConllFilenames[index] = workingDirectory + "stacking-" + index + "-" + TRAINING_CONLL;
    }
    String[] stackingTestingConllFilenames = new String[stackingTestingDirectories.length];
    for (int index = 0; index < stackingTestingDirectories.length; index++) {
      stackingTestingConllFilenames[index] = workingDirectory + "stacking-" + index + "-" + TESTING_CONLL;
    }
    
    ParserOptions trainingOpts = makeTrainOptions(trainingConll, modelFileName, props);
    ParserOptions testingOpts = makeTestOptions(testingConll, modelFileName, outputFileName, props);
    DataSet dataset = trainingOpts.dataset;
    
    String baseDirectory = props.getProperty("base.directory");
    if (baseDirectory != null) {
      BioNLP11FilesystemInformation.setBaseDirectory(baseDirectory);
    }
    
    //
    // Step 1: convert the BioNLP directories to our extended CoNLL format
    //
    MachineReading.setConsoleLevel(READER_LOG_LEVEL);
    // converting the training corpus
    BioNLPToCoNLL.convert(dataset, props, 
        props.getProperty("train.dir"),
        Integer.parseInt(props.getProperty("trigger.train.beam")),
        Boolean.parseBoolean(props.getProperty("trigger.merge.gold.triggers", "true")),
        Boolean.parseBoolean(props.getProperty("document.level.parsing", "false")),
        trainingConll, READER_LOG_LEVEL, props.getProperty("stacking.train.dir", null));
    
    // convert stacking training corpora (if available)
    for (int index = 0; index < stackingTrainingDirectories.length; index++) {
      String stackingTrainingDirectory = stackingTrainingDirectories[index];
      String stackingTrainingConll = stackingTrainingConllFilenames[index];
      
      BioNLPToCoNLL.convert(dataset, props, 
          stackingTrainingDirectory,
          Integer.parseInt(props.getProperty("trigger.train.beam")),
          Boolean.parseBoolean(props.getProperty("trigger.merge.gold.triggers", "true")),
          Boolean.parseBoolean(props.getProperty("document.level.parsing", "false")),
          stackingTrainingConll, READER_LOG_LEVEL, props.getProperty("train.dir"));
    }
    trainingOpts.stackingtrainfiles = stackingTrainingConllFilenames;
    
    // converting the testing corpus
    int testBeam = Integer.parseInt(props.getProperty("trigger.test.beam"));
    BioNLPToCoNLL.convert(dataset, props, 
        props.getProperty("test.dir"), 
        testBeam,
        testBeam >= 0 ? false : true,
        Boolean.parseBoolean(props.getProperty("document.level.parsing", "false")),
        testingConll, READER_LOG_LEVEL);
    
    // convert stacking testing corpora (if available)
    for (int index = 0; index < stackingTestingDirectories.length; index++) {
      String stackingTestingDirectory = stackingTestingDirectories[index];
      String stackingTestingConll = stackingTestingConllFilenames[index];
      
      BioNLPToCoNLL.convert(dataset, props, 
          stackingTestingDirectory, 
          testBeam,
          testBeam >= 0 ? false : true,
          Boolean.parseBoolean(props.getProperty("document.level.parsing", "false")),
          stackingTestingConll, READER_LOG_LEVEL);
    }
    testingOpts.stackingtestfiles = stackingTestingConllFilenames;    
    
    MachineReading.setConsoleLevel(LOG_LEVEL);
    
    //
    // Step 2: train/test/evaluate msteventextractor
    //
    // train
    if (FileSystem.existsAndNonEmpty(modelFile)) {
      logger.severe("Model file " + modelFileName + " already exists. Will NOT re-train the parser.");
    } else {
      logger.severe("Training parser...");
      DependencyParser.trainFromOptions(trainingOpts);
    }
    
    // test
    if (FileSystem.existsAndNonEmpty(outputFile)) {
      logger.severe("MST output file " + outputFileName + " already exists. Will NOT re-run and will NOT evaluate the parser.");
    } else {
      logger.severe("Running parser...");
      DependencyParser.testFromOptions(testingOpts);

      // evaluate
      CorpusMeta testMeta = readMetaInfo(testingConll);
      assert (testMeta != null);
      if (testMeta.hasGold) {
        if (testKBest <= 1) {
          logger.severe("Evaluating parser:");
          DependencyEvaluator.evaluate(testingConll, outputFileName, "BIONLP");
        } else {
          logger.severe("Will not evaluate parser in k-best mode!");
        }
      } else {
        logger.severe("Test corpus does not have gold info. Skipping evaluation.");
      }
    }
    
    //
    // STEP 3: generate the BioNLP format from our extended CoNLL format
    //         run this step only if we are not testing in K-best mode
    //
    if (testKBest <= 1) {
      File outDir = new File(workingDirectory + OUTPUT_DIR_BIONLP);
      logger.severe("Generating BioNLP formatted-output in " + outDir.getAbsolutePath());
      // this is trivial to regenerate, so it's Ok to delete old executions
      if (outDir.exists()) {
        File[] files = outDir.listFiles();
        for (File f : files) {
          f.delete();
        }
        assert outDir.delete() : "Failed to delete temporary directory " + outDir.getAbsolutePath();
      }
      outDir.mkdir();
      CoNLLToBioNLP converter = new CoNLLToBioNLP(dataset);
      converter.convert(outputFileName, outDir.getAbsolutePath());
    } else {
      logger.severe("Cannot generate BioNLP output from k-best parser output!");
    }

    logger.severe("DONE.");
  }
  
  private static ParserOptions makeGeneralOptions(String modelFile,
      Properties props) {
    ParserOptions opts = new ParserOptions();
    
    opts.modelName = modelFile;
    opts.format = "BIONLP";
    opts.decodeType = props.getProperty("mst.decode.type", "non-proj").toLowerCase();
    assert opts.decodeType.equals("proj") || opts.decodeType.equals("non-proj"); 
    
    opts.secondOrder = props.getProperty("mst.second.order").equalsIgnoreCase("true");
    opts.makeFeatureMatchingPattern(props.getProperty("features"));
    
    opts.dataset = DataSet.load(props);
    
    int minFeatureCount = Integer.parseInt(props.getProperty("features.min.count", "0"));
    boolean collectFeatureCounts = Boolean.parseBoolean(props.getProperty("features.collect.counts", "false"));
    String featureCountsFilename = props.getProperty("features.counts.filename");
    if (minFeatureCount > 0) {
      opts.featureCountPruning = true;
      opts.minFeatureCount = minFeatureCount;
      assert !collectFeatureCounts && featureCountsFilename != null;
    }
    if (collectFeatureCounts) {
      opts.collectFeatureCounts = collectFeatureCounts;
      assert featureCountsFilename != null;
    }
    opts.featureCountsFilename = featureCountsFilename;
    if (featureCountsFilename != null && featureCountsFilename.length() > 0 && minFeatureCount == 0 && !collectFeatureCounts) {
      throw new RuntimeException("Feature counts file specified (features.counts.filename) but minimum feature count (features.min.count) is 0 but we're not collecting feature counts (features.collect.counts).  Drop the feature count filename, increase the min count, or enable feature collection.");
    }
    opts.filterOriginalMSTFeatures = Boolean.parseBoolean(props.getProperty("features.filter.mst.features", "false"));
    String conjoinDistances = props.getProperty("features.conjoin.sentence.distance", "false");
    
    if (conjoinDistances.equalsIgnoreCase("true")) {
      opts.conjoinSentenceDistance = conjoiningMethod.BOTH;  
    } else if (conjoinDistances.equalsIgnoreCase("distance-only")) {
      opts.conjoinSentenceDistance = conjoiningMethod.DISTANCE_ONLY;  
    } else {
      opts.conjoinSentenceDistance = conjoiningMethod.NONE;
    }
    
    return opts;
  }
  
  private static ParserOptions makeTrainOptions(String trainFile, String modelFile, Properties props) {
    ParserOptions opts = makeGeneralOptions(modelFile, props);
    opts.trainfile = trainFile;
    opts.numIters = Integer.parseInt(props.getProperty("mst.num.iters", "10"));
    opts.trainK = Integer.parseInt(props.getProperty("mst.kbest.train", "1"));

    int minFeatureCount = Integer.parseInt(props.getProperty("features.min.count", "0"));
    boolean collectFeatureCounts = Boolean.parseBoolean(props.getProperty("features.collect.counts", "false"));
    String featureCountsFilename = props.getProperty("features.counts.filename");
    
    if (minFeatureCount > 0) {
      System.out.println("Minimum feature count: " + minFeatureCount);
    }
    if (collectFeatureCounts) {
      System.out.println("Feature count collection mode.");
    }
    if (featureCountsFilename != null && featureCountsFilename.length() > 0) {
      System.out.println("Feature counts filename: "  + featureCountsFilename);
    }
    
    // options for dumping edge features statistics and feature names (used in one from of feature selection)
    opts.edgeFeatureStatisticsFilename = props.getProperty("edge.feature.statistics.filename");
    opts.featureNamesFilename = props.getProperty("features.names.filename");
    
    opts.makeForest();
    return opts;
  }

  private static ParserOptions makeTestOptions(String testFile, String modelFile, String outFile, Properties props) {
    ParserOptions opts = makeGeneralOptions(modelFile, props);
    opts.testfile = testFile;
    opts.outfile = outFile;
    opts.testK = Integer.parseInt(props.getProperty("mst.kbest.test", "1"));
    
    opts.makeForest();
    return opts;
  }
}
