/**
 * This class computes the scores for the gold BioNLP dependencies using the official BioNLP scorer
 * The goal is to estimate the quality of our conversion to a format that is a tree, hence parsable
 * This class converts the corpus in the test.dir property
 */
package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.msteventextractor.CoNLLToBioNLP;
import edu.stanford.nlp.ie.machinereading.msteventextractor.GENIA09DataSet;
import edu.stanford.nlp.util.StringUtils;

public class RepresentationChecker {
  
  static final String TESTING_CONLL = "testing.conll";
  static final String OUTPUT_DIR_BIONLP = "output";
  
  private static Level LOG_LEVEL = Level.INFO;
  private static Level READER_LOG_LEVEL = Level.SEVERE;
  private static Logger logger = Logger.getLogger(BioNLPToCoNLL.class.getName());
  
  static class CorpusMeta {
    boolean hasGold;
  }
  
  static CorpusMeta readMetaInfo(String output) throws IOException {
    BufferedReader is = new BufferedReader(new FileReader(output + ".meta"));
    CorpusMeta meta = new CorpusMeta();
    String line;
    while((line = is.readLine()) != null){
      if(line.startsWith("hasGold")){
        String [] bits = line.split("\\s+");
        meta.hasGold = Boolean.parseBoolean(bits[1]);
      }
    }
    is.close();
    return meta;
  }
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    System.err.println("Using the following properties:\n" + props);
    logger.setLevel(LOG_LEVEL);
    
    //
    // Sanity checks
    //
    // make sure the working directory exists
    File wdir = new File(props.getProperty("working.dir"));
    if(! wdir.exists() || ! wdir.isDirectory() || ! wdir.canWrite()){
      throw new RuntimeException("Working directory " + wdir.getAbsolutePath() + " must exist and be a writable directory!");
    }
    
    //
    // Step 1: convert the BioNLP directories to our extended CoNLL format
    //
    MachineReading.setConsoleLevel(READER_LOG_LEVEL);
    // converting the testing corpus
    int testBeam = Integer.parseInt(props.getProperty("trigger.test.beam"));
    GENIA09DataSet dataset = new GENIA09DataSet();
    BioNLPToCoNLL.convert(dataset, props, 
        props.getProperty("test.dir"), 
        testBeam,
        testBeam >= 0 ? false : true,
        Boolean.parseBoolean(props.getProperty("document.level.parsing", "false")),
        wdir.getAbsolutePath() + File.separator + TESTING_CONLL, READER_LOG_LEVEL);
    MachineReading.setConsoleLevel(LOG_LEVEL);
    
    //
    // STEP 2: generate the BioNLP format from our extended CoNLL format
    //
    File outDir = new File(wdir.getAbsolutePath() + File.separator + OUTPUT_DIR_BIONLP);
    logger.severe("Generating BioNLP formatted-output in " + outDir.getAbsolutePath());
    // this is trivial to regenerate, so it's Ok to delete old executions
    if(outDir.exists()){
      File [] files = outDir.listFiles();
      for(File f: files){
        f.delete();
      }
      assert outDir.delete(): "Failed to delete temporary directory " + outDir.getAbsolutePath();
    }
    outDir.mkdir();
    CoNLLToBioNLP converter = new CoNLLToBioNLP(dataset);
    converter.convert(wdir.getAbsolutePath() + File.separator + TESTING_CONLL, outDir.getAbsolutePath());
    
    //
    // score manually using the official BioNLP scorer
    // run: eval_approx_recursive test.dir working.dir/output
    //
    
    logger.severe("DONE.");
  }
}
