package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.nlp.classify.Classifier;
import edu.stanford.nlp.classify.Dataset;
import edu.stanford.nlp.classify.GeneralDataset;
import edu.stanford.nlp.classify.LinearClassifier;
import edu.stanford.nlp.classify.LinearClassifierFactory;
import edu.stanford.nlp.classify.LogPrior;
import edu.stanford.nlp.classify.SVMLightClassifier;
import edu.stanford.nlp.classify.SVMLightClassifierFactory;
import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.RelationFeatureFactory;
import edu.stanford.nlp.ie.machinereading.domains.bionlp.BioNLPFormatReader.HasGoldAnnotation;
import edu.stanford.nlp.ie.machinereading.msteventextractor.DataSet;
import edu.stanford.nlp.ie.machinereading.structure.AnnotationUtils;
import edu.stanford.nlp.ie.machinereading.structure.EntityMention;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.DocumentDirectoryAnnotation;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.DocumentIdAnnotation;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.EntityMentionsAnnotation;
import edu.stanford.nlp.ling.BasicDatum;
import edu.stanford.nlp.ling.CoreAnnotation;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.Datum;
import edu.stanford.nlp.ling.IndexedWord;
import edu.stanford.nlp.ling.CoreAnnotations.CharacterOffsetBeginAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.CharacterOffsetEndAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.DistSimAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.SentencesAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TextAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TokensAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TreeAnnotation;
import edu.stanford.nlp.objectbank.ObjectBank;
import edu.stanford.nlp.optimization.DiffFunction;
import edu.stanford.nlp.optimization.HybridMinimizer;
import edu.stanford.nlp.optimization.Minimizer;
import edu.stanford.nlp.optimization.QNMinimizer;
import edu.stanford.nlp.optimization.StochasticInPlaceMinimizer;
import edu.stanford.nlp.pipeline.Annotation;
import edu.stanford.nlp.pipeline.StanfordCoreNLP;
import edu.stanford.nlp.process.WordShapeClassifier;
import edu.stanford.nlp.stats.ClassicCounter;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.stats.Counters;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.semgraph.SemanticGraph;
import edu.stanford.nlp.trees.semgraph.SemanticGraphEdge;
import edu.stanford.nlp.trees.semgraph.SemanticGraphFactory;
import edu.stanford.nlp.util.CoreMap;
import edu.stanford.nlp.util.ErasureUtils;
import edu.stanford.nlp.util.Index;
import edu.stanford.nlp.util.MutableInteger;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;
import edu.stanford.nlp.util.Triple;

public class TriggerClassifier {
  
  private static Logger logger = Logger.getLogger(TriggerClassifier.class.getName());
  
  private static enum CLASSIFIER_TYPE { LR, SVM1, SVM2, SGD_LR };
  private CLASSIFIER_TYPE classifierType;

  // TODO what's the deal with useAllFeatures?  even when true, features are still filtered so probably this option can disappear
  // also, there are about three different mechanisms for picking features which is poor
  private boolean useAllFeatures = true;
  
  private boolean doErrorAnalysis = false;
  private boolean enableDistSim = false;
  private boolean documentLevelParsing = false;
  private boolean dumpSvmLight = false;
  private boolean useFineGrained = false;
  private boolean enablePosFeatures = false;
  private boolean enableShapeFeatures = false;
  private boolean enableCharTrigrams = false;
  
  // dataset and fields loaded from it
  private DataSet dataset;
  private Set<String> labelsForErrorAnalysis;
  
  private Level logLevel;
  private StanfordCoreNLP processor;
  private HashMap<String, Set<String>> knownTriggers;
  private Classifier<String, String> classifier;
  private Index<String> featureIndex;
  private Index<String> labelIndex;
  private double tolerance;
  private double sigma;
  private Set<String> featuresToUse;
  
  private int maxLeftSurface;
  private int maxRightSurface;
  private int maxDependency;
  
  /** Distributional similarity lexicon */
  private HashMap<String, String> lexicon;
  
  public TriggerClassifier(Properties props) {
    dataset = DataSet.load(props);
    labelsForErrorAnalysis = dataset.getLabelsForTriggerErrorAnalysis();
    
    String triggerFeatures = props.getProperty("trigger.features");
    if (triggerFeatures == null) {
      featuresToUse = null;
      logger.info("trigger.features property not defined. Will use ALL features.");
    } else {
      String[] toks = triggerFeatures.split(",");
      featuresToUse = new HashSet<String>(Arrays.asList(toks));
      logger.info("The following trigger features are enabled: " + featuresToUse);
    }

    maxLeftSurface = (props.containsKey("left.surface.context") ? Integer.parseInt(props.getProperty("left.surface.context")) : 1);
    maxRightSurface = (props.containsKey("right.surface.context") ? Integer.parseInt(props.getProperty("right.surface.context")) : 1);
    logger.info("Surface context: [" + maxLeftSurface + ", " + maxRightSurface + "]");
    maxDependency = (props.containsKey("length.dependency.context") ? Integer.parseInt(props.getProperty("length.dependency.context")) : 1);
    logger.info("Dependency context: " + maxDependency);

    String ct = props.getProperty("trigger.classifier");
    if (ct == null) {
      classifierType = CLASSIFIER_TYPE.LR;
    } else if (ct.equalsIgnoreCase("lr")) {
      classifierType = CLASSIFIER_TYPE.LR;
    } else if (ct.equalsIgnoreCase("svm1")) {
      classifierType = CLASSIFIER_TYPE.SVM1;
    } else if (ct.equalsIgnoreCase("svm2")) {
      classifierType = CLASSIFIER_TYPE.SVM2;
    } else if (ct.equalsIgnoreCase("sgd_lr")) {
      classifierType = CLASSIFIER_TYPE.SGD_LR;
    } else {
      throw new RuntimeException("Unknown classifier type: " + ct);
    }

    enableDistSim = Boolean.valueOf(props.getProperty("trigger.distsim.enabled", "false"));

    if (props.containsKey("trigger.distsim.lexicon")) {
      initLexicon(props.getProperty("trigger.distsim.lexicon"));
    }

    if (props.containsKey("trigger.lr.tolerance")) {
      tolerance = Double.parseDouble(props.getProperty("trigger.lr.tolerance"));
      logger.info("Using LR tolerance = " + tolerance);
    } else {
      tolerance = 1e-4;
    }
    if (props.containsKey("trigger.lr.sigma")) {
      sigma = Double.parseDouble(props.getProperty("trigger.lr.sigma"));
      logger.info("Using LR sigma = " + sigma);
    } else {
      sigma = 1.0;
    }

    documentLevelParsing = Boolean.parseBoolean(props.getProperty("document.level.parsing", "false"));
    doErrorAnalysis = Boolean.parseBoolean(props.getProperty("trigger.do.error.analysis", "false"));
    dumpSvmLight = Boolean.parseBoolean(props.getProperty("trigger.dump.svm.light", "false"));
    useFineGrained = Boolean.parseBoolean(props.getProperty("trigger.use.fine.grained", "false"));
    enablePosFeatures = Boolean.parseBoolean(props.getProperty("trigger.enable.pos.features", "false"));
    enableShapeFeatures = Boolean.parseBoolean(props.getProperty("trigger.enable.shape.features", "false"));
    enableCharTrigrams = Boolean.parseBoolean(props.getProperty("trigger.enable.char.trigrams", "false"));
  }
  
  private void initLexicon(String fn) {
    if (enableDistSim) {
      logger.info("Loading distributional similarity lexicon: " + fn);
      lexicon = new HashMap<String, String>();
      File f = new File(fn);
      if (f.exists()) {
        for (String word : ObjectBank.getLineIterator(fn)) {
          String[] bits = word.split("\\s+");
          lexicon.put(bits[0].toLowerCase(), bits[1]);
        }
      }
      logger.info("Done loading.");
    }
  }
  
  private String getDistSim(CoreLabel l) {
    String v = l.get(DistSimAnnotation.class);
    if (v != null)
      return v;

    v = lexicon.get(l.word().toLowerCase());
    if (v == null)
      v = "";
    l.set(DistSimAnnotation.class, v);
    return v;
  }
  
  private boolean useFeature(String s) {
    if (useAllFeatures)
      return true;
    if (featuresToUse == null)
      return true;
    return featuresToUse.contains(s.toLowerCase());
  }
  
  public static class Candidate {
    int position;
    CoreMap sentence;
    String type;

    public Candidate(String t, int p, CoreMap s) {
      this.type = t;
      this.position = p;
      this.sentence = s;
    }

    @Override
    public String toString() {
      StringBuffer os = new StringBuffer();
      List<CoreLabel> toks = sentence.get(TokensAnnotation.class);
      int span = 4;
      int start = Math.max(position - span, 0);
      int end = Math.min(position + span + 1, toks.size());
      for (int i = start; i < end; i++) {
        if (i > start)
          os.append(" ");
        if (i == position) {
          os.append("[[");
        }
        os.append(toks.get(i).word());
        if (i == position) {
          os.append("]]");
        }
      }
      return os.toString();
    }
  }
  
  // no longer used
  public void diffSets(String trainDir, String testDir) throws Exception {
    Annotation trainCorpus = readCorpus(trainDir);
    buildTriggerGazetteer(trainCorpus);
    Annotation testCorpus = readCorpus(testDir);
    List<Candidate> trainCands = extractCandidates(trainCorpus, true);
    List<Candidate> testCands = extractCandidates(testCorpus, true);
    Set<String> trainTriggers = extractHeadWords(trainCands);
    Set<String> testTriggers = extractHeadWords(testCands);
    System.out.printf("Found %d unique train triggers and %d unique test triggers.\n", trainTriggers.size(), testTriggers.size());
    for (String t : testTriggers) {
      if (!trainTriggers.contains(t)) {
        System.out.println("NOT FOUND: " + t);
      }
    }
  }
  
  private Set<String> extractHeadWords(List<Candidate> cands) {
    Set<String> heads = new HashSet<String>();
    for (Candidate cand : cands) {
      heads.add(cand.sentence.get(TokensAnnotation.class).get(cand.position).word().toLowerCase());
    }
    return heads;
  }
  
  private List<String> extractDocIds(List<Candidate> cands) {
    List<String> docids = new ArrayList<String>();
    for (Candidate cand : cands) {
      String docid = cand.sentence.get(DocumentIdAnnotation.class);
      assert (docid != null);
      docids.add(docid);
    }
    return docids;
  }
  
  private List<Integer> findSplitPoints(List<String> docids, int foldCount) {
    List<Integer> splitPoints = new ArrayList<Integer>();
    int foldSize = docids.size() / foldCount;
    logger.info("Finding split points for " + foldCount + " folds and " + docids.size() + " datums.");
    for (int i = 0; i < foldCount; i++) {
      int start = i * foldSize;
      logger.info("Start for fold #" + i + " set initially at " + start);
      if (start > 0 && start < docids.size() - 1) {
        while (start < docids.size() && docids.get(start).equals(docids.get(start - 1)))
          start++;
        logger.info("Start for fold #" + i + " moved to " + start);
      }
      splitPoints.add(start);
    }
    return splitPoints;
  }
  
  private void makeFold(List<Datum<String, String>> filteredAllDatums, 
      int fold, 
      List<Integer> splitPoints, 
      Collection<Datum<String, String >> trainDatums, 
 Collection<Datum<String, String>> testDatums) {
    int start = splitPoints.get(fold);
    int end = filteredAllDatums.size();
    if (fold < splitPoints.size() - 1)
      end = splitPoints.get(fold + 1);
    for (int i = 0; i < filteredAllDatums.size(); i++) {
      if (i < start)
        trainDatums.add(filteredAllDatums.get(i));
      else if (i < end)
        testDatums.add(filteredAllDatums.get(i));
      else
        trainDatums.add(filteredAllDatums.get(i));
    }
  }
  
  private Score evaluateFeatureClassesWithCrossValidation(Set<String> classes, 
      Collection<Datum<String, String>> allDatums,
      List<Integer> splitPoints) {
    List<Datum<String, String>> filteredAllDatums = filterDatums(allDatums, classes);
    for (Datum<String, String> d : filteredAllDatums) {
      logger.finest("DATUM: " + d.label() + " => " + d.asFeatures());
    }

    // this will contain the overall score on all folds
    Score allScores = new Score();

    // train/test on each fold
    double minF1 = Double.MAX_VALUE, maxF1 = Double.MIN_VALUE; // keep track of
                                                               // F1 variation
    for (int fold = 0; fold < splitPoints.size(); fold++) {
      Collection<Datum<String, String>> trainDatums = new ArrayList<Datum<String, String>>();
      Collection<Datum<String, String>> testDatums = new ArrayList<Datum<String, String>>();
      makeFold(filteredAllDatums, fold, splitPoints, trainDatums, testDatums);

      logger.info("Train/test sizes in fold #" + fold + ": " + trainDatums.size() + "/" + testDatums.size());
      logger.info("Training model in fold #" + fold + " using features: " + StringUtils.join(classes, ","));
      classifier = null; // make sure the model is reset (some paranoia is
                         // useful)
      GeneralDataset<String, String> dataset = new Dataset<String, String>();
      dataset.addAll(trainDatums);
      trainModel(dataset);

      logger.info("Evaluating current model in fold #" + fold + "...");
      Score s = evaluate(testDatums, null, true, 0.0, null);
      logger.info("Score in fold #" + fold + " for features " + StringUtils.join(classes, ",") + " is: " + s.toStringShort());
      logger.fine(s.toString());

      double f1 = s.f1();
      if (f1 < minF1)
        minF1 = f1;
      if (f1 > maxF1)
        maxF1 = f1;

      allScores.add(s);
    }

    double f1Var = maxF1 - minF1;
    logger.info("F1 variation for features " + StringUtils.join(classes, ",") + " is: " + f1Var);

    return allScores;
  }
  
  public Set<String> featureSelectionWithCrossValidation(String trainDir, Properties props) throws Exception {
    boolean maximizeRecall = props.containsKey("trigger.feature.selection.max.recall");
    
    logger.info("Reading corpora...");
    Annotation fullCorpus = readCorpus(trainDir);

    // build a gazetteer of known trigger words
    buildTriggerGazetteer(fullCorpus);

    logger.info("Extracting trigger candidates...");
    List<Candidate> allCands = extractCandidates(fullCorpus, true);
    List<String> docids = extractDocIds(allCands);
    assert (docids.size() == allCands.size());
    List<Integer> splitPoints = findSplitPoints(docids, 10);
    assert (splitPoints.size() == 10);
    fullCorpus = null;

    logger.info("Generating ALL features...");
    Collection<Datum<String, String>> allDatums = toDatums(allCands);
    logger.info("Found " + allDatums.size() + " training datums.");
    allCands = null;

    logger.info("Extracting feature classes...");
    Set<String> allFeatureClasses = extractFeatureClasses(allDatums);
    logger.info("Found " + allFeatureClasses.size() + " feature classes:");
    logger.info(StringUtils.join(allFeatureClasses, ","));

    Set<String> selectedFeatureClasses = new HashSet<String>();
    Score bestScore = new Score();
    if (true) {
      selectedFeatureClasses = new HashSet<String>(Arrays.asList("token"));
      bestScore = evaluateFeatureClassesWithCrossValidation(selectedFeatureClasses, allDatums, splitPoints);
    }
    logger.info("Initial score is: " + bestScore.toStringShort());

    while (selectedFeatureClasses.size() < allFeatureClasses.size()) {
      logger.info("Curently selected feature set is: " + StringUtils.join(selectedFeatureClasses, ","));
      String iterBestFeat = null;
      Score iterBestScore = bestScore;

      double oldScore = maximizeRecall ? iterBestScore.r() : iterBestScore.f1();
      for (String feat : allFeatureClasses) {
        if (selectedFeatureClasses.contains(feat))
          continue;

        selectedFeatureClasses.add(feat);
        Score s = evaluateFeatureClassesWithCrossValidation(selectedFeatureClasses, allDatums, splitPoints);
        
        double newScore = maximizeRecall ? s.r() : s.f1();
        if (newScore > oldScore) {
          logger.info("Found better score for feature " + feat + ": " + s.toStringShort());
          iterBestScore = s;
          iterBestFeat = feat;
        }
        selectedFeatureClasses.remove(feat);
      }

      if (iterBestFeat != null) {
        assert (oldScore > bestScore.f1());
        logger.info("Iteration #" + selectedFeatureClasses.size() + ": chose feature class " + iterBestFeat);
        logger.info("Iteration #" + selectedFeatureClasses.size() + ": current score is " + iterBestScore.toStringShort());
        selectedFeatureClasses.add(iterBestFeat);
        logger.info("Iteration #" + (selectedFeatureClasses.size() - 1) + ": current feature set is "
            + StringUtils.join(selectedFeatureClasses, ","));
        bestScore = iterBestScore;
      } else {
        logger.info("Iteration #" + selectedFeatureClasses.size() + ": failed to improve. Stopping.");
        break;
      }
    }

    return selectedFeatureClasses;
  }
  
  public Set<String> featureSelection(String trainDir, String testDir) throws Exception {
    logger.info("Reading corpora...");
    Annotation trainCorpus = readCorpus(trainDir);
    Annotation testCorpus = readCorpus(testDir);
    
    // build a gazetteer of known trigger words
    buildTriggerGazetteer(trainCorpus);
    
    logger.info("Extracting trigger candidates...");
    List<Candidate> trainCands = extractCandidates(trainCorpus, true);
    List<Candidate> testCands = extractCandidates(testCorpus, true);
    trainCorpus = null; testCorpus = null;
    
    logger.info("Generating ALL features...");
    Collection<Datum<String, String>> trainDatums = toDatums(trainCands);
    Collection<Datum<String, String>> testDatums = toDatums(testCands);
    logger.info("Found " + trainDatums.size() + " training datums and " + testDatums.size() + " test datums.");
    trainCands = null; testCands = null;
    
    logger.info("Extracting feature classes...");
    Set<String> allFeatureClasses = extractFeatureClasses(trainDatums);
    logger.info("Found " + allFeatureClasses.size() + " feature classes:");
    logger.info(StringUtils.join(allFeatureClasses, ","));
    
    Set<String> selectedFeatureClasses = new HashSet<String>();
    Score bestScore = new Score();
    if (true) {
      selectedFeatureClasses = new HashSet<String>(Arrays.asList(
          "token", // P 54.68 R 24.56 F1 33.89
          "firstdep" // P 71.91 R 46.44 F1 56.43
      ));
      bestScore = evaluateFeatureClasses(selectedFeatureClasses, trainDatums, testDatums);
    }
    logger.info("Initial score is: " + bestScore.toStringShort());
    
    while (selectedFeatureClasses.size() < allFeatureClasses.size()) {
      logger.info("Curently selected feature set is: " + StringUtils.join(selectedFeatureClasses, ","));
      String iterBestFeat = null;
      Score iterBestScore = bestScore;

      for (String feat : allFeatureClasses) {
        if (selectedFeatureClasses.contains(feat))
          continue;

        selectedFeatureClasses.add(feat);
        Score s = evaluateFeatureClasses(selectedFeatureClasses, trainDatums, testDatums);
        if (s.f1() > iterBestScore.f1()) {
          logger.info("Found better score for feature " + feat + ": " + s.toStringShort());
          iterBestScore = s;
          iterBestFeat = feat;
        }
        selectedFeatureClasses.remove(feat);
      }

      if (iterBestFeat != null) {
        assert (iterBestScore.f1() > bestScore.f1());
        logger.info("Iteration #" + selectedFeatureClasses.size() + ": chose feature class " + iterBestFeat);
        logger.info("Iteration #" + selectedFeatureClasses.size() + ": current score is " + iterBestScore.toStringShort());
        selectedFeatureClasses.add(iterBestFeat);
        logger.info("Iteration #" + (selectedFeatureClasses.size() - 1) + ": current feature set is "
            + StringUtils.join(selectedFeatureClasses, ","));
        bestScore = iterBestScore;
      } else {
        logger.info("Iteration #" + selectedFeatureClasses.size() + ": failed to improve. Stopping.");
        break;
      }
    }
    
    return selectedFeatureClasses;
  }
  
  private Score evaluateFeatureClasses(Set<String> classes, 
      Collection<Datum<String, String>> trainDatums,
      Collection<Datum<String, String>> testDatums) {
    Collection<Datum<String, String>> filteredTrainDatums = filterDatums(trainDatums, classes);
    for (Datum<String, String> d : filteredTrainDatums) {
      logger.finest("TRAIN DATUM: " + d.label() + " => " + d.asFeatures());
    }
    Collection<Datum<String, String>> filteredTestDatums = filterDatums(testDatums, classes);

    logger.fine("Training model using features: " + StringUtils.join(classes, ","));
    classifier = null; // make sure the model is reset (some paranoia is useful)
    GeneralDataset<String, String> dataset = new Dataset<String, String>();
    dataset.addAll(filteredTrainDatums);
    trainModel(dataset);

    logger.fine("Evaluating current model...");
    Score s = evaluate(filteredTestDatums, null, true, 0.0, null);
    logger.fine("Score for features " + StringUtils.join(classes, ",") + " is: " + s.toStringShort());
    logger.fine(s.toString());
    return s;
  }
  
  private List<Datum<String, String>> filterDatums(Collection<Datum<String, String>> datums, Set<String> classes) {
    List<Datum<String, String>> filteredDatums = new ArrayList<Datum<String, String>>();
    for (Datum<String, String> d : datums) {
      Collection<String> feats = d.asFeatures();
      String l = d.label();
      Collection<String> filteredFeats = new ArrayList<String>();
      for (String f : feats) {
        String[] toks = f.split(":");
        if (classes.contains(toks[0])) {
          filteredFeats.add(f);
        }
      }
      Datum<String, String> fd = new BasicDatum<String, String>(filteredFeats, l);
      filteredDatums.add(fd);
    }
    return filteredDatums;
  }
  
  public Set<String> extractFeatureClasses(Collection<Datum<String, String>> trainDatums) {
    Set<String> classes = new HashSet<String>();
    for (Datum<String, String> datum : trainDatums) {
      Collection<String> feats = datum.asFeatures();
      for (String f : feats) {
        String[] toks = f.split(":");
        classes.add(toks[0]);
      }
    }
    return classes;
  }
  
  public Score[] test(String dir, boolean saveFiles) throws Exception {
    // read the whole corpus in memory
    Annotation corpus = readCorpus(dir);
    return testCorpus(corpus, dir, saveFiles);
  }
  
  private Score[] testCorpus(Annotation corpus, String dir, boolean saveFiles) throws Exception {
    boolean hasGold = corpus.get(HasGoldAnnotation.class);
    if (hasGold)
      logger.info("Found gold annotations in the test corpus.");
    else
      logger.info("Did not find gold annotations in the test corpus.");

    // build +/- candidates
    logger.info("Extracting trigger candidates (testing)...");
    List<Candidate> cands = extractCandidates(corpus, hasGold);
    corpus = null;
    logger.info("Generating features (testing)...");
    Collection<Datum<String, String>> datums = toDatums(cands);
    assert (datums.size() == cands.size());
    datums = filterDatums(datums, featuresToUse);
    for (Datum<String, String> d : datums) {
      logger.finest("TEST DATUM: " + d.label() + " => " + d.asFeatures());
    }

    if (dumpSvmLight) {
      saveSvmLight(datums, "test.trigger");
    }

    logger.info("Evaluating...");
    // return evaluate(datums, cands, hasGold, 0);
    Score[] scores = new Score[10];
    int scorePosition = 0;
    for (int beam = 0; beam < 100; beam += 10) {
      List<Triple<String, Double, Double>> predictions = new ArrayList<Triple<String, Double, Double>>();
      Score s = evaluate(datums, cands, hasGold, ((double) beam) / 100.00, predictions);
      if (s != null) {
        logger.info("Score for beam " + beam + " is: " + s.toStringShort());
        logger.info(s.toString());
      }
      scores[scorePosition++] = s;

      // save files with triggers
      if (saveFiles) {
        if (classifier instanceof SVMLightClassifier<?, ?>) {
          savePredictionsForStacking(cands, predictions);
        } else {
          saveFiles(cands, predictions, beam);
        }
      }
    }
    return scores;
  }
  
  private void savePredictionsForStacking(List<Candidate> cands, List<Triple<String, Double, Double>> predictions) {
    assert (predictions.size() == cands.size());
    Set<Pair<String, String>> docIds = new HashSet<Pair<String, String>>();
    for (Candidate c : cands) {
      String docid = c.sentence.get(DocumentIdAnnotation.class);
      assert (docid != null);
      String path = c.sentence.get(DocumentDirectoryAnnotation.class);
      assert (path != null);
      docIds.add(new Pair<String, String>(docid, path));
    }

    // create one file with predicted triggers for each docid
    for (Pair<String, String> docid : docIds) {
      String fn = docid.second + File.separator + docid.first + ".triggers.svm";
      try {
        logger.fine("Saving predicted triggers to " + fn + "...");
        PrintStream os = new PrintStream(new FileOutputStream(fn));
        MutableInteger tid = new MutableInteger(1000);
        for (int i = 0; i < cands.size(); i++) {
          if (docid.first.equals(cands.get(i).sentence.get(DocumentIdAnnotation.class))) {
            savePredictedTrigger(os, cands.get(i), predictions.get(i), tid);
          }
        }
        os.close();
      } catch (IOException e) {
        System.err.println("ERROR: cannot save trigger file " + fn);
        e.printStackTrace();
        System.exit(1);
      }
    }
  }
  
  private void saveFiles(List<Candidate> cands, List<Triple<String, Double, Double>> predictions, int beam) {
    assert (predictions.size() == cands.size());
    Set<Pair<String, String>> docIds = new HashSet<Pair<String, String>>();
    for (Candidate c : cands) {
      String docid = c.sentence.get(DocumentIdAnnotation.class);
      assert (docid != null);
      String path = c.sentence.get(DocumentDirectoryAnnotation.class);
      assert (path != null);
      docIds.add(new Pair<String, String>(docid, path));
    }

    // create one file with predicted triggers for each docid
    for (Pair<String, String> docid : docIds) {
      String filename = dataset.getFilesystemInformation().getTriggerFilename(docid.second(), docid.first(), beam);
      try {
        logger.severe("Saving predicted triggers to " + filename + "...");
        PrintStream os = new PrintStream(new FileOutputStream(filename));
        MutableInteger tid = new MutableInteger(1000);
        for (int i = 0; i < cands.size(); i++) {
          if (docid.first.equals(cands.get(i).sentence.get(DocumentIdAnnotation.class))) {
            savePredictedTrigger(os, cands.get(i), predictions.get(i), tid);
          }
        }
        os.close();
      } catch (IOException e) {
        System.err.println("ERROR: cannot save trigger file " + filename);
        e.printStackTrace();
        System.exit(1);
      }
    }
  }
     
  private void savePredictedTrigger(PrintStream os, Candidate cand, Triple<String, Double, Double> pred, MutableInteger tid) {
    if (pred.first.equals("O"))
      return;
    List<CoreLabel> tokens = cand.sentence.get(TokensAnnotation.class);
    int charStart = tokens.get(cand.position).get(CharacterOffsetBeginAnnotation.class);
    int charEnd = tokens.get(cand.position).get(CharacterOffsetEndAnnotation.class);
    String text = tokens.get(cand.position).word();
    os.printf("T%d\t%s %d %d %s\t%f %f\n", tid.intValue(), pred.first, charStart, charEnd, text, pred.second,
        (pred.third != null ? pred.third : 0));
    tid.incValue(1);
  }
  
  private Score evaluate(Collection<Datum<String, String>> datums, 
      List<Candidate> cands, 
      boolean annotated,
      double beam,
      List<Triple<String, Double, Double>> predictions) {
    Score score = null;
    int candPos = 0;
    for (Datum<String, String> datum : datums) {
      Candidate cand = null;
      if (cands != null)
        cand = cands.get(candPos++);
      Triple<String, Double, Double> prediction = getPrediction(datum, beam);
      if (predictions != null)
        predictions.add(prediction);
      if (annotated) {
        if (score == null)
          score = new Score();
        if (!datum.label().equals("O"))
          score.incTotal(datum.label());
        if (!prediction.first.equals("O")) {
          score.incPredicted(prediction.first);
          if (prediction.first.equals(datum.label())) {
            score.incCorrect(datum.label());
          }
        }

        // some error analysis for the worst classes
        if (doErrorAnalysis && labelsForErrorAnalysis.contains(datum.label()) && !datum.label().equals(prediction.first)) {
          logger.info("RECALL error: gold/predicted " + datum.label() + "/" + prediction);
          logger.info("Candidate: " + cand);
        }
      }
    }
    return score;
  }

  private Triple<String, Double, Double> getPrediction(Datum<String, String> datum, double beam) {
    if (classifier instanceof LinearClassifier<?, ?>) {
      LinearClassifier<String, String> lc = (LinearClassifier<String, String>) classifier;
      Counter<String> probs = lc.probabilityOf(datum);
      List<Pair<String, Double>> sorted = Counters.toSortedListWithCounts(probs);
      logger.finest("Predictions: " + sorted);
      if (!sorted.get(0).first.equals("O")) {
        return new Triple<String, Double, Double>(sorted.get(0).first, sorted.get(0).second, null);
      } else {
        double diff = sorted.get(0).second - sorted.get(1).second;
        if (diff < beam) {
          return new Triple<String, Double, Double>(sorted.get(1).first, sorted.get(1).second, sorted.get(0).second);
        } else {
          return new Triple<String, Double, Double>(sorted.get(0).first, sorted.get(0).second, null);
        }
      }
    } else {
      // TODO: add probability predictions for SVM classifier
      // For now it's fine, because we use only basic SVM predictions for
      // stacking
      return new Triple<String, Double, Double>(classifier.classOf(datum), 0.9999, null);
    }
  }
  
  @SuppressWarnings("unchecked")
  Pair<Annotation, Annotation>[] makeCrossValidationSplits(Annotation corpus, int foldCount) {
    Pair<Annotation, Annotation>[] folds = new Pair[foldCount];
    List<CoreMap> sentences = corpus.get(SentencesAnnotation.class);
    boolean hasGold = corpus.get(HasGoldAnnotation.class);
    int foldSize = sentences.size() / foldCount;
    for (int fold = 0; fold < foldCount; fold++) {
      int start = foldSize * fold;
      int end = foldSize * (fold + 1);
      logger.fine("Using start=" + start + " and end=" + end + " fold #" + fold + ".");

      Annotation train = new Annotation("");
      Annotation test = new Annotation("");
      train.set(HasGoldAnnotation.class, hasGold);
      test.set(HasGoldAnnotation.class, hasGold);

      for (int i = 0; i < start; i++) {
        AnnotationUtils.addSentence(train, sentences.get(i));
      }
      for (int i = start; i < end; i++) {
        AnnotationUtils.addSentence(test, sentences.get(i));
      }
      for (int i = end; i < sentences.size(); i++) {
        AnnotationUtils.addSentence(train, sentences.get(i));
      }

      folds[fold] = new Pair<Annotation, Annotation>(train, test);
    }
    return folds;
  }
  
  public void crossValidation(String dir, boolean saveFiles) throws Exception {
    // read the whole corpus in memory
    Annotation corpus = readCorpus(dir);

    // generate the 10 versions of train/test for cross validation
    Pair<Annotation, Annotation>[] splits = makeCrossValidationSplits(corpus, 10);
    Score[] scores = new Score[10];
    for (int i = 0; i < scores.length; i++)
      scores[i] = new Score();
    // call trainCorpus, testCorpus in each fold
    for (int fold = 0; fold < splits.length; fold++) {
      logger.info("Training fold #" + fold + "...");
      logger.info("Fold #" + fold + " contains " +
          splits[fold].first.get(SentencesAnnotation.class).size() + 
          " training sentences and " +
          splits[fold].second.get(SentencesAnnotation.class).size() + 
          " testing sentences.");
      
      trainCorpus(splits[fold].first, dir);
      logger.info("Evaluating fold #" + fold + "...");
      Score[] foldScores = testCorpus(splits[fold].second, dir, saveFiles);
      assert (foldScores.length == scores.length);
      for (int i = 0; i < foldScores.length; i++) {
        scores[i].add(foldScores[i]);
      }
    }

    // report overall scores
    int beam = 0;
    for (int i = 0; i < scores.length; i++, beam += 10) {
      logger.info("OVERALL CROSS-VALIDATION SCORE for beam " + beam + " is: " + scores[i].toStringShort());
      logger.info(scores[i].toString());
    }
  }
  
  public void train(String dir) throws Exception {
    // read the whole corpus in memory
    Annotation corpus = readCorpus(dir);
    trainCorpus(corpus, dir);
  }
   
  private void trainCorpus(Annotation corpus, String dir) throws Exception {
    // build a gazetteer of known trigger words
    buildTriggerGazetteer(corpus);

    // build +/- candidates
    logger.info("Extracting trigger candidates (training)...");
    List<Candidate> cands = extractCandidates(corpus, true);
    corpus = null;
    logger.info("Generating features (training)...");
    Collection<Datum<String, String>> datums = toDatums(cands);
    datums = filterDatums(datums, featuresToUse);
    for (Datum<String, String> d : datums) {
      logger.finest("TRAIN DATUM: " + d.label() + " => " + d.asFeatures());
    }
    GeneralDataset<String, String> dataset = new Dataset<String, String>();
    dataset.addAll(datums);
    featureIndex = dataset.featureIndex();
    labelIndex = dataset.labelIndex();
    System.err.println("Label index has " + labelIndex.size() + " labels: " + labelIndex);
    cands = null;

    if (dumpSvmLight) {
      saveSvmLight(datums, "train.trigger");
      logger.info("Saved training corpus in SVM light format in train.trigger.");
    }

    // train
    logger.info("Training model...");
    trainModel(dataset);
  }
  
  private void saveSvmLight(Collection<Datum<String, String>> datums, String fn) {
    try {
      PrintStream os = new PrintStream(new FileOutputStream(fn));
      for (Datum<String, String> d : datums) {
        int label = labelIndex.indexOf(d.label());
        if (label < 0) {
          System.err.println("Skipping unknown label: " + d.label());
          continue;
        }
        Counter<String> feats = Counters.asCounter(d.asFeatures());
        Counter<Integer> ifeats = new ClassicCounter<Integer>();
        for (String f : feats.keySet()) {
          int i = featureIndex.indexOf(f);
          if (i >= 0) {
            ifeats.incrementCount(i, feats.getCount(f));
          }
        }
        List<Integer> sortedFeats = new ArrayList<Integer>(ifeats.keySet());
        Collections.sort(sortedFeats);
        os.print(label);
        for (Integer f : sortedFeats) {
          if (f >= 0) {
            os.print(" " + f + ":" + ifeats.getCount(f));
          }
        }
        os.println();
      }
      os.close();
    } catch (IOException e) {
      System.err.println("Cannot save datums to " + fn);
      e.printStackTrace();
      System.exit(1);
    }
  }
  
  @SuppressWarnings("unchecked")
  private void trainModel(GeneralDataset<String, String> dataset) {
    if (classifierType == CLASSIFIER_TYPE.LR) {
      LinearClassifierFactory<String, String> factory = new LinearClassifierFactory<String, String>(new QNMinimizer(15), tolerance, false,
          sigma);
      classifier = factory.trainClassifier(dataset);
    } else if (classifierType == CLASSIFIER_TYPE.SVM1) {
      // classifier = LibSVMClassifier.trainLinearSVM(dataset, 1.0);
      SVMLightClassifierFactory<String, String> factory = new SVMLightClassifierFactory<String, String>();
      factory.setC(sigma); // sigma actually stores the C values for SVM
      factory.setUseSigmoid(true);
      factory.setDeleteTempFilesOnExitFlag(false);
      classifier = factory.trainClassifier(dataset);
    } else if (classifierType == CLASSIFIER_TYPE.SVM2) {
      throw new RuntimeException("Classifier type " + CLASSIFIER_TYPE.SVM2 + " not supported anymore!");
      // not really supported anymore
      // classifier = LibSVMClassifier.trainPolynomialSVM(dataset, 1 /* gamma */, 0 /* coef0 */, 2 /* degree */, 1 /* C */);
    } else if (classifierType == CLASSIFIER_TYPE.SGD_LR) {
      // TODO HybridMinimizer for SGN->QN should be packaged better
      
      int sgdPasses = 50;
      int tuneSampleSize = 1000;

      Minimizer<DiffFunction> sgdMinimizer = new StochasticInPlaceMinimizer(sigma, sgdPasses, tuneSampleSize);
      Minimizer<DiffFunction> qnMinimizer = new QNMinimizer(15);
      Minimizer<DiffFunction> hybridMinimizer = new HybridMinimizer(sgdMinimizer, qnMinimizer, sgdPasses);

      LinearClassifierFactory<String, String> factory = new LinearClassifierFactory<String, String>(hybridMinimizer, tolerance, false,
          LogPrior.LogPriorType.QUADRATIC.ordinal(), sigma);
      classifier = factory.trainClassifier(dataset);
    }
  }
  
  private Collection<Datum<String, String>> toDatums(List<Candidate> cands) {
    Collection<Datum<String, String>> datums = new ArrayList<Datum<String, String>>();
    for (Candidate cand : cands) {
      Datum<String, String> d = toDatum(cand);
      datums.add(d);
    }
    return datums;
  }

  @SuppressWarnings("unused")
  private GeneralDataset<String, String> toDataset(List<Candidate> cands) {
    GeneralDataset<String, String> dataset = new Dataset<String, String>();
    for (Candidate cand : cands) {
      Datum<String, String> d = toDatum(cand);
      dataset.add(d);
    }
    return dataset;
  }
  
  private Datum<String, String> toDatum(Candidate cand) {
    Collection<String> feats = generateFeatures(cand);
    Datum<String, String> datum = new BasicDatum<String, String>(feats, cand.type);
    return datum;
  }
  
  public static class EntSigAnnotation implements CoreAnnotation<String[]> {
    public Class<String[]> getType() {
      return ErasureUtils.<Class<String[]>> uncheckedCast(List.class);
    }
  }
  
  private String [] getEntitySignature(CoreMap sent) {
    if (sent.get(EntSigAnnotation.class) != null) {
      return sent.get(EntSigAnnotation.class);
    } else {
      List<CoreLabel> tokens = sent.get(TokensAnnotation.class);
      String[] sig = new String[tokens.size()];
      for (int i = 0; i < sig.length; i++)
        sig[i] = "O";
      if (sent.get(EntityMentionsAnnotation.class) != null) {
        for (EntityMention e : sent.get(EntityMentionsAnnotation.class)) {
          boolean isTrigger = isTrigger(e);
          //int headPos = isTrigger(e) ? BioNLPToCoNLL.triggerHead(e, tokens) : BioNLPToCoNLL.entityHead(e, tokens);
          //sig[headPos] = e.getType();
          if (!isTrigger) {
            int headPos = BioNLPFormatReader.entityHead(e, tokens);
            sig[headPos] = e.getType();
          }
        }

        logger.finest("Found following signature for sentence: " + sent.get(TextAnnotation.class));
        logger.finest("Containing entities: " + sent.get(EntityMentionsAnnotation.class));
        logger.finest("Signature: " + Arrays.asList(sig));
      }
      sent.set(EntSigAnnotation.class, sig);
      return sig;
    }
  }
  
  private Collection<String> generateFeatures(Candidate cand) {
    Collection<String> features = new ArrayList<String>();
    List<CoreLabel> tokens = cand.sentence.get(TokensAnnotation.class);
    String [] entSig = getEntitySignature(cand.sentence);
    CoreLabel token = tokens.get(cand.position);
    
    addTokenFeatures("token", null, token, null, features);
    addSurfaceContextFeatures(maxLeftSurface, maxRightSurface, cand.position, tokens, entSig, features);
    addSurfaceNGramFeatures(2, cand.position, tokens, entSig, features);

    //
    // bag of word features
    //
    addBagOfWords("full", tokens, 0, tokens.size(), features);
    for (int i = 1; i <= 5; i++) {
      addBagOfWords("window-" + i, tokens, Math.max(0, cand.position - i), Math.min(tokens.size(), cand.position + i + 1), features);
    }
    
    //
    // entity count features
    //
    for (int i = 1; i <= 5; i++) {
      addEntityCounts("window-" + i, tokens, entSig, Math.max(0, cand.position - i), Math.min(tokens.size(), cand.position + i + 1),
          features);
    }
    addEntityCounts("full", tokens, entSig, 0, tokens.size(), features);
    
    // syntactic features
    SemanticGraph sg = getSemanticGraph(cand.sentence);
    addDependencyFeatures(cand.sentence.get(TextAnnotation.class), cand, tokens, entSig, sg, features);
    
    return features;
  }
  
  private void addDependencyFeatures(String text, Candidate cand, List<CoreLabel> tokens, String[] entSig, SemanticGraph graph,
      Collection<String> features) {
    logger.finest("Sentence text: " + text);
    logger.finest("My position: " + cand.position + " for word " + tokens.get(cand.position).word());

    IndexedWord myNode = null;
    try {
      myNode = graph.getNodeByIndex(cand.position + 1);
    } catch (IllegalArgumentException e) {
      // this is fine: some nodes may get collapsed in the dependency graph
      return;
    }
    assert (myNode != null);
    CoreLabel myToken = tokens.get(cand.position);

    HashSet<String> seenFirstDeps = new HashSet<String>();
    HashSet<String> seenFullPaths = new HashSet<String>();
    
    // info about the head token 
    IndexedWord headNode = graph.getParent(myNode);
    if(headNode == null){
      features.add("dephead:NIL");
      logger.finest("Head is NIL.");
    } else {
      int headPos = headNode.index() - 1;
      CoreLabel headToken = tokens.get(headPos);
      logger.finest("Head indexedWord is: " + headNode.toString());
      logger.finest("Head position is " + headPos + ". Head token is " + headToken.word());
      addTokenFeatures("dephead", null, headToken, entSig[headPos], features);
    }

    // compute dependency paths to all other nodes in the corpus
    for (int dest = 0; dest < tokens.size(); dest++) {
      if (dest == cand.position)
        continue;
      IndexedWord destNode = null;
      try {
        destNode = graph.getNodeByIndex(dest + 1);
        CoreLabel destToken = tokens.get(dest);
        logger.finest("Finding shortest path between " + myToken.word() + "(" + cand.position + ") and " + destToken.word() + "(" + dest
            + ") in sentence " + text);
        List<SemanticGraphEdge> edgePath = graph.getShortestPathEdges(myNode, destNode);
        if (edgePath == null) {
          // this happens for tokens like parentheses
          // not sure why, but let's just move on
          continue;
        }

        List<String> specificRelations = RelationFeatureFactory.dependencyPathAsList(edgePath, myNode, false);
        logger.finest("Dependency path between " + myToken.word() + " and " + destToken.word() + ": " + specificRelations);

        // only look at dependencies shorter or equal that a max
        if (specificRelations.size() <= maxDependency && specificRelations.size() > 0) {
          // full path
          String prefix = "deppath-" + specificRelations.size();
          String suffix = StringUtils.join(specificRelations, "+");
          addTokenFeatures(prefix, suffix, destToken, entSig[dest], features);

          if (!seenFullPaths.contains(suffix)) {
            addTokenFeatures("srcdeppath-" + specificRelations.size(), suffix, myToken, entSig[cand.position], features);
            seenFullPaths.add(suffix);
          }

          // collapsed path
          if (specificRelations.size() >= 1) {
            suffix = specificRelations.get(0);
            addTokenFeatures("firstdep", suffix, destToken, entSig[dest], features);

            if (!seenFirstDeps.contains(suffix)) {
              addTokenFeatures("firstdepsrc", suffix, myToken, entSig[cand.position], features);
              seenFirstDeps.add(suffix);
            }

            suffix = specificRelations.get(specificRelations.size() - 1);
            addTokenFeatures("lastdep", suffix, destToken, entSig[dest], features);
          }

          // add combo features if syntactic distance is 1 or 2
          if (specificRelations.size() <= 2) {
            String path = StringUtils.join(specificRelations, "+");
            addDoubleTokenDistanceFeatures("synpathcombo-" + specificRelations.size(), path, myToken, destToken, entSig[cand.position],
                entSig[dest], features);
          }
        }
      } catch (IllegalArgumentException e) {
        // this is fine: some nodes get collapsed
      }
    }
  }
  
  public static class DependencyGraphAnnotation implements CoreAnnotation<SemanticGraph> {
    public Class<SemanticGraph> getType() {
      return SemanticGraph.class;
    }
  }

  private SemanticGraph getSemanticGraph(CoreMap sent) {
    if (sent.containsKey(DependencyGraphAnnotation.class)) {
      return sent.get(DependencyGraphAnnotation.class);
    }

    Tree tree = sent.get(TreeAnnotation.class);
    // collapsed version
    SemanticGraph g = SemanticGraphFactory.makeFromTree(tree, true, true, true, false, true);
    sent.set(DependencyGraphAnnotation.class, g);
    return g;
  }
  
  private void addEntityCounts(String prefix, List<CoreLabel> tokens, String[] entSig, int start, int end, Collection<String> features) {
    int count = 0;
    for (int i = start; i < end; i++) {
      if (!entSig.equals("O") && !isTrigger(entSig[i])) {
        count++;
        String name = prefix + ".entityfound";
        if (useFeature(name)) {
          features.add(name);
        }
      }
    }
    String name = prefix + ".entitycount";
    if (useFeature(name)) {
      features.add(name + ":" + Integer.toString(count));
    }
  }

  private void addBagOfWords(String prefix, List<CoreLabel> tokens, int start, int end, Collection<String> features) {
    String name = prefix + ".bow";
    if (useFeature(name)) {
      for (int i = start; i < end; i++) {
        String w = tokens.get(i).word();
        features.add(name + ":" + w);
      }
    }
  }
  
  private void addSurfaceNGramFeatures(int maxSpan, int position, List<CoreLabel> tokens, String[] entSig, Collection<String> features) {
    CoreLabel myToken = tokens.get(position);

    // combo with previous tokens
    for (int i = Math.max(position - maxSpan, 0); i < position; i++) {
      String prefix = "prevcombo-" + (position - i);
      CoreLabel destToken = tokens.get(i);
      addDoubleTokenDistanceFeatures(prefix, null, destToken, myToken, entSig[i], entSig[position], features);
    }

    // combo with following tokens
    for (int i = position + 1; i < Math.min(position + maxSpan + 1, tokens.size()); i++) {
      String prefix = "nextcombo-" + (i - position);
      CoreLabel destToken = tokens.get(i);
      addDoubleTokenDistanceFeatures(prefix, null, myToken, destToken, entSig[position], entSig[i], features);
    }
  }
   
  private void addSurfaceContextFeatures(int leftMax, int rightMax, int position, List<CoreLabel> tokens, String[] entSig,
      Collection<String> features) {
    for (int i = position - 1; i >= 0 && i >= position - leftMax; i--) {
      addTokenFeatures("prev" + (position - i), null, tokens.get(i), entSig[i], features);
    }
    for (int i = position + 1; i < tokens.size() && i <= position + rightMax; i++) {
      addTokenFeatures("next" + (i - position), null, tokens.get(i), entSig[i], features);
    }
  }

  private String[][] extractTokenFeaturesForCombo(CoreLabel token, String entLabel) {
    String word = token.word();
    String pos = null; // token.tag();
    String distsim = null; // getDistSim(token);
    return new String[][] { { "lemma", dataset.lemmatizeWordForTriggerClassification(word, pos) }, { "pos", pos },
        { "ent", (!isTrigger(entLabel) && !entLabel.equals("O") ? entLabel : null) }, { "distsim", distsim } };
  }
  
  private void addDoubleTokenDistanceFeatures(String prefix, String path, CoreLabel srcToken, CoreLabel dstToken, String srcEnt,
      String dstEnt, Collection<String> features) {
    String[][] src = extractTokenFeaturesForCombo(srcToken, srcEnt);
    String[][] dst = extractTokenFeaturesForCombo(dstToken, dstEnt);
    if (path == null)
      path = "";

    for (int i = 0; i < src.length; i++) {
      assert (src[i].length == 2);
      if (src[i][1] == null)
        continue;
      for (int j = 0; j < dst.length; j++) {
        assert (dst[j].length == 2);
        if (dst[j][1] == null)
          continue;

        String name, value;
        if (useFineGrained) {
          name = prefix + "X" + src[i][0] + "X" + dst[j][0];
          value = path + "X" + src[i][1] + "X" + dst[j][1];
        } else {
          // keep distsim fine-grained so we can toggle it
          if (src[i][0].equals("distsim") && dst[j][0].equals("distsim")) {
            name = prefix + "X" + src[i][0] + "X" + dst[j][0];
            value = path + "X" + src[i][1] + "X" + dst[j][1];
          } else if (src[i][0].equals("distsim")) {
            name = prefix + "X" + src[i][0] + "X";
            value = path + "X" + src[i][1] + "X" + dst[j][0] + "-" + dst[j][1];
          } else if (dst[j][0].equals("distsim")) {
            name = prefix + "XX" + dst[j][0];
            value = path + "X" + src[i][0] + "-" + src[i][1] + "X" + dst[j][1];
          } else {
            name = prefix + "XX";
            value = path + "X" + src[i][0] + "-" + src[i][1] + "X" + dst[j][0] + "-" + dst[j][1];
          }
        }
        if (useFeature(name))
          features.add(name + ":" + value);
      }
    }
  }
  
  private void addTokenFeatures(String prefix, String suffix, CoreLabel token, String entLabel, Collection<String> features) {
    String word = token.word();
    String pos = token.tag();
    String lemma = dataset.lemmatizeWordForTriggerClassification(word, pos);
    String name, value;
    
    /*
    // lemma generalizes better than word. let's keep just that
    String name = prefix + ".word";
    String value = word + (suffix != null ? "-" + suffix : "");
    if(useFeature(name)) features.add(name + ":" + value);
    */
    
    // Character n-grams
    if (enableCharTrigrams) {
      addCharacterNgramFeatures(prefix, suffix, word, features);
    }
    
    // LEMMA
    name = prefix;
    if (useFineGrained) {
      name += ".lemma";
      value = "";
    } else {
      value = "lemma.";
    }
    value += lemma + (suffix != null ? "-" + suffix : "");
    if (useFeature(name))
      features.add(name + ":" + value);
    
    // POS
    if (enablePosFeatures) {
      name = prefix;
      if (useFineGrained) {
        name += ".pos";
        value = "";
      } else {
        value = "pos.";
      }
      value += pos + (suffix != null ? "-" + suffix : "");
      if (useFeature(name))
        features.add(name + ":" + value);
    }
    
    // SHAPE
    if (enableShapeFeatures) {
      /*
      // no reason to have two shape classifiers
      name = prefix + ".shape";
      if(useFeature(name)){
        String shape = WordShapeClassifier.wordShape(word, WordShapeClassifier.WORDSHAPECHRIS2);
        value = shape + (suffix != null ? "-" + suffix : "");
        logger.finest("CHRIS2 shape for " + word + ": " + lemma + ", " + shape);
        features.add(name + ":" + value);
      }
      */
      name = prefix;
      if (useFineGrained) {
        name += ".shape";
        value = "";
      } else {
        value = "shape.";
      }
      if (useFeature(name)) {
        String shape = WordShapeClassifier.wordShape(word, WordShapeClassifier.WORDSHAPEDAN2BIO);
        value += shape + (suffix != null ? "-" + suffix : "");
        logger.finest("DAN2BIO shape for " + word + ": " + lemma + ", " + shape);
        features.add(name + ":" + value);
      }
    }

    // ENTITY LABEL
    name = prefix;
    if (useFineGrained) {
      name += ".ent";
      value = "";
    } else {
      value = "ent.";
    }
    if (entLabel != null && !entLabel.equals("O") && !isTrigger(entLabel) && useFeature(name)) {
      value += entLabel + (suffix != null ? "-" + suffix : "");
      features.add(name + ":" + value);
    }
    
    // GAZETTEER LABEL
    name = prefix;
    if (useFineGrained) {
      name += ".gaz";
      value = "";
    } else {
      value = "gaz";
    }
    if (useFeature(name) && knownTriggers.containsKey(word.toLowerCase())) {
      Set<String> labels = knownTriggers.get(word.toLowerCase());
      for (String label : labels) {
        value = label + (suffix != null ? "-" + suffix : "");
        features.add(name + ":" + value);
      }
    }
    
    if (enableDistSim) {
      name = prefix + ".distsim"; // keep this fine grained. this may or may not
                                  // work, so we need to toggle it
      String distsim = null;
      if (useFeature(name) && (distsim = getDistSim(token)) != null) {
        logger.finest("Distsim for " + token.word() + " is: " + distsim);
        if (distsim.length() > 0) {
          value = distsim + (suffix != null ? "-" + suffix : "");
          features.add(name + ":" + value);
        }
      }
    }
  }
  
  private void addCharacterNgramFeatures(String prefix, String suffix, String word, Collection<String> features) {
    if (word.length() >= 3) {
      String name = prefix;
      String value = "";
      if (useFineGrained) {
        name += ".cgram";
      } else {
        value = "cgram.";
      }
      for (int start = 0; start < word.length() - 3; start++) {
        String substr = word.substring(start, start + 3);
        // System.err.println("trigram for " + word + ": " + substr);
        if (useFeature(name)) {
          String v = value + substr + (suffix != null ? "-" + suffix : "");
          features.add(name + ":" + v);
        }
      }
    }
  }
  
  @SuppressWarnings("unused")
  private boolean hasPunc(String s) {
    for (int i = 0; i < s.length(); i++) {
      char c = s.charAt(i);
      if (c == '.')
        return true;
    }
    return false;
  }
  
  @SuppressWarnings("unused")
  private boolean hasNum(String s) {
    for (int i = 0; i < s.length(); i++) {
      char c = s.charAt(i);
      if (Character.isDigit(c))
        return true;
    }
    return false;
  }
  
  private boolean isTrigger(EntityMention em) {
    return isTrigger(em.getType());
  }
  private boolean isTrigger(String s) {
    return BioNLPFormatReader.isTrigger(s);
  }
  
  private List<Candidate> extractCandidates(Annotation corpus, boolean annotated) {
    logger.info("Extracting candidates...");
    List<Candidate> cands = new ArrayList<Candidate>();
    int posCount = 0, negCount = 0;

    for (CoreMap sentence : corpus.get(SentencesAnnotation.class)) {
      List<CoreLabel> tokens = sentence.get(TokensAnnotation.class);

      // keep track of entity positions
      HashSet<Integer> entityPositions = new HashSet<Integer>();
      if (sentence.get(EntityMentionsAnnotation.class) != null) {
        for (EntityMention em : sentence.get(EntityMentionsAnnotation.class)) {
          if (!isTrigger(em)) {
            int h = BioNLPFormatReader.entityHead(em, tokens);
            entityPositions.add(h);
          }
        }
      }

      // positive candidates
      Set<Integer> seenPos = new HashSet<Integer>();
      if (annotated && sentence.get(EntityMentionsAnnotation.class) != null) {
        for (EntityMention em : sentence.get(EntityMentionsAnnotation.class)) {
          if (isTrigger(em)) {
            int h = BioNLPFormatReader.triggerHead(em, tokens);
            if (!entityPositions.contains(h)) {
              cands.add(new Candidate(em.getType(), h, sentence));
              seenPos.add(h);
              posCount++;
            }
          }
        }
      }

      // negative (or unknown) candidates
      for (int i = 0; i < tokens.size(); i++) {
        if (seenPos.contains(i))
          continue;
        if (entityPositions.contains(i))
          continue;
        if (filterCandidate(tokens, i)) {
          cands.add(new Candidate("O", i, sentence));
          negCount++;
        }
      }
    }

    if (annotated)
      logger.info("Extracted " + posCount + " positive candidates and " + negCount + " negative candidates.");
    else
      logger.info("Extracted " + negCount + " candidates.");
    return cands;
  }
  
  private boolean filterCandidate(List<CoreLabel> tokens, int position) {
    CoreLabel t = tokens.get(position);

    // this is a known trigger word
    if (knownTriggers.containsKey(t.word().toLowerCase())) {
      return true;
    }

    // this is a POS tag we must consider (nouns, plural nouns, all verbs)
    if (t.tag().equals("NN") || t.tag().equals("NNS") || t.tag().startsWith("VB")) {
      return true;
    }

    return false;
  }
  
  private void buildTriggerGazetteer(Annotation corpus) {
    logger.info("Building trigger gazetteer...");
    knownTriggers = new HashMap<String, Set<String>>();
    int total = 0;
    for (CoreMap sentence : corpus.get(SentencesAnnotation.class)) {
      List<CoreLabel> tokens = sentence.get(TokensAnnotation.class);
      if (sentence.get(EntityMentionsAnnotation.class) != null) {
        for (EntityMention em : sentence.get(EntityMentionsAnnotation.class)) {
          if (!dataset.getEntityTypes().contains(em.getType())) {
            int h = BioNLPFormatReader.triggerHead(em, tokens);
            CoreLabel t = tokens.get(h);
            String k = t.word().toLowerCase();
            String[] label = em.getType().split("/");
            Set<String> labels = knownTriggers.get(k);
            if (labels == null) {
              labels = new HashSet<String>();
              knownTriggers.put(k, labels);
            }
            for (String l : label)
              labels.add(l);
            total++;
          }
        }
      }
    }
    logger.info("Built the trigger gazetteer with " + knownTriggers.size() + " unique known triggers, covering " + total
        + " trigger instances.");
  }
  
  private Annotation readCorpus(String dir) throws Exception {
    // reader
    boolean useOfflineAnnotations = true;
    BioNLPFormatReader reader = new BioNLPFormatReader(useOfflineAnnotations, -1, true, documentLevelParsing, false, dataset);
    reader.setProcessor(processor);

    // read the BioNLP format
    MachineReading.setConsoleLevel(Level.SEVERE);
    reader.setLoggerLevel(Level.SEVERE);
    Annotation corpus = reader.parse(dir);

    // logging
    logger.info("DONE reading the corpus");
    MachineReading.setConsoleLevel(logLevel);
    logger.setLevel(logLevel);

    return corpus;
  }
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    String trainDir = props.getProperty("trigger.train.dir");
    String testDir = props.getProperty("trigger.test.dir");
    assert trainDir != null : "Training directory must be specified!";
    assert testDir != null : "Testing directory must be specified!";

    TriggerClassifier tc = new TriggerClassifier(props);
    tc.logLevel = Level.FINE;

    // tc.diffSets(trainDir, testDir); System.exit(0);

    boolean saveFiles = Boolean.parseBoolean(props.getProperty("trigger.save.files", "false"));
    boolean featureSelection = props.containsKey("trigger.feature.selection")
        && props.getProperty("trigger.feature.selection").equalsIgnoreCase("true");
    if (featureSelection) {
      //
      // do feature selection between train and dev
      //
      logger.info("Running in feature selection mode.");
      // Set<String> feats = tc.featureSelection(trainDir, testDir);
      Set<String> feats = tc.featureSelectionWithCrossValidation(trainDir, props);
      logger.info("Selected features: " + StringUtils.join(feats, ","));
    } else if (saveFiles) {
      //
      // generate all trigger files: in train using xval, in test using train
      //
      logger.info("Generating trigger files in training using cross validation...");
      tc.crossValidation(trainDir, saveFiles);
      logger.info("Completed cross validation.");
      logger.info("Training on the entire training partition...");
      tc.train(trainDir);
      logger.info("Generating trigger files in the testing corpus...");
      Score[] score = tc.test(testDir, saveFiles);
      if (score != null) {
        for (int i = 0; i < score.length; i++) {
          logger.info("Score #" + i + " in testing: " + score[i]);
        }
      }
    } else {
      // regular run: train followed by evaluation on development
      tc.train(trainDir);
      Score[] score = tc.test(testDir, false);
      if (score != null) {
        for (int i = 0; i < score.length; i++) {
          logger.info("Score #" + i + " in testing: " + score[i]);
        }
      }
    }
  }
}
