package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.stanford.nlp.ie.machinereading.domains.bionlp.TriggerSequenceClassifier.TriggerInfo;
import edu.stanford.nlp.io.IOUtils;
import edu.stanford.nlp.util.CollectionValuedMap;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;

public class TriggerMajorityVote {
  private static Logger logger = Logger.getLogger(TriggerMajorityVote.class.getName());
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    String inputDirs = props.getProperty("majority.vote.input.dirs");
    String outputDir = props.getProperty("majority.vote.output.dir");
    int minCount = Integer.parseInt(props.getProperty("majority.vote.min.count"));
    
    Map<String, DocInfo> inputDirToDocInfo = new HashMap<String, DocInfo>();
    for (String inputDir : inputDirs.split(",")) {
      inputDirToDocInfo.put(inputDir, readTriggerInformation(inputDir));
    }
    
    aggregateTriggers(inputDirToDocInfo, minCount, outputDir);
  }

  private static void aggregateTriggers(Map<String, DocInfo> inputDirToDocInfo,
      int minCount, String outputDir) throws IOException {
    new File(outputDir).mkdirs();
    
    Set<Pair<String, Integer>> allKeys = new HashSet<Pair<String, Integer>>();
    int numSourceModels = 0;
    for (DocInfo docInfo : inputDirToDocInfo.values()) {
      allKeys.addAll(docInfo.allKeys());
      numSourceModels++;
    }
    
    for (Pair<String, Integer> key : allKeys) {
      logger.info("Considering triggers in " + key);
      CollectionValuedMap<String, TriggerInfo> stringsToTriggers = new CollectionValuedMap<String, TriggerInfo>();
      for (DocInfo docInfo : inputDirToDocInfo.values()) {
        List<TriggerInfo> triggers = docInfo.get(key);
        for (TriggerInfo trigger : triggers) {
          // TODO later could try multiplying these by scores for weighted majority voting
          stringsToTriggers.add(triggerInfoHashString(trigger), trigger);
        }
      }
      
      File outputFile = new File(outputDir, key.first() + ".triggers.b" + key.second());
      BufferedWriter bufferedOutput = new BufferedWriter(new FileWriter(outputFile));
      int idCounter = 1000;

      for (Collection<TriggerInfo> groupedTriggers : stringsToTriggers.values()) {
        int count = groupedTriggers.size();
        TriggerInfo trigger = groupedTriggers.iterator().next();
        
        if (count >= minCount) {          
          // include the trigger in the output
          logger.info("Selected trigger: " + trigger.format(idCounter));
          bufferedOutput.write(trigger.format(idCounter));
        }
        
        idCounter++;
      }
      
      bufferedOutput.close();
    }
  }
  
  public static String triggerInfoHashString(TriggerInfo triggerInfo) {
    List<Object> pieces = new ArrayList<Object>();
    pieces.add(triggerInfo.type);
    pieces.add(triggerInfo.startOrig);
    pieces.add(triggerInfo.endOrig);
    return pieces.toString();
  }

  public static class DocInfo {
    Map<Pair<String, Integer>, List<TriggerInfo>> idToTriggerInfo;
    
    public DocInfo() {
      idToTriggerInfo = new HashMap<Pair<String,Integer>, List<TriggerInfo>>();
    }
    
    public List<TriggerInfo> get(Pair<String, Integer> key) {
      return idToTriggerInfo.get(key);
    }

    public void put(String docId, int beamValue, List<TriggerInfo> triggers) {
      Pair<String, Integer> key = new Pair<String, Integer>(docId, beamValue);
      idToTriggerInfo.put(key, triggers);
    }

    public Collection<Pair<String, Integer>> allKeys() {
      return idToTriggerInfo.keySet();
    }
  }
  
  private static DocInfo readTriggerInformation(String inputDir) throws IOException {
    logger.info("Reading " + inputDir);
    File[] filenames = new File(inputDir).listFiles();
    
    Pattern beamRegex = Pattern.compile("([0-9]+).triggers.b([0-9]+)$");
    DocInfo docInfo = new DocInfo();
    
    for (File file : filenames) {
      String filename = file.getName();
      Matcher matcher = beamRegex.matcher(filename);
      if (!matcher.find()) {
        continue;
      }
      
      String docId = matcher.group(1);
      int beamValue = Integer.parseInt(matcher.group(2));
      docInfo.put(docId, beamValue, parseTriggerFile(file));
    }
    
    return docInfo;
  }
  
  // TODO unfortunately copied/adapted from BioNLP09Reader -- should update getPredictedTrigger to use TriggerInfo
  private static TriggerInfo parseTriggerLine(String line) {
    assert line.startsWith(BioNLPFormatReader.ENTITY_TYPE_PREFIX) : "Line \"" + line + "\" should be an entity annotation!";
    
    TriggerInfo trigInfo = new TriggerInfo();
    
    // example: T1 Protein 87 92 c-Fos  0.99 0
    String[] bigParts = line.split("\t");
    assert(bigParts.length == 3);
    trigInfo.id = bigParts[0];
    String [] parts = bigParts[1].split("\\s+", 4);
    assert(parts.length == 4);
    trigInfo.type = parts[0];
    trigInfo.startOrig = Integer.valueOf(parts[1]);
    trigInfo.endOrig = Integer.valueOf(parts[2]);
    trigInfo.text = parts[3];
    String [] probs = bigParts[2].split("\\s+");
    assert(probs.length == 2);
    trigInfo.myProb = Double.valueOf(probs[0]);
    trigInfo.firstProb = Double.valueOf(probs[1]);
    
    return trigInfo;
  }
  
  private static List<TriggerInfo> parseTriggerFile(File triggerFile) {
    ArrayList<TriggerInfo> triggers = new ArrayList<TriggerInfo>();
    for (String line : IOUtils.readLines(triggerFile)) {
      if (line.startsWith(BioNLPFormatReader.ENTITY_TYPE_PREFIX)) {
        triggers.add(parseTriggerLine(line));
      }
    }
    return triggers;
  }
}