package edu.stanford.nlp.ie.machinereading.domains.bionlp;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.stanford.nlp.ie.crf.CRFClassifier;
import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.domains.bionlp.BioNLPFormatReader.OriginalGoldAnnotation;
import edu.stanford.nlp.ie.machinereading.msteventextractor.GENIA09DataSet;
import edu.stanford.nlp.ie.machinereading.structure.EntityMention;
import edu.stanford.nlp.ie.machinereading.structure.EventMention;
import edu.stanford.nlp.ie.machinereading.structure.ExtractionObject;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.EntityMentionsAnnotation;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.EventMentionsAnnotation;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.CoreAnnotations.AnswerAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.PartOfSpeechAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.SentencesAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.StackedNamedEntityTagAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TokensAnnotation;
import edu.stanford.nlp.pipeline.Annotation;
import edu.stanford.nlp.pipeline.StanfordCoreNLP;
import edu.stanford.nlp.sequences.SeqClassifierFlags;
import edu.stanford.nlp.stats.ClassicCounter;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.util.CoreMap;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;
import edu.stanford.nlp.util.Triple;

public class TriggerSequenceClassifier {
  private static Logger logger = Logger.getLogger(TriggerSequenceClassifier.class.getName());
  
  private static final boolean JUST_GENERATE_FILES = false;
  private static final boolean DISPLAY_LENGTH_HISTOGRAM = false;
  
  /** histogram of trigger lengths */
  static Counter<Integer> lengthHisto = new ClassicCounter<Integer>();
  
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    String input = props.getProperty("trigger.train.dir");
    String output = props.getProperty("trigger.test.dir");
    boolean documentLevelParsing = Boolean.parseBoolean(props.getProperty("document.level.parsing", "false"));
    int beam = Integer.parseInt(props.getProperty("trigger.train.beam", "0"));
    boolean useClassifierStacking = Boolean.parseBoolean(props.getProperty("trigger.use.stacking", "true"));
    if (useClassifierStacking) {
      assert beam >= 0;
    }
    
    logger.info("Training using corpus from " + input);
    StanfordCoreNLP processor = new StanfordCoreNLP(props, false);
    CRFClassifier crf = train(input, processor, Level.FINE, documentLevelParsing, beam, useClassifierStacking);
    
    if(output != null){
      logger.info("Testing using corpus from: " + output);
      evaluate(output, processor, crf, Level.FINE, documentLevelParsing, beam, useClassifierStacking);
    }
  }
  
  private static void saveBio(String fn, Collection<List<CoreLabel>> sents) throws IOException {
    PrintStream os = new PrintStream(new FileOutputStream(fn));
    for(List<CoreLabel> sent: sents){
      String prev = null;
      for(CoreLabel l: sent){
        String a = l.get(AnswerAnnotation.class);
        String ne = a;
        if(! ne.equals("O")){
          if(prev != null && a.equals(prev)){
            ne = "I-" + a;
          } else {
            ne = "B-" + a;
          }
        }
        os.println(l.word() + " " + l.tag() + " " + ne);
        prev = a;
      }
      os.println();
    }
    os.close();
    System.err.println("BIO file saved to " + fn);
  }
  
  public static void evaluate(String testDir, StanfordCoreNLP processor, CRFClassifier crf, Level logLevel, boolean documentLevelParsing, int beam,
      boolean useClassifierStacking) throws Exception {
    Collection<List<CoreLabel>> testSents = readCorpus(testDir, processor, logLevel, documentLevelParsing, beam, useClassifierStacking);
    
    if(JUST_GENERATE_FILES){
      saveBio("trigger.test.seq", testSents);
      return;
    }
    
    Score scor = score(testSents, crf);
    logger.severe("TESTING SCORE: " + scor);
  }
  
  private static Score score(Collection<List<CoreLabel>> testSents, CRFClassifier crf) {
    Score score = new Score();
    
    for(List<CoreLabel> sent: testSents) {
      List<String> gold = extractAnnotations(sent);
      List<CoreLabel> preds = crf.classify(sent);
      List<String> sys = extractAnnotations(preds);
      eval(score, gold, sys);
    }
    
    return score;
  }
  
  private static void eval(Score score, List<String> gold, List<String> sys) {
    List<Triple<Integer, Integer, String>> goldEnts = extractEntities(gold);
    List<Triple<Integer, Integer, String>> sysEnts = extractEntities(sys);
    score.incTotal("all", goldEnts.size());
    score.incPredicted("all", sysEnts.size());
    for(Triple<Integer, Integer, String> se: sysEnts){
      for(Triple<Integer, Integer, String> ge: goldEnts){
        if(se.equals(ge)){
          score.incCorrect("all");
          break;
        }
      }
    }
  }
  
  private static List<Triple<Integer, Integer, String>> extractEntities(List<String> bios) {
    List<Triple<Integer, Integer, String>> ents = new ArrayList<Triple<Integer,Integer,String>>();
    for(int i = 0; i < bios.size(); ){
      if(! bios.get(i).equals("O")){
        int start = i;
        String type = bios.get(i);
        while(i < bios.size() && bios.get(i).equals(type)) i ++;
        int end = i;
        ents.add(new Triple<Integer, Integer, String>(start, end, type));
      } else {
        i ++;
      }
    }
    return ents;
  }
  
  private static List<String> extractAnnotations(List<CoreLabel> sent) {
    List<String> annos = new ArrayList<String>();
    for(CoreLabel l: sent){
      annos.add(l.get(AnswerAnnotation.class));
    }
    return annos;
  }
  
  public static CRFClassifier train(String inputDir, StanfordCoreNLP processor, Level logLevel, boolean documentLevelParsing, int beam,
      boolean useClassifierStacking) throws Exception {
    Collection<List<CoreLabel>> trainSents = readCorpus(inputDir, processor, logLevel, documentLevelParsing, beam, useClassifierStacking);
    
    if(JUST_GENERATE_FILES) {
      saveBio("trigger.train.seq", trainSents);
      return null;
    }
    
    // train the CRF
    logger.info("Started training the trigger CRF...");
    
    Properties props = StringUtils.propFileToProperties("/u/nlp/data/ner/goodClassifiers/conll.crf.chris2009.prop");
    SeqClassifierFlags flags = new SeqClassifierFlags(props);
    flags.useInPlaceSGD = true;
    flags.useQN = false;
    CRFClassifier crf = new CRFClassifier(flags);
    crf.train(trainSents);
    logger.info("Completed training.");
    crf.serializeClassifier("trigger.model.crf");
    logger.severe("CRF classifier saved to file trigger.model.crf.");
    
    Score scor = score(trainSents, crf);
    logger.severe("TRAINING SCORE: " + scor);
    
    return crf;
  }
   
  private static Collection<List<CoreLabel>> readCorpus(String inputDir, StanfordCoreNLP processor, Level logLevel, boolean documentLevelParsing, int beam,
      boolean useClassifierStacking) throws Exception {    
    // reader
    boolean useOfflineAnnotations = true;
    BioNLPFormatReader reader = new BioNLPFormatReader(useOfflineAnnotations, beam, true, documentLevelParsing, false, new GENIA09DataSet());
    reader.setProcessor(processor);
    
    // read the BioNLP format
    MachineReading.setConsoleLevel(Level.SEVERE);
    reader.setLoggerLevel(Level.SEVERE);
    Annotation corpus = reader.parse(inputDir);
    
    // logging
    logger.info("DONE reading the corpus");
    MachineReading.setConsoleLevel(logLevel);
    logger.setLevel(logLevel);
    
    // generate training data
    Collection<List<CoreLabel>> sentences = new ArrayList<List<CoreLabel>>();
    for(CoreMap sent: corpus.get(SentencesAnnotation.class)){
      List<CoreLabel> annotatedSent = prepareSent(sent, useClassifierStacking);
      if(annotatedSent != null){
        sentences.add(annotatedSent);
      }
    }
    
    if (DISPLAY_LENGTH_HISTOGRAM) {
      // display the length histogram
      Set<Integer> lengths = lengthHisto.keySet();
      Integer [] sortedLengths = new Integer[lengths.size()];
      lengths.toArray(sortedLengths);
      Arrays.sort(sortedLengths);
      System.err.println("Trigger length histogram:");
      int total = 0;
      for(Integer l: sortedLengths) total += (int) lengthHisto.getCount(l);
      for(Integer l: sortedLengths){
        int c = (int) lengthHisto.getCount(l);
        double p = 100.00 * (double) c / (double) total;
        System.err.println("\t" + l + ": " + c + " " + p + "%");
      }
    }
   
    return sentences;
  }
    
  static class TriggerInfo {
    /** entity id, if a known entity (i.e., given in the .a1 file) */
    String id;
    /** entity type */
    String type;
    /** span start in the original sentence */
    int startOrig;
    /** span end in the original sentence */
    int endOrig;
    /** position of head in the original sentence */
    int head;
    
    int length() { return endOrig - startOrig; }
    
    String text;
    double myProb;
    double firstProb;

    public String format(int idCounter) {
      String formatted = String.format("%s%s\t%s %s %s %s\t%s %s\n", BioNLPFormatReader.ENTITY_TYPE_PREFIX, idCounter, type, startOrig, endOrig, text, myProb, firstProb);
      return formatted;
    }
  };
  
  private static List<CoreLabel> prepareSent(CoreMap sentence, boolean useClassifierStacking) {
    List<CoreLabel> tokens = sentence.get(TokensAnnotation.class);
    if(sentence.get(EventMentionsAnnotation.class) != null){
      logger.finest("Have the following events in this sentence:");
      for(EventMention e: sentence.get(EventMentionsAnnotation.class)){
        logger.finest(e.toString());
        ExtractionObject p = e.getSingleParent(sentence);
        if(p != null){
          logger.finest("Parent is: " + p.toString());
        } else {
          logger.finest("Parent is null");
        }
      }
    }
    
    // copy non-gold annotations from TriggerClassifier (which the reader should store as the types inside the EntityMentions) to the tag stacking annotation.
    if (useClassifierStacking && sentence.get(EntityMentionsAnnotation.class) != null) {
      for (EntityMention mention: sentence.get(EntityMentionsAnnotation.class)) {
        if (BioNLPFormatReader.isTrigger(mention) && !BioNLPFormatReader.isGoldTrigger(mention)) {
          for (int index = mention.getExtentTokenStart(); index < mention.getExtentTokenEnd(); index++) {
            CoreLabel token = tokens.get(index);

            assert mention.getType() != null;
            // mention types will be overridden by predicted entity types
            token.set(StackedNamedEntityTagAnnotation.class, mention.getType());
          }
        }
      }
    }

    // first extract all anchor mentions in this sentence
    Map<EntityMention, TriggerInfo> trigs = new HashMap<EntityMention, TriggerInfo>();
    if(sentence.get(EntityMentionsAnnotation.class) != null){
      for(EntityMention em: sentence.get(EntityMentionsAnnotation.class)){
        if(BioNLP09Reader.isGoldTrigger(em)){
          TriggerInfo ei = new TriggerInfo();
          // make sure we get the original types, not one that could be clobbered by predictions
          String type = em.attributeMap().get(OriginalGoldAnnotation.class);
          if (type == null) {
            type = em.getType(); // fall back to original type when necessary
          }
          ei.type = type;
          ei.head = BioNLPFormatReader.triggerHead(em, tokens);
          // whole head span
          //ei.startOrig = em.getExtentTokenStart();
          //ei.endOrig = em.getExtentTokenEnd();
          // just the head token
          ei.startOrig = ei.head;
          ei.endOrig = ei.head + 1;
          ei.id = em.getObjectId();
          trigs.put(em, ei);
        }
      }
    }

    // stats
    Set<EntityMention> triggerKeys = trigs.keySet();
    for(EntityMention trigger: triggerKeys) {
      TriggerInfo info = trigs.get(trigger);
      lengthHisto.incrementCount(info.length(), 1);
      if(info.length() > 1){
        logger.fine("MULTI-WORD TRIGGER: " + trigger.getExtentString() + "\tHEAD: " + tokens.get(info.head).word());
      }
    }
    
    // create a new sentence with triggers stored in AnswerAnnotation
    List<CoreLabel> annotatedTokens = new ArrayList<CoreLabel>();
    for(CoreLabel t: tokens){
      CoreLabel al = new CoreLabel(t);
      al.set(AnswerAnnotation.class, "O");
      annotatedTokens.add(al);
    }
    for(EntityMention trigger: triggerKeys){
      TriggerInfo info = trigs.get(trigger);
      for(int i = info.startOrig; i < info.endOrig; i ++){
        logger.fine("Setting token #" + i + " to: " + info.type);
        annotatedTokens.get(i).set(AnswerAnnotation.class, info.type); 
      }
    }
    
    addAnnotationsForFeatures(sentence, annotatedTokens);
    
    // log and return
    StringBuffer buf = new StringBuffer();
    for(int i = 0; i < annotatedTokens.size(); i ++){
      if(i > 0) buf.append(" ");
      CoreLabel l = annotatedTokens.get(i);
      buf.append(l.word() + "/" + l.tag() + "/" + l.get(AnswerAnnotation.class));
    }
    logger.fine("Annotated sentence: " + buf.toString());
    return annotatedTokens;
  } 
  
  static void addAnnotationsForFeatures(CoreMap sentence, List<CoreLabel> annotatedTokens) {
    List<CoreLabel> tokens = sentence.get(TokensAnnotation.class);
    
    // add entity (proteins) info to the annotation, as POS tags
    if(sentence.get(EntityMentionsAnnotation.class) != null){
      for(EntityMention em: sentence.get(EntityMentionsAnnotation.class)){
        if(! BioNLPFormatReader.isTrigger(em)){
          int head = BioNLPFormatReader.entityHead(em, tokens);
          annotatedTokens.get(head).set(PartOfSpeechAnnotation.class, em.getType());
          logger.fine("Added entity type " + em.getType() + " at position " + head);
        }
      }
    }
  }
  
  static class TextOrderingComparator implements Comparator<Pair<EntityMention, TriggerInfo>> {
    @Override
    public int compare(
        Pair<EntityMention, TriggerInfo> o1,
        Pair<EntityMention, TriggerInfo> o2) {
      if(o1.second.startOrig < o2.second.startOrig) return -1;
      else if(o1.second.startOrig == o2.second.startOrig) return 0;
      return 1;
    }
        
  }
}