package edu.stanford.nlp.ie.machinereading.domains.nfl;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;
import java.util.logging.Logger;

import edu.stanford.nlp.ie.machinereading.Extractor;
import edu.stanford.nlp.ie.machinereading.structure.EntityMention;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations;
import edu.stanford.nlp.ie.machinereading.structure.RelationMention;
import edu.stanford.nlp.pipeline.StanfordCoreNLP;
import edu.stanford.nlp.util.CoreMap;

/**
 * 
 * Generates candidate relations based on input sentences
 * 
 * @author Andrey Gusev
 * 
 */
public abstract class GenericCandidateRelationsGenerator {

	// used to detect NFL entities
	protected final Extractor extractor;
	private final Properties props;

	public static final Logger logger = Logger
	    .getLogger(GenericCandidateRelationsGenerator.class.getName());

	public GenericCandidateRelationsGenerator(Properties props) {
		this.props = props;
		this.extractor = getExtractor(props);
	}

	/**
	 * Reads sentences from target file, generates entities and writes candidate
	 * relations based on entities to outputFile
	 * 
	 * @param sentenceFile
	 *          - the file with target sentences
	 * @param outputFile
	 *          - candiate relations will be written to this file
	 * @throws Exception
	 */
	public void run(File sentenceFile, File outputFile) throws Exception {
		List<String> sentences = readSentences(sentenceFile);

		Map<String, List<RelationMention>> candidateRelations = getCandiadateRelations(sentences);

		writeRelationsToFile(candidateRelations, outputFile);
	}

	/**
	 * reads sentence file
	 */
	private List<String> readSentences(File file) throws Exception {

		List<String> sentences = new ArrayList<String>();

		BufferedReader in = new BufferedReader(new FileReader(file));

		String line = null;
		while ((line = in.readLine()) != null) {
			line = line.trim();
			if (line.length() > 0) {
				// remove gameid
				line = processSentenceLine(line);
				sentences.add(line);
			}
		}

		return sentences;
	}

	/**
	 * creates new Extractor needed to identify entities
	 */
	protected abstract Extractor getExtractor(Properties props);

	/**
	 * Allows to process the lines that contains the sentence
	 * 
	 * @param line
	 *          - raw sentence line
	 * @return processed sentence
	 */
	protected abstract String processSentenceLine(String line);

	/**
	 * Based on sentence and extractors annotates the sentence
	 */
	protected abstract CoreMap getSentenceAnnotation(String sentence,
	    StanfordCoreNLP pipe);

	/**
	 * processed annotated sentence into map of entity mentions
	 * 
	 * @param sentenceAnnotation
	 *          - annoteted sentences
	 * @return Map of entity types to list of EntityMentions
	 */
	protected abstract Map<String, List<EntityMention>> getEntityMap(
	    CoreMap sentenceAnnotation);

	/**
	 * 
	 * @param entities
	 *          - map of entity mentions
	 * @param sentence
	 *          - original sentence
	 * @param sentenceAnnotation
	 *          - annotated sentence
	 * @param relations
	 *          - argument where candidate relations should be added
	 */
	protected abstract void generateRelations(
	    Map<String, List<EntityMention>> entities, String sentence,
	    CoreMap sentenceAnnotation, Map<String, List<RelationMention>> relations);

	/**
	 * Generates all possible relations based on entities detected for each
	 * sentence
	 */
	private Map<String, List<RelationMention>> getCandiadateRelations(
	    List<String> sentences) {

		Map<String, List<RelationMention>> retVal = new HashMap<String, List<RelationMention>>();

		StanfordCoreNLP pipe = new StanfordCoreNLP(props, false);

		for (String sentence : sentences) {

			// pre create arrays
			retVal.put(sentence, new ArrayList<RelationMention>());

			CoreMap sentenceAnnotation = getSentenceAnnotation(sentence, pipe);

			if (sentenceAnnotation
			    .get(MachineReadingAnnotations.EntityMentionsAnnotation.class) != null) {
				// get entities
				Map<String, List<EntityMention>> typeToEntities = getEntityMap(sentenceAnnotation);
				// now generate relations
				generateRelations(typeToEntities, sentence, sentenceAnnotation, retVal);
			}
		}
		return retVal;
	}

	/**
	 * Write candidate relations and sentences
	 */
	protected void writeRelationsToFile(
	    Map<String, List<RelationMention>> relations, File file)
	    throws IOException {

		FileWriter writer = new FileWriter(file);

		for (Entry<String, List<RelationMention>> sentenceWithRelations : relations
		    .entrySet()) {
			writer.write(sentenceWithRelations.getKey());
			writer.write("\n\n");
			for (RelationMention mention : sentenceWithRelations.getValue()) {
				writer.write(mention.toString());
				writer.write("\n");
			}
			writer.write("\n\n\n");

		}
		writer.close();

	}

}
