/**
 * This class generates extra relations for the sentences crawled from Yahoo! Sports
 * This works on the .rel files created just with teamScoringAll relations, and it creates teamInGame, gameWinner, and gameLoser relations
 * @author Mihai
 */
package edu.stanford.nlp.ie.machinereading.domains.nfl;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.logging.Level;

import edu.stanford.nlp.ie.machinereading.GenericDataSetReader;
import edu.stanford.nlp.ie.machinereading.MachineReading;
import edu.stanford.nlp.ie.machinereading.domains.nfl.NFLReader.NFLTokenizer;
import edu.stanford.nlp.ie.machinereading.structure.EntityMention;
import edu.stanford.nlp.ie.machinereading.structure.ExtractionObject;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations;
import edu.stanford.nlp.ie.machinereading.structure.RelationMention;
import edu.stanford.nlp.ie.machinereading.structure.Span;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations.RelationMentionsAnnotation;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.CoreAnnotations.SentencesAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TextAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.TokensAnnotation;
import edu.stanford.nlp.pipeline.Annotation;
import edu.stanford.nlp.util.CoreMap;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;

public class NFLRelationGeneratorForCrawls {
  public static void main(String[] args) throws Exception {
    Properties props = StringUtils.argsToProperties(args);
    MachineReading.setConsoleLevel(Level.FINEST);
    
    //
    // Read the crawled corpus
    //
    GenericDataSetReader reader = new StringBasedNFLReader(false, false, false);
    reader.setLoggerLevel(Level.SEVERE);
    Annotation corpus = reader.parse(props.getProperty("crawledRelPath"));
    String outputDir = props.getProperty("generatedRelPath");
    
    //
    // some stats
    //
    System.err.println("Read a corpus with " + corpus.get(SentencesAnnotation.class).size() + " sentences.");
    List<CoreMap> sentences = corpus.get(SentencesAnnotation.class);
    int validSentCount = 0;
    for(CoreMap sent: sentences){
      List<RelationMention> rels = sent.get(RelationMentionsAnnotation.class);
      if(rels != null && rels.size() == 2) validSentCount ++;
    }
    System.err.println("Found " + validSentCount + " valid sentences.");
    
    //
    // For each sentence:
    // - manually annotate NFLGame
    // - automatically generate the missing relations
    //
    BufferedReader is = new BufferedReader(new InputStreamReader(System.in));
    for(CoreMap sent: sentences){
      List<RelationMention> rels = sent.get(RelationMentionsAnnotation.class);
      Pair<EntityMention, EntityMention> winnerAndLoser = findWinnerAndLoser(rels);
      if(winnerAndLoser != null){
        String docid = sent.get(MachineReadingAnnotations.DocumentIdAnnotation.class);
        String text = sent.get(TextAnnotation.class);
        System.out.println("\nDoc id: " + docid);
        System.out.println("Sentence: " + text);
        System.out.println("Winner: " + winnerAndLoser.first());
        System.out.println("Loser: " + winnerAndLoser.second());
        System.out.print("Please input the NFLGame string: ");
        String input = is.readLine();
        input = input.trim();
        if(input.length() == 0){
          System.out.println("No NFLGame selected. Skipping sentence...");
          continue;
        } else {
          EntityMention game = makeGame(sent, input);
          if(game == null){
            System.out.println("Cannot match game text \"" + input + "\" to sentence. Skipping sentence...");
          } else {
            RelationMention inGameWinner = makeGameRelation("teamInGame", game, winnerAndLoser.first(), sent);
            rels.add(inGameWinner);
            RelationMention inGameLoser = makeGameRelation("teamInGame", game, winnerAndLoser.second(), sent);
            rels.add(inGameLoser);
            RelationMention gameWinner = makeGameRelation("gameWinner", game, winnerAndLoser.first(), sent);
            rels.add(gameWinner);
            RelationMention gameLoser = makeGameRelation("gameLoser", game, winnerAndLoser.second(), sent);
            rels.add(gameLoser);
            System.out.println("Created the following relations:");
            System.out.println(inGameWinner);
            System.out.println(inGameLoser);
            System.out.println(gameWinner);
            System.out.println(gameLoser);
            
            String fn = findFileName(outputDir, docid); 
            PrintStream os = new PrintStream(new FileOutputStream(fn));
            os.println("<relations docid=\"" + docid + "\">");
            for(RelationMention rel: rels) {
              saveAsXml(os, rel);
            }
            os.println("</relations>");
            os.close();
          }
        }
      }
    }
  }
  
  private static void saveAsXml(PrintStream os, RelationMention rel) {
    CoreMap sent = rel.getSentence();
    List<CoreLabel> tokens = sent.get(TokensAnnotation.class);
    int relStart = tokens.get(0).beginPosition();
    int relEnd = tokens.get(tokens.size() - 1).endPosition() - 1;
    os.println("  <relation end=\"" + relEnd + "\" start=\"" + relStart + "\" type=\"" + rel.getType() + "\">" + sent.get(TextAnnotation.class));
    for(ExtractionObject m: rel.getArgs()){
      int argStart = tokens.get(m.getExtentTokenStart()).beginPosition();
      int argEnd = tokens.get(m.getExtentTokenEnd()).endPosition() - 1;
      os.println("    <arg end=\"" + argEnd + "\" start=\"" + argStart + "\" type=\"" + m.getType() + "\">" + m.getExtentString() + "</arg>");
    }
    os.println("  </relation>");
  }
  
  private static String findFileName(String dir, String docid) {
    for(int count = 1; ; count ++){
      File f = new File(dir + File.separator + "nfl.game." + docid + "." + count + ".rel");
      if(! f.exists()) return f.getAbsolutePath();
    }
  }
  
  private static RelationMention makeGameRelation(String type, EntityMention game, EntityMention team, CoreMap sent) {
    List<ExtractionObject> argList = new ArrayList<ExtractionObject>();
    argList.add(game);
    argList.add(team);
    Span relExtent = new Span(Integer.MAX_VALUE, Integer.MIN_VALUE);
    for(ExtractionObject o: argList){
      Span extent = o.getExtent();
      if(extent.start() < relExtent.start()){
        relExtent.setStart(extent.start());
      }
      if(extent.end() > relExtent.end()){
        relExtent.setEnd(extent.end());
      }
    }
    RelationMention rel = new RelationMention(
        RelationMention.makeUniqueId(),
        sent,
        relExtent,
        type,
        null,
        argList);
    return rel;
  }
  
  private static EntityMention makeGame(CoreMap sent, String text) {
    NFLTokenizer tokenizer = new NFLTokenizer(text);
    List<CoreLabel> argTokens = tokenizer.tokenize();
    List<CoreLabel> tokens = sent.get(TokensAnnotation.class);
    assert(tokens != null);

    List<Span> matches = new ArrayList<Span>();
    for(int offset = 0; offset < tokens.size() - argTokens.size(); offset ++){
      if(argMatches(argTokens, tokens, offset)){
        Span match = new Span(offset, offset + argTokens.size()); 
        matches.add(match);
      }
    }
    
    if(matches.size() != 1) return null;
    
    EntityMention game = new EntityMention(EntityMention.makeUniqueId(),
        sent,
        matches.get(0),
        matches.get(0), // head span is the same as extent span in NFL
        "NFLGame",
        null,
        null);
    return game;
  }
  
  private static boolean argMatches(List<CoreLabel> argTokens, List<CoreLabel> tokens, int offset) {
    for(int i = 0; i < argTokens.size(); i ++){
      if(! argTokens.get(i).word().equals(tokens.get(i + offset).word())){
        return false;
      }
    }
    return true;
  }
  
  private static Pair<EntityMention, EntityMention> findWinnerAndLoser(List<RelationMention>  rels) {
    if(rels != null && rels.size() == 2 && 
        rels.get(0).getType().equals("teamScoringAll") && 
        rels.get(1).getType().equals("teamScoringAll")){
      int score0, score1;
      EntityMention team0, team1;
      List<ExtractionObject> args0 = rels.get(0).getArgs();
      List<ExtractionObject> args1 = rels.get(1).getArgs();
      if(args0.get(0).getType().equals("FinalScore")){
        score0 = Integer.valueOf(rels.get(0).getArg(0).getExtentString());
        team0 = (EntityMention) rels.get(0).getArg(1);
      } else {
        score0 = Integer.valueOf(rels.get(0).getArg(1).getExtentString());
        team0 = (EntityMention) rels.get(0).getArg(0);
      }
      if(args1.get(0).getType().equals("FinalScore")){
        score1 = Integer.valueOf(rels.get(1).getArg(0).getExtentString());
        team1 = (EntityMention) rels.get(1).getArg(1);
      } else {
        score1 = Integer.valueOf(rels.get(1).getArg(1).getExtentString());
        team1 = (EntityMention) rels.get(1).getArg(0);
      }
      if(score0 >= score1){
        return new Pair<EntityMention, EntityMention>(team0, team1);
      } else {
        return new Pair<EntityMention, EntityMention>(team1, team0);
      }
    }
    
    return null;
  }
}
