package edu.stanford.nlp.ie.machinereading.domains.nfl;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import edu.stanford.nlp.ie.machinereading.Extractor;
import edu.stanford.nlp.ie.machinereading.domains.nfl.NFLReader.NFLTokenizer;
import edu.stanford.nlp.ie.machinereading.domains.nfl.NFLRelation.ARG_TYPE;
import edu.stanford.nlp.ie.machinereading.structure.AnnotationUtils;
import edu.stanford.nlp.ie.machinereading.structure.EntityMention;
import edu.stanford.nlp.ie.machinereading.structure.ExtractionObject;
import edu.stanford.nlp.ie.machinereading.structure.MachineReadingAnnotations;
import edu.stanford.nlp.ie.machinereading.structure.RelationMention;
import edu.stanford.nlp.ie.machinereading.structure.Span;
import edu.stanford.nlp.ling.CoreAnnotations;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.pipeline.Annotation;
import edu.stanford.nlp.pipeline.StanfordCoreNLP;
import edu.stanford.nlp.util.CoreMap;

/**
 * 
 * @author Andrey Gusev
 * 
 *         Based on identified entities generates all possible relations for NFL
 *         domain
 * 
 */
public class NFLRelationsGenerator extends GenericCandidateRelationsGenerator {

	private static final boolean USE_AMT_FORMAT = true;
	private static final boolean USE_AMT_4_OPTION_FORMAT = false;

	/**
	 * 
	 * @param args
	 *          - file to write sentences to and optional override for default
	 *          location of gazetteer
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {

		if (args == null || args.length != 3) {
			System.out.println("Usage: NFLRelationsGenerator fileWithSentence outFile propertiesFile");
			return;
		}

		Properties props = getProperties(args[2]);
		if (props == null) {
			throw new RuntimeException("ERROR: failed to find Properties in the given arguments!");
		}

		NFLRelationsGenerator generator = new NFLRelationsGenerator(props, USE_AMT_FORMAT, USE_AMT_4_OPTION_FORMAT);
		generator.run(new File(args[0]), new File(args[1]));
	}

	private final boolean useAmtFormatForRelation;
	private final boolean useAmt4OptionFormat;

	public NFLRelationsGenerator(Properties props, boolean useAmtFormatForRelation, boolean useAmt4OptionFormat) {
		super(props);
		this.useAmtFormatForRelation = useAmtFormatForRelation;
		this.useAmt4OptionFormat = useAmt4OptionFormat;
	}

	@Override
	protected Extractor getExtractor(Properties props) {
		String gazetteerLocation = props.getProperty("entityGazetteerPath");
		return new NFLMaxRecallEntityExtractor(gazetteerLocation);
	}

	@Override
	protected String processSentenceLine(String line) {
		// handles "gameid:" string in the beginning of each sentence
		line = line.substring(line.indexOf(":") + 1);
		line = line.replace(",", "");
		return line.trim();
	}

	@Override
	protected CoreMap getSentenceAnnotation(String sentence, StanfordCoreNLP pipe) {
		Annotation dataset = new Annotation("");
		// annotate
		NFLTokenizer tokenizer = new NFLTokenizer(sentence);
		List<CoreLabel> tokens = tokenizer.tokenize();
		Annotation annotatedSentence = new Annotation(sentence);
		List<CoreLabel> labels = tokens; // AnnotationUtils.wordsToCoreLabels(tokens);
		annotatedSentence.set(CoreAnnotations.TokensAnnotation.class, labels);

		AnnotationUtils.addSentence(dataset, annotatedSentence);

		pipe.annotate(dataset);

		extractor.annotate(dataset);

		List<CoreMap> sentenceAnnotation = dataset.get(CoreAnnotations.SentencesAnnotation.class);

		return sentenceAnnotation.get(0);
	}

	@Override
	protected Map<String, List<EntityMention>> getEntityMap(CoreMap sentenceAnnotation) {
		Map<String, List<EntityMention>> typeToEntities = new HashMap<String, List<EntityMention>>();

		// pre-create arrays
		typeToEntities.put(NFLMaxRecallEntityExtractor.NFL_TEAM, new ArrayList<EntityMention>());
		typeToEntities.put(NFLMaxRecallEntityExtractor.FINAL_SCORE, new ArrayList<EntityMention>());
		typeToEntities.put(NFLMaxRecallEntityExtractor.NFL_GAME, new ArrayList<EntityMention>());
		typeToEntities.put(NFLMaxRecallEntityExtractor.NFL_PLAYOFF_GAME, new ArrayList<EntityMention>());
		typeToEntities.put(NFLMaxRecallEntityExtractor.DATE, new ArrayList<EntityMention>());

		for (EntityMention ent : sentenceAnnotation.get(MachineReadingAnnotations.EntityMentionsAnnotation.class)) {
			List<EntityMention> entities = typeToEntities.get(ent.getType());
			entities.add(ent);
		}

		return typeToEntities;
	}

	@Override
	protected void generateRelations(Map<String, List<EntityMention>> entities, String sentence,
			CoreMap sentenceAnnotation, Map<String, List<RelationMention>> relations) {

		Span span = new Span(0, sentence.length());

		// now create candiate relation mentions
		for (EntityMention team : entities.get(ARG_TYPE.NFL_TEAM)) {
			for (EntityMention score : entities.get(ARG_TYPE.FINAL_SCORE)) {

				// TeamScoringAll
				relations.get(sentence).add(
						new RelationMention(RelationMention.makeUniqueId(), sentenceAnnotation, span, NFLRelation.TeamScoringAll
								.getRelationType(), null, score, team));
			}
		}

		for (EntityMention game : entities.get(ARG_TYPE.NFL_GAME)) {
			for (EntityMention score : entities.get(ARG_TYPE.FINAL_SCORE)) {

				// TeamFinalScore
				relations.get(sentence).add(
						new RelationMention(RelationMention.makeUniqueId(), sentenceAnnotation, span, NFLRelation.TeamFinalScore
								.getRelationType(), null, game, score));
			}
		}

		for (EntityMention team : entities.get(ARG_TYPE.NFL_TEAM)) {
			for (EntityMention game : entities.get(ARG_TYPE.NFL_GAME)) {

				// TeamInGame
				relations.get(sentence).add(
						new RelationMention(RelationMention.makeUniqueId(), sentenceAnnotation, span, NFLRelation.TeamInGame
								.getRelationType(), null, game, team));

				// GameWinner
				relations.get(sentence).add(
						new RelationMention(RelationMention.makeUniqueId(), sentenceAnnotation, span, NFLRelation.GameWinner
								.getRelationType(), null, team, game));

				// GameLooser
				relations.get(sentence).add(
						new RelationMention(RelationMention.makeUniqueId(), sentenceAnnotation, span, NFLRelation.GameLoser
								.getRelationType(), null, team, game));
			}
		}

		for (EntityMention game : entities.get(ARG_TYPE.NFL_GAME)) {
			for (EntityMention date : entities.get(ARG_TYPE.DATE)) {

				// GameDate
				relations.get(sentence).add(
						new RelationMention(RelationMention.makeUniqueId(), sentenceAnnotation, span, NFLRelation.GameDate
								.getRelationType(), null, game, date));
			}
		}

	}

	private static final String HEADING_4_OPTIONS = "relation,sentence,yes_full_relation,no_full_relation,no_entity_1,"
			+ "human_not_entity_1,no_entity_2,human_not_entity_2\n";

	private static final String HEADING_2_OPTIONS = "relation,sentence,yes_full_relation,no_full_relation\n";

	/**
	 * There are 8 parameters that are required by template for this task at AMT,
	 * they should be returned in this order in csv file
	 * 
	 * @relation - relation in human readable form
	 * @sentence - target sentence which may have this relation
	 * @yes_full_relation - parameter which will be returned when relation holds
	 * @no_full_relation - parameter which will be returned when relation does not
	 *                   hold
	 * 
	 * @Note if USE_AMT_4_OPTION_FORMAT is true the following will also be used
	 * 
	 * @no_entity_1 - parameter returned when entity 1 of the relation does not
	 *              hold
	 * @human_not_entity_1 - parameter which represent human readable form of the
	 *                     above parameter
	 * @no_entity_2 - parameter returned when entity 2 of the relation does not
	 *              hold
	 * @human_not_entity_2 - parameter which represent human readable form of the
	 *                     above parameter
	 * 
	 * @Note - yes_full_relation,no_full_relation,no_entity_1,no_entity_2 should
	 *       contain all the information to parse out both relation and entities
	 *       involved in relation
	 * 
	 *       Write candidate relations and sentences
	 */
	@Override
	protected void writeRelationsToFile(Map<String, List<RelationMention>> relations, File file) throws IOException {

		// if regular format is used call super
		if (!useAmtFormatForRelation) {
			super.writeRelationsToFile(relations, file);
			return;
		}

		// create CSV file for input into AMT task with eight variables described
		// above
		FileWriter writer = new FileWriter(file);

		// first write headings - these are eight variables
		writer.write(useAmt4OptionFormat ? HEADING_4_OPTIONS : HEADING_2_OPTIONS);
		int counter = 0;
		for (Entry<String, List<RelationMention>> sentenceWithRelations : relations.entrySet()) {

			String sentence = sentenceWithRelations.getKey();
			for (RelationMention mention : sentenceWithRelations.getValue()) {

				List<ExtractionObject> args = mention.getArgs();
				Map<String, ExtractionObject> typeToEntity = new HashMap<String, ExtractionObject>();
				for (ExtractionObject arg : args) {
					typeToEntity.put(arg.getType(), arg);
				}

				// relation variable
				writer.write(getHumanRedableRelation(mention, typeToEntity).replace(",", "") + ",");
				// sentence variable
				// NOTE: this string should not be changed otherwise offsets will be
				// wrong for entities
				writer.write(sentence + ",");
				// yes_full_relation variable
				writer.write(getPositiveRelation(mention, typeToEntity) + ",");
				// no_full_relation variable
				writer.write(getNegativeOverallRelation(mention, typeToEntity));
				if (useAmt4OptionFormat) {
					// no_entity_1 variable
					writer.write("," + getNegativeEntity1Relation(mention, typeToEntity) + ",");
					// human_not_entity_1 variable
					writer.write(getHumanReadableEntityNegative(mention, typeToEntity, 1).replace(",", "") + ",");
					// no_entity_2 variable
					writer.write(getNegativeEntity2Relation(mention, typeToEntity) + ",");
					// human_not_entity_1 variable
					writer.write(getHumanReadableEntityNegative(mention, typeToEntity, 2).replace(",", ""));
				}
				writer.write("\n");
				counter++;
			}
		}
		writer.close();
	}

	private String getHumanReadableEntityNegative(RelationMention mention, Map<String, ExtractionObject> typeToEntity,
			int argPos) {

		NFLRelation relation = NFLRelation.relationToEnum.get(mention.getType());

		String argType = relation.getArgType(argPos);
		ExtractionObject arg = typeToEntity.get(argType);
		String humanReadableType = null;
		if (ARG_TYPE.NFL_TEAM.equals(argType)) {
			humanReadableType = "nfl team";
		} else if (ARG_TYPE.FINAL_SCORE.equals(argType)) {
			humanReadableType = "score";
		} else if (ARG_TYPE.NFL_GAME.equals(argType)) {
			humanReadableType = "nfl game";
		} else if (ARG_TYPE.DATE.equals(argType)) {
			humanReadableType = "date";
		}

		if (humanReadableType == null) {
			throw new RuntimeException("humanReadableType should not be null, arg:" + arg);
		}

		return " '" + arg.getValue() + "' does not represent '" + humanReadableType + "'";

	}

	private String getPositiveRelation(RelationMention mention, Map<String, ExtractionObject> typeToEntity) {
		return "1_" + getRelationForAMT(mention, typeToEntity);
	}

	private String getNegativeOverallRelation(RelationMention mention, Map<String, ExtractionObject> typeToEntity) {
		return "2_" + getRelationForAMT(mention, typeToEntity);
	}

	private String getNegativeEntity1Relation(RelationMention mention, Map<String, ExtractionObject> typeToEntity) {
		return "3_" + getRelationForAMT(mention, typeToEntity);
	}

	private String getNegativeEntity2Relation(RelationMention mention, Map<String, ExtractionObject> typeToEntity) {
		return "4_" + getRelationForAMT(mention, typeToEntity);
	}

	private String getRelationForAMT(RelationMention mention, Map<String, ExtractionObject> typeToEntity) {
		String type = mention.getType();
		StringBuilder sb = new StringBuilder();
		sb.append(type).append("_");

		NFLRelation relation = NFLRelation.relationToEnum.get(type);
		sb.append(getEntityForAMT((EntityMention) typeToEntity.get(relation.getFirstArgType()))).append("_");
		sb.append(getEntityForAMT((EntityMention) typeToEntity.get(relation.getSecondArgType())));
		return sb.toString().trim().replace(",", "");
	}

	private String getEntityForAMT(EntityMention ent) {
		return "[" + ent.getValue().replace(" ", "") + "_" + ent.getType() + "_" + ent.getHeadTokenStart() + "_"
				+ ent.getHeadTokenEnd() + "]";
	}

	private String getHumanRedableRelation(RelationMention mention, Map<String, ExtractionObject> typeToEntity) {

		StringBuilder sb = new StringBuilder();
		NFLRelation relation = NFLRelation.relationToEnum.get(mention.getType());

		switch (relation) {
		case TeamScoringAll:
			sb.append("The team '").append(typeToEntity.get(relation.getFirstArgType()).getValue()).append("'");
			sb.append(" scored '").append(typeToEntity.get(relation.getSecondArgType()).getValue()).append("' points.");
			break;
		case TeamFinalScore:
			sb.append("In the game between two teams described by '").append(
					typeToEntity.get(relation.getFirstArgType()).getValue()).append("'");
			sb.append(" one of the teams scored '").append(typeToEntity.get(relation.getSecondArgType()).getValue()).append(
					"' points.");
			break;
		case TeamInGame:
			sb.append("The team '").append(typeToEntity.get(relation.getFirstArgType()).getValue()).append("'");
			sb.append(" participated in the game between two teams described by '").append(
					typeToEntity.get(relation.getSecondArgType()).getValue()).append("'.");
			break;
		case GameWinner:
			sb.append("The team '").append(typeToEntity.get(relation.getFirstArgType()).getValue()).append("'");
			sb.append(" is the winner of the game between two teams described by '").append(
					typeToEntity.get(relation.getSecondArgType()).getValue()).append("'.");
			break;
		case GameLoser:
			sb.append("The team '").append(typeToEntity.get(relation.getFirstArgType()).getValue()).append("'");
			sb.append(" is the loser of the game between two teams described by '").append(
					typeToEntity.get(relation.getSecondArgType()).getValue()).append("'.");
			break;
		case GameDate:
			sb.append("The game between two teams described by '").append(
					typeToEntity.get(relation.getFirstArgType()).getValue()).append("'");
			sb.append(" occurred on the date: '").append(typeToEntity.get(relation.getSecondArgType()).getValue()).append(
					"'.");
			break;
		}

		return sb.toString().trim();
	}

	private static Properties getProperties(String fileName) {

		try {
			InputStream is = new BufferedInputStream(new FileInputStream(fileName));
			Properties result = new Properties();
			result.load(is);
			// trim all values
			for (Object propKey : result.keySet()) {
				String newVal = result.getProperty((String) propKey);
				result.setProperty((String) propKey, newVal.trim());
			}
			is.close();
			return result;
		} catch (IOException e) {
			logger.severe("ERROR: failed to read properties from file " + fileName);
			e.printStackTrace();
			return null;
		}
	}

}
