package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.*;
import java.util.*;

import edu.stanford.nlp.ie.machinereading.msteventextractor.DependencyInstance.SentenceInfo;

/**
 * A reader for files in BioNLP's CoNLL format.
 */
public class BioNLPReader extends DependencyReader {

  protected boolean discourseMode = false;
  private boolean useNoTypeLabel = false;

  public BioNLPReader(boolean discourseMode) {
    this.discourseMode = discourseMode;
  }

  public DependencyInstance getNext() throws IOException {

    ArrayList<String[]> lineList = new ArrayList<String[]>();
    
    List<SentenceInfo> sentences = new ArrayList<SentenceInfo>();
    String docId = null;

    String line = inputReader.readLine();
    while (line != null && !line.equals("") && !line.startsWith("*")) {
      if(line.startsWith("#")){
        String [] toks = line.split("\\s+");
        if(line.startsWith("#DOC")){
          docId = toks[1];
        } else if(line.startsWith("#SENT")){
          int j = 0;
          OrigToken [] origSentence = new OrigToken[(toks.length - 1) / 4];
          for(int i = 1; i < toks.length; i += 4){
            OrigToken t = new OrigToken();
            t.word = toks[i];
            t.pos = toks[i + 1];
            t.startCharOffset = Integer.parseInt(toks[i + 2]);
            t.endCharOffset = Integer.parseInt(toks[i + 3]);
            origSentence[j ++] = t;
          }
          SentenceInfo si = new SentenceInfo();
          si.originalSentence = origSentence;
          if(sentences.size() > 0){
            SentenceInfo last = sentences.get(sentences.size() - 1);
            assert(last.originalSentence != null && last.originalTreeAsString != null);
          }
          sentences.add(si);
        } else if(line.startsWith("#TREE: ")){
          assert(sentences.size() > 0);
          SentenceInfo last = sentences.get(sentences.size() - 1);
          last.originalTreeAsString = line.substring(7);
        }
      } else {
        lineList.add(line.split("\t"));
      }
      line = inputReader.readLine();
      // System.out.println("## "+line);
    }

    int length = lineList.size();

    if (length == 0) {
      inputReader.close();
      return null;
    }

    String[] forms = new String[length + 1];
    String[] lemmas = new String[length + 1];
    String[] cpos = new String[length + 1];
    String[] pos = new String[length + 1];
    String[][] feats = new String[length + 1][];
    String[] deprels = new String[length + 1];
    int[] heads = new int[length + 1];
    int [] startOrigs = new int[length + 1]; // start token in the original sentence
    int [] endOrigs = new int[length + 1]; // end token in the original sentence
    int [] headOrigs = new int[length + 1];
    String [] ids = new String [length + 1]; // entity ids
    int [] sents = new int[length + 1];
    double [] triggerProbs = new double[length + 1];
    double [] firstTriggerProbs = new double[length + 1];

    forms[0] = "<root>";
    lemmas[0] = "<root-LEMMA>";
    cpos[0] = "<root-CPOS>";
    pos[0] = "<root-POS>";
    deprels[0] = useNoTypeLabel ? "<no-type>" : "root";
    heads[0] = -1;
    startOrigs[0] = -1;
    endOrigs[0] = -1;
    headOrigs[0] = -1;
    ids[0] = "<root>";
    sents[0] = -1;
    triggerProbs[0] = -1;
    firstTriggerProbs[0] = 0;

    for (int i = 0; i < length; i++) {
      String[] info = lineList.get(i);
      forms[i + 1] = normalize(info[1]);
      lemmas[i + 1] = normalize(info[2]);
      cpos[i + 1] = info[3];
      pos[i + 1] = info[4];
      feats[i + 1] = info[5].split("\\|");
      deprels[i + 1] = labeled ? normalizeLabel(info[7]) : (useNoTypeLabel ?"<no-type>" : "root");
      heads[i + 1] = Integer.parseInt(info[6]);
      int startOrig = Integer.parseInt(info[10]);
      int endOrig = Integer.parseInt(info[11]);
      int headOrig = Integer.parseInt(info[12]);
      startOrigs[i + 1] = startOrig;
      endOrigs[i + 1] = endOrig;
      headOrigs[i + 1] = headOrig;
      assert startOrig != -1;
      assert endOrig != -1;
      assert headOrig != -1;
      assert startOrig <= headOrig;
      assert headOrig <= endOrig;
      ids[i + 1] = info[13];
      sents[i + 1] = Integer.parseInt(info[14]);
      
      String [] probBits = info[15].split(",");
      assert(probBits != null && probBits.length == 2);
      triggerProbs[i + 1] = Double.parseDouble(probBits[0]);
      firstTriggerProbs[i + 1] = Double.parseDouble(probBits[1]);
    }

    feats[0] = new String[feats[1].length];
    for (int i = 0; i < feats[1].length; i++)
      feats[0][i] = "<root-feat>" + i;

    // The following stuff is for discourse and can be safely
    // ignored if you are doing sentential parsing. (In theory it
    // could be useful for sentential parsing.)
    if (discourseMode) {
      String[][] extended_feats = new String[feats[0].length][length + 1];
      for (int i = 0; i < extended_feats.length; i++) {
        for (int j = 0; j < length + 1; j++)
          extended_feats[i][j] = feats[j][i];
      }

      feats = extended_feats;
    }

    ArrayList<RelationalFeature> rfeats = new ArrayList<RelationalFeature>();

    while (line != null && !line.equals("")) {
      rfeats.add(new RelationalFeature(length, line, inputReader));
      line = inputReader.readLine();
    }

    RelationalFeature[] rfeatsList = new RelationalFeature[rfeats.size()];
    rfeats.toArray(rfeatsList);

    // End of discourse stuff.

    assert(docId != null);
    assert(sentences.size() > 0);
    return new DependencyInstance(forms, lemmas, cpos, pos, feats, deprels,
        heads, rfeatsList, docId, sentences, startOrigs, endOrigs, headOrigs, ids, sents, 
        triggerProbs, firstTriggerProbs);
  }
  
  private static String normalizeLabel(String s){
    if(s.startsWith("Theme")) return "Theme";
    return s;
  }

  protected boolean fileContainsLabels(String file) throws IOException {
    BufferedReader in = new BufferedReader(new FileReader(file));

    String line;
    boolean ret = false;
    while((line = in.readLine()) != null){
      if(line.trim().startsWith("#")) continue;
      if(line.trim().length() > 0){
        ret = true;
        break;
      }
    }
    in.close();
    return ret;
  }

}
