package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.stanford.nlp.ie.machinereading.msteventextractor.DependencyInstance.SentenceInfo;
import edu.stanford.nlp.ling.IndexedWord;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.semgraph.SemanticGraph;
import edu.stanford.nlp.trees.semgraph.SemanticGraphEdge;
import edu.stanford.nlp.trees.semgraph.SemanticGraphFactory;
import edu.stanford.nlp.util.Pair;

public class CoNLLToBioNLP {
  private static final boolean REMOVE_EVENTS_WITH_ZERO_ARGUMENTS = true;
  private static final boolean REMOVE_EVENTS_WITHOUT_THEME = true;
  private static final boolean DUPLICATE_MULTIPLE_THEME_EVENTS = true;
  private static final boolean DUPLICATE_MULTIPLE_CAUSE_REGULATIONS = true;
  private static final boolean DUPLICATE_MULTIPLE_CAUSE_CATALYSES = true;
  private static final boolean SPLIT_JOINT_ANCHOR_TYPES = true;
  private static final boolean CONVERT_CAUSE_TO_THEME = true;
  private static final boolean CONVERT_ROOT_TO_THEME = true;
  private static final boolean REMOVE_ROOT = false; // enabled only when CONVERT_ROOT_TO_THEME == false 
  
  /** If true, generate anchors' offsets based on extent; otherwise based on syntactic head */
  private static final boolean OFFSETS_BASED_ON_EXTENT = false;
  
  private int CRT_EVENT_ID;
  
  private boolean NEEDS_RECURSION = false;
  private DataSet dataset;
  
  public static void main(String[] args) throws Exception {
    String inputFile = args[0];
    String outputDir = args[1];
    String datasetName = args[2];
    DataSet d;
    if (datasetName.equalsIgnoreCase("genia")) {
      d = new GENIA11DataSet();
    } else if (datasetName.equalsIgnoreCase("epi")) {
      d = new EpigeneticsDataSet();
    } else if (datasetName.equalsIgnoreCase("infect")) {
      d = new InfectiousPlusPlusDataSet();
    } else {
      throw new RuntimeException("Invalid dataset name: " + datasetName + " (must be genia, epi, or infect)");
    }

    CoNLLToBioNLP converter = new CoNLLToBioNLP(d);
    converter.convert(inputFile, outputDir);
  }
  
  public CoNLLToBioNLP(DataSet dataset) {
    this(dataset, -1);
  }
  
  public CoNLLToBioNLP(DataSet dataset, int crtId) {
    CRT_EVENT_ID = crtId;
    NEEDS_RECURSION = false;
    this.dataset = dataset;
  }
  
  public void convert(String inputFile, String outputDir) throws Exception {
    DependencyReader reader = DependencyReader.createDependencyReader("BIONLP");
    reader.startReading(inputFile);
    DependencyInstance dep;
    int count = 0;
    while ((dep = reader.getNext()) != null) {
      count++;
      try {
        String fileName = outputDir + File.separator + dep.docId + ".a2";
        CRT_EVENT_ID = findLastUsedEventId(fileName) + 1;
        String output = convert(dep);
        PrintStream os = new PrintStream(new FileOutputStream(fileName, true));
        os.print(output);
        os.close();
      } catch (RuntimeException re) {
        System.err.println("Runtime exception while processing " + dep.docId);
        re.printStackTrace();
        throw re;
      }
    }
    System.err.println("Read " + count + " sentences.");
    if (!OFFSETS_BASED_ON_EXTENT) {
      System.err.println("Converted to BioNLP using trigger heads (approximate matching).");
    }
  }
  
  private static int findLastUsedEventId(String fn) throws IOException {
    int maxEvent = 0;
    try {
      BufferedReader is = new BufferedReader(new FileReader(fn));
      String line;
      while ((line = is.readLine()) != null) {
        String[] toks = line.split("\\s+");
        if (toks[0].startsWith("E")) {
          int id = Integer.parseInt(toks[0].substring(1));
          if (id > maxEvent)
            maxEvent = id;
        }
      }
      is.close();
    } catch (IOException e) {
      // ignored
    }
    return maxEvent; 
  }
  
  static class Entity {
    String id;
    String type;
    Pair<Integer, Integer> offsets;
    Pair<Integer, Integer> headOffsets;
    int sentence;
    String text;
    String headText;
    
    public Entity() {}
    
    public Entity(Entity other, int suffix) {
      this.id = other.id + Integer.toString(suffix);
      this.type = other.type;
      this.offsets = new Pair<Integer, Integer>(other.offsets.first, other.offsets.second);
      this.headOffsets = new Pair<Integer, Integer>(other.headOffsets.first, other.headOffsets.second);
      this.sentence = other.sentence;
      this.text = other.text;
      this.headText = other.headText;
    }
  };
  
  static class Event {
    String id;
    String type;
    Entity anchor;
    List<String> argIds;
    List<String> argNames;
    DependencyInstance dep;
    
    public Event(int id, String type, Entity anchor, DependencyInstance dep) {
      this.id = "E" + Integer.toString(id);
      this.type = type;
      this.anchor = anchor;
      this.argIds = new java.util.ArrayList<String>();
      this.argNames = new ArrayList<String>();
      this.dep = dep;
    }
    
    public Event(String id, String type, Entity anchor, DependencyInstance dep) {
      this.id = id;
      this.type = type;
      this.anchor = anchor;
      this.argIds = new java.util.ArrayList<String>();
      this.argNames = new ArrayList<String>();
      this.dep = dep;
    }
    
    public void addArg(String id, String name) {
      argIds.add(id);
      argNames.add(name);
    }
    
    public void removeArg(String id) {
      List<String> newArgIds = new ArrayList<String>();
      List<String> newArgNames = new ArrayList<String>();
      for(int i = 0; i < argIds.size(); i ++){
        if(! argIds.get(i).equalsIgnoreCase(id)){
          newArgIds.add(argIds.get(i));
          newArgNames.add(argNames.get(i));
        }
      }
      argIds = newArgIds;
      argNames = newArgNames;
    }
    
    @Override
    public String toString() {
      StringBuffer os = new StringBuffer();
      os.append(id + " " + type + ":" + anchor.id);
      for(int i = 0; i < argIds.size(); i ++)
        os.append(" " + argNames.get(i) + ":" + argIds.get(i));
      return os.toString();
    }
  };
  
  private static Entity makeEntity(DependencyInstance dep, int position) {
    Entity e = new Entity();
    e.id = dep.ids[position];
    e.sentence = dep.sentenceIndex[position];
    SentenceInfo si = dep.sentences.get(e.sentence);
    e.type = dep.postags[position];
    int startInOriginal = dep.startInOriginal[position];
    int start = si.originalSentence[startInOriginal].startCharOffset;
    int endInOriginal = dep.endInOriginal[position];
    int end = si.originalSentence[endInOriginal - 1].endCharOffset;
    e.offsets = new Pair<Integer, Integer>(start, end);
    int hstart = si.originalSentence[dep.headInOriginal[position]].startCharOffset;
    int hend = si.originalSentence[dep.headInOriginal[position]].endCharOffset;
    e.headOffsets = new Pair<Integer, Integer>(hstart, hend);
    StringBuffer text = new StringBuffer();
    for (int i = startInOriginal; i < endInOriginal; i++) {
      if (i > startInOriginal) {
        text.append(" ");
      }
      text.append(si.originalSentence[i].word);
    }
    e.text = text.toString();
    e.headText = si.originalSentence[dep.headInOriginal[position]].word;
    return e;
  }
  
  static void printEventIds(String h, Collection<Event> events) {
    System.err.print("EVENTS " + h + ":");
    for (Event e : events) {
      System.err.print(" " + e.id);
    }
    System.err.println();
  }
  
  private Collection<Event> postProcessing(Collection<Event> events, 
      Set<String> zeroArgEvents, 
      String docid, 
      Map<String, Integer> idToTokenIndex,
      Map<String, Integer> idToSentenceIndex) {
    // remove events with zero args
    if (REMOVE_EVENTS_WITH_ZERO_ARGUMENTS) {
      events = removeZeroArgEvents(events, zeroArgEvents);
    }
    
    NEEDS_RECURSION = true;
    while (NEEDS_RECURSION) {
      NEEDS_RECURSION = false;
      
      // convert root arguments to Theme
      if(CONVERT_ROOT_TO_THEME){
        convertRoots(events);
      } else if(REMOVE_ROOT){
        Set<String> empties = removeRoots(events);
        events = removeZeroArgEvents(events, empties);
      }
      
      // duplicate events that must contain a single theme
      if (DUPLICATE_MULTIPLE_THEME_EVENTS) {
        events = duplicateMultipleThemeEvents(events, idToTokenIndex, idToSentenceIndex);
        // printIdToSentence("after multiple theme", idToSentenceIndex);
        // printEventIds("after multiple theme", events);
      }

      // duplicate regulation events if they have multiple causes
      if (DUPLICATE_MULTIPLE_CAUSE_REGULATIONS) {
        events = duplicateMultipleCauseRegulations(events, idToTokenIndex, idToSentenceIndex);
        // printIdToSentence("after multiple cause", idToSentenceIndex);
        // printEventIds("after multiple cause", events);
      }
      
      // duplicate catalysis events if they have multiple causes
      if (DUPLICATE_MULTIPLE_CAUSE_CATALYSES) {
        events = duplicateMultipleCauseCatalyses(events, idToTokenIndex, idToSentenceIndex);
        // printIdToSentence("after multiple cause", idToSentenceIndex);
        // printEventIds("after multiple cause", events);
      }

      // split events where the anchor has joint types
      // this also handles special cases where a regulation event is joint with something else
      // this also handles special cases where a catalysis event is joint with something else
      if (SPLIT_JOINT_ANCHOR_TYPES) {
        events = splitJointAnchorTypes(events, idToTokenIndex, idToSentenceIndex);
        // printIdToSentence("after split joint anchor", idToSentenceIndex);
        // printEventIds("after split joint anchor", events);
      }
      
      if (dataset.splitBindingEvents) {
        events = splitBindingEvents(events, idToTokenIndex, idToSentenceIndex);
        // printIdToSentence("after split bind", idToSentenceIndex);
        // printEventIds("after split bind", events);
      }
      
      // only regulation events can have causes. for all others, convert to Theme
      if (CONVERT_CAUSE_TO_THEME) {
        convertCauses(events);
      }
    }
    
    // remove regulation events wo themes
    if (REMOVE_EVENTS_WITHOUT_THEME) {
      Set<String> regWoTheme = findEventsWithoutTheme(events, docid);
      events = removeZeroArgEvents(events, regWoTheme);
    }
    
    // only a protein can be a Theme for a non-regulation event
    // so, if any event Themes found replace them with the protein at the end of the chain
    // NOT NEEDED: handled using soft constraints in the parser
    
    // sanity checks
    // make sure there are no regulation events wo a theme
    for (Event e : events) {
      if (dataset.eventsRequiringTheme.contains(e.type)) {
        boolean foundTheme = false;
        for (int i = 0; i < e.argNames.size(); i++) {
          if (e.argNames.get(i).toLowerCase().contains("theme")) {
            foundTheme = true;
            break;
          }
        }
        if (!foundTheme) {
          System.err.println("Found regulation event without a theme: " + e + " in document " + docid);
        }
      }
    }
    
    return events;
  }

  private static void extractRegulationsOrCatalyses(String [] types, Set<String> regs, Set<String> noRegs) {
    for(String t: types){
      String tl = t.toLowerCase();
      if(tl.contains("regulation") || tl.contains("catalysis")){
        regs.add(t);
      } else {
        noRegs.add(t);
      }
    }
  }
  
  private Collection<Event> splitJointAnchorTypes(Collection<Event> events, Map<String, Integer> idToTokenIndex, Map<String, Integer> idToSentenceIndex) {
    // XXX domain specific
    Collection<Event> newEvents = new ArrayList<Event>();
    HashMap<String, Set<String>> replacements = new HashMap<String, Set<String>>();
    for(Event e: events) {
      String [] types = e.type.split("/");
      if (types.length > 1) {
        Set<String> regsOrCats = new HashSet<String>();
        Set<String> noRegsOrCats = new HashSet<String>();
        extractRegulationsOrCatalyses(types, regsOrCats, noRegsOrCats);
        // contains no regulation or catalysis
        if(regsOrCats.size() == 0){
          int crt = 0;
          for(String t: noRegsOrCats){
            Entity na = new Entity();
            na.id = e.anchor.id + "0" + Integer.toString(crt ++);
            na.type = t;
            na.offsets = new Pair<Integer, Integer>(e.anchor.offsets.first, e.anchor.offsets.second);
            na.headOffsets = new Pair<Integer, Integer>(e.anchor.headOffsets.first, e.anchor.headOffsets.second);
            na.sentence = e.anchor.sentence;
            na.text = e.anchor.text;
            na.headText = e.anchor.headText;
            idToTokenIndex.put(na.id, idToTokenIndex.get(e.anchor.id));
            idToSentenceIndex.put(na.id, idToSentenceIndex.get(e.anchor.id));
            Event ne = new Event(CRT_EVENT_ID ++, t, na, e.dep);
            for(int i = 0; i < e.argIds.size(); i ++){
              ne.addArg(e.argIds.get(i), e.argNames.get(i));
            }
            newEvents.add(ne);
            
            Set<String> rids = new HashSet<String>();
            rids.add(ne.id);
            replacements.put(e.id, rids);
            for(String rid: rids){
              idToTokenIndex.put(rid, idToTokenIndex.get(e.anchor.id));
              idToSentenceIndex.put(rid, idToSentenceIndex.get(e.anchor.id));
            }
          }
        } 
        // contains 1 regulation or catalysis
        // note: regulation and catalysis behave EXACTLY the same, so we use the same code for post-processing
        else if(regsOrCats.size() == 1) {
          // TODO: handle situations with multiple no-regulation events (does not occur in training)
          assert(noRegsOrCats.size() == 1);
          int crt = 0;
          
          //
          // in these situations we:
          // (a) split events;
          // (b) move all non-cause args to the non-regulation (or non-catalysis) event 
          // (c) assign the non-regulation/non-catalysis event as a Theme to the regulation/catalysis event
          //
          
          // this is the non-regulation (or non-catalysis) event
          Entity nra = new Entity();
          // XXX: this generates an entity anchor id = old id x 100 + crt;
          // this usually works, but we may run into problem for sentences that contain 100x entities
          nra.id = e.anchor.id + "0" + Integer.toString(crt ++);
          nra.type = noRegsOrCats.iterator().next();
          nra.offsets = new Pair<Integer, Integer>(e.anchor.offsets.first, e.anchor.offsets.second);
          nra.headOffsets = new Pair<Integer, Integer>(e.anchor.headOffsets.first, e.anchor.headOffsets.second);
          nra.sentence = e.anchor.sentence;
          nra.text = e.anchor.text;
          nra.headText = e.anchor.headText;
          idToTokenIndex.put(nra.id, idToTokenIndex.get(e.anchor.id));
          idToSentenceIndex.put(nra.id, idToSentenceIndex.get(e.anchor.id));
          Event nre = new Event(CRT_EVENT_ID ++, nra.type, nra, e.dep);
          for(int i = 0; i < e.argIds.size(); i ++){
            // move everything but the Cause args
            if(! e.argNames.get(i).equalsIgnoreCase("cause")){
              nre.addArg(e.argIds.get(i), e.argNames.get(i));
            }
          }
          
          // this is the regulation (or catalysis) event
          Entity ra = new Entity();
          ra.id = e.anchor.id + "0" + Integer.toString(crt ++);
          ra.type = regsOrCats.iterator().next();
          ra.offsets = new Pair<Integer, Integer>(e.anchor.offsets.first, e.anchor.offsets.second);
          ra.headOffsets = new Pair<Integer, Integer>(e.anchor.headOffsets.first, e.anchor.headOffsets.second);
          ra.sentence = e.anchor.sentence;
          ra.text = e.anchor.text;
          ra.headText = e.anchor.headText;
          idToTokenIndex.put(ra.id, idToTokenIndex.get(e.anchor.id));
          idToSentenceIndex.put(ra.id, idToSentenceIndex.get(e.anchor.id));
          Event re = new Event(CRT_EVENT_ID ++, ra.type, ra, e.dep);
          for(int i = 0; i < e.argIds.size(); i ++){
            // move the Cause args
            if(e.argNames.get(i).equalsIgnoreCase("cause")){
              re.addArg(e.argIds.get(i), e.argNames.get(i));
            }
          }
          // add the non-regulation (or non-catalysis) event as Theme
          re.addArg(nre.id, "Theme");
          
          newEvents.add(re);
          newEvents.add(nre);
          
          Set<String> rids = new HashSet<String>();
          rids.add(nre.id);
          rids.add(re.id);
          replacements.put(e.id, rids);
          
          for(String rid: rids){
            idToTokenIndex.put(rid, idToTokenIndex.get(e.anchor.id));
            idToSentenceIndex.put(rid, idToSentenceIndex.get(e.anchor.id));
          }
        }
        // contains more than 1 regulation or catalysis
        else {
          // TODO: not sure what to do here...
          System.err.println("Found joint type with multiple regulations: " + e.type);
          newEvents.add(e);
        }
      }
      
      // just a regular event
      else {
        newEvents.add(e);
      }
    }
    
    // replace arguments with the changed event ids
    replaceArgs(newEvents, replacements);
    
    return newEvents;
  }
  
  private Collection<Event> duplicateMultipleThemeEvents(Collection<Event> events, Map<String, Integer> idToTokenIndex, Map<String, Integer> idToSentenceIndex) {
    Collection<Event> newEvents = new ArrayList<Event>();
    HashMap<String, Set<String>> replacements = new HashMap<String, Set<String>>();
    for (Event e : events) {
      boolean found = false;
      int themeCount = 0;
      for (String t : dataset.singleThemeEvents) {
        themeCount = 0;
        if (e.type.contains(t)) {
          for (int i = 0; i < e.argIds.size(); i++) {
            if (e.argNames.get(i).contains("Theme")) {
              // System.err.println("Found theme at #" + i + ": " +
              // e.argIds.get(i));
              themeCount++;
            }
          }
          if (themeCount > 1) {
            found = true;
            break;
          }
        }
      }

      // found a single-theme event that has multiple themes
      if (found) {
        Event[] es = new Event[themeCount];
        // make the new events, one for each theme
        for (int i = 0; i < es.length; i++) {
          if (i == 0) {
            es[i] = new Event(e.id, e.type, e.anchor, e.dep);
          } else {
            es[i] = new Event(CRT_EVENT_ID++, e.type, e.anchor, e.dep);
          }
          // copy all non-theme args
          for (int j = 0; j < e.argIds.size(); j++) {
            if (!e.argNames.get(j).contains("Theme")) {
              es[i].addArg(e.argIds.get(j), e.argNames.get(j));
            }
          }
        }

        // copy one theme to each argument
        int crt = 0;
        for (int i = 0; i < e.argIds.size(); i++) {
          if (e.argNames.get(i).contains("Theme")) {
            es[crt++].addArg(e.argIds.get(i), e.argNames.get(i));
          }
        }

        for (Event ne : es) {
          newEvents.add(ne);
          // System.err.println("\t" + ne);
          assert (ne.argIds.size() > 0);
        }

        Set<String> rids = new HashSet<String>();
        for (Event ne : es)
          rids.add(ne.id);
        assert (rids.size() > 0);
        replacements.put(e.id, rids);
        for (String rid : rids) {
          idToTokenIndex.put(rid, idToTokenIndex.get(e.anchor.id));
          idToSentenceIndex.put(rid, idToSentenceIndex.get(e.anchor.id));
        }
      }
      // just a regular event
      else {
        newEvents.add(e);
      }
    }
    
    // replace arguments with the changed event ids
    replaceArgs(newEvents, replacements);
    
    return newEvents;
  }
  
  private void replaceArgs(Collection<Event> events, HashMap<String, Set<String>> replacements) {
    for(Event event: events){
      List<String> newIds = new ArrayList<String>();
      List<String> newNames = new ArrayList<String>();
      
      for(int i = 0; i < event.argIds.size(); i ++){
        String id = event.argIds.get(i);
        String name = event.argNames.get(i);
        
        // not replaced
        if(! replacements.containsKey(id)){
          newIds.add(id);
          newNames.add(name);
        } 
        // found replacements
        else {
          Set<String> ids = replacements.get(id);
          assert(ids != null);
          for(String newId: ids){
            newIds.add(newId);
            newNames.add(name);
          }
          NEEDS_RECURSION = true;
        }
      }
      
      event.argIds = newIds;
      event.argNames = newNames;
    }
  }
  
  private Collection<Event> duplicateMultipleCauseRegulations(Collection<Event> events, Map<String, Integer> idToTokenIndex, Map<String, Integer> idToSentenceIndex) {
    Collection<Event> newEvents = new ArrayList<Event>();
    HashMap<String, Set<String>> replacements = new HashMap<String, Set<String>>();
    for (Event e : events) {
      boolean found = false;
      int causeCount = 0;
      if (e.type.toLowerCase().contains("regulation")) {
        causeCount = 0;
        for (int i = 0; i < e.argIds.size(); i++) {
          if (e.argNames.get(i).equalsIgnoreCase("Cause")) {
            causeCount++;
          }
        }
        if (causeCount > 1) {
          found = true;
        }
      }
      
      // found a multiple cause regulation event
      if (found) {
        Event[] es = new Event[causeCount];
        // make the new events, one for each cause
        for (int i = 0; i < es.length; i++) {
          if (i == 0) {
            es[i] = new Event(e.id, e.type, e.anchor, e.dep);
          } else {
            es[i] = new Event(CRT_EVENT_ID++, e.type, e.anchor, e.dep);
          }
          // copy all non-cause args
          for (int j = 0; j < e.argIds.size(); j++) {
            if (!e.argNames.get(j).equalsIgnoreCase("Cause")) {
              es[i].addArg(e.argIds.get(j), e.argNames.get(j));
            }
          }
        }

        // copy one cause to each argument
        int crt = 0;
        for (int i = 0; i < e.argIds.size(); i++) {
          if (e.argNames.get(i).equalsIgnoreCase("Cause")) {
            es[crt++].addArg(e.argIds.get(i), e.argNames.get(i));
          }
        }

        for (Event ne : es) {
          newEvents.add(ne);
          assert (ne.argIds.size() > 0);
        }

        Set<String> rids = new HashSet<String>();
        for (Event ne : es) {
          rids.add(ne.id);
        }
        assert (rids.size() > 0);
        replacements.put(e.id, rids);
        for (String rid : rids) {
          idToTokenIndex.put(rid, idToTokenIndex.get(e.anchor.id));
          idToSentenceIndex.put(rid, idToSentenceIndex.get(e.anchor.id));
        }
      }
      // just a regular event
      else {
        newEvents.add(e);
      }
    }

    // replace arguments with the changed event ids
    replaceArgs(newEvents, replacements);

    return newEvents;
  }
  
  private Collection<Event> duplicateMultipleCauseCatalyses(Collection<Event> events, Map<String, Integer> idToTokenIndex, Map<String, Integer> idToSentenceIndex) {
    Collection<Event> newEvents = new ArrayList<Event>();
    HashMap<String, Set<String>> replacements = new HashMap<String, Set<String>>();
    for (Event e : events) {
      boolean found = false;
      int causeCount = 0;
      if (e.type.toLowerCase().contains("catalysis")) {
        causeCount = 0;
        for (int i = 0; i < e.argIds.size(); i++) {
          if (e.argNames.get(i).equalsIgnoreCase("Cause")) {
            causeCount++;
          }
        }
        if (causeCount > 1) {
          found = true;
        }
      }
      
      // found a multiple cause catalysis event
      if (found) {
        Event[] es = new Event[causeCount];
        // make the new events, one for each cause
        for (int i = 0; i < es.length; i++) {
          if (i == 0) {
            es[i] = new Event(e.id, e.type, e.anchor, e.dep);
          } else {
            es[i] = new Event(CRT_EVENT_ID++, e.type, e.anchor, e.dep);
          }
          // copy all non-cause args
          for (int j = 0; j < e.argIds.size(); j++) {
            if (!e.argNames.get(j).equalsIgnoreCase("Cause")) {
              es[i].addArg(e.argIds.get(j), e.argNames.get(j));
            }
          }
        }

        // copy one cause to each argument
        int crt = 0;
        for (int i = 0; i < e.argIds.size(); i++) {
          if (e.argNames.get(i).equalsIgnoreCase("Cause")) {
            es[crt++].addArg(e.argIds.get(i), e.argNames.get(i));
          }
        }

        for (Event ne : es) {
          newEvents.add(ne);
          assert (ne.argIds.size() > 0);
        }

        Set<String> rids = new HashSet<String>();
        for (Event ne : es) {
          rids.add(ne.id);
        }
        assert (rids.size() > 0);
        replacements.put(e.id, rids);
        for (String rid : rids) {
          idToTokenIndex.put(rid, idToTokenIndex.get(e.anchor.id));
          idToSentenceIndex.put(rid, idToSentenceIndex.get(e.anchor.id));
        }
      }
      // just a regular event
      else {
        newEvents.add(e);
      }
    }

    // replace arguments with the changed event ids
    replaceArgs(newEvents, replacements);

    return newEvents;
  }
  
  private static int getOrigTokenIndex(String id, DependencyInstance dep, Map<String, Integer> idToTokenIndex) {
    assert (id != null);
    assert (idToTokenIndex != null);
    if (!idToTokenIndex.containsKey(id))
      throw new RuntimeException("Unknown id " + id);
    int collapsedIndex = idToTokenIndex.get(id);
    assert (collapsedIndex >= 0);
    int origIndex = dep.headInOriginal[collapsedIndex];
    return origIndex;
  }

  private static int getOrigSentenceIndex(String id, DependencyInstance dep, Map<String, Integer> idToSentenceIndex) {
    assert (id != null);
    assert (idToSentenceIndex != null);
    if (!idToSentenceIndex.containsKey(id))
      throw new RuntimeException("Unknown id " + id);
    int sentIndex = idToSentenceIndex.get(id);
    assert (sentIndex >= 0);
    return sentIndex;
  }
  
  private Collection<Event> splitBindingEvents(Collection<Event> events, Map<String, Integer> idToTokenIndex, Map<String, Integer> idToSentenceIndex) {
    Collection<Event> newEvents = new ArrayList<Event>();
    HashMap<String, Set<String>> replacements = new HashMap<String, Set<String>>();
    boolean verbose = false;
    HashMap<Integer, SemanticGraph> sentenceToUncollapsedSemanticGraph = new HashMap<Integer, SemanticGraph>();
    
    for (Event event: events) {
      // we only care about binding events
      if (!event.type.equalsIgnoreCase("Binding")) {
        newEvents.add(event);
        continue;
      }
      int numThemeArgs = 0;
      for (String argName : event.argNames) {
        if (argName.toLowerCase().startsWith("theme")) {
          numThemeArgs++;
        }
      }
      
      // ...which have multiple theme arguments
      if (numThemeArgs <= 1) {
        newEvents.add(event);
        continue;
      }
      
      // anchor position in the original sentence. we need this to compute dependency paths
      int origAnchorIndex = getOrigTokenIndex(event.anchor.id, event.dep, idToTokenIndex);
      int origAnchorSentenceIndex = getOrigSentenceIndex(event.anchor.id, event.dep, idToSentenceIndex);
      SemanticGraph graph;
      
      if (sentenceToUncollapsedSemanticGraph.containsKey(origAnchorSentenceIndex)) {
        graph = sentenceToUncollapsedSemanticGraph.get(origAnchorSentenceIndex);
      } else {
        // create the uncollapsed basic dependency tree of this sentence
        Tree tree = event.dep.sentences.get(origAnchorSentenceIndex).originalTree;
        graph = SemanticGraphFactory.makeFromTree(tree,
            false, false, false, false, true); // uncollapsed version!
        sentenceToUncollapsedSemanticGraph.put(origAnchorSentenceIndex, graph);
      }
      
      IndexedWord anchorNode = graph.getNodeByIndexSafe(origAnchorIndex + 1);
      if (anchorNode == null) {
        // the anchor is something weird like ":" which isn't in the SemanticGraph
        newEvents.add(event);
        continue;
      }

      // go through all arguments and:
      // - put all non-theme args in a single bin
      // - cluster all theme args by the first dependency from anchor to arg if they are in the same sentence
      // - cluster all theme args by sentence id if they are in different sentences
      Collection<String> nonThemeArgs = new HashSet<String>();
      HashMap<String, Collection<String>> pathToThemes = new HashMap<String, Collection<String>>();
      int nullIndex = 1;
      boolean splittable = false;
      for (int i = 0; i < event.argIds.size(); i++) {
        String argId = event.argIds.get(i);
        if (!event.argNames.get(i).equalsIgnoreCase("theme")) {
          // found a non-theme
          String n = event.argNames.get(i) + ":" + argId;
          nonThemeArgs.add(n);
        } else {
          // found a theme. add to map where key is the first dep in path
          int origArgIndex = getOrigTokenIndex(argId, event.dep, idToTokenIndex);
          int origArgSentenceIndex = getOrigSentenceIndex(argId, event.dep, idToSentenceIndex);

          // they are in different sentence: index by sentence id
          if (origArgSentenceIndex != origAnchorSentenceIndex) {
            if (addThemeWithPath(pathToThemes, "sentid" + origArgSentenceIndex, argId) > 1)
              splittable = true;
          }

          // they are in the same sentence: index by first dependency in path
          // between anchor and arg
          else {
            IndexedWord argNode;
            try {
              argNode = graph.getNodeByIndex(origArgIndex + 1);
            } catch (IllegalArgumentException iae) {
              continue;
            }
            assert (argNode != null);

            List<SemanticGraphEdge> edgePath = graph.getShortestPathEdges(anchorNode, argNode);
            if (edgePath == null || edgePath.size() == 0) {
              // somehow we didn't find a path towards this arg
              // create a new cluster for this arg
              if (addThemeWithPath(pathToThemes, "key" + nullIndex, argId) > 1)
                splittable = true;
              nullIndex++;
            } else {
              if (addThemeWithPath(pathToThemes, edgePath.get(0).getRelation().toString() + "-" + edgePath.get(0).getDependent().index(),
                  argId) > 1)
                splittable = true;
            }
          }
        }
      } // end looping thru all args
      
      // split in new events here
      List<Collection<String>> argGroups  = new ArrayList<Collection<String>>(pathToThemes.values());
      Collection<Collection<String>> splitEventArgs = new ArrayList<Collection<String>>();
      createNewSplitEvents(nonThemeArgs, argGroups, 0, splitEventArgs);

      if (splittable && verbose) {
        printSplits(System.err, event, pathToThemes, idToTokenIndex, splitEventArgs);
      }
      
      if (splitEventArgs.size() > 1) {
        Set<String> rids = new HashSet<String>();

        // split this event into multiple Binding events
        for (Collection<String> argCombo : splitEventArgs) {
          Event newEvent = new Event(CRT_EVENT_ID++, event.type, event.anchor, event.dep);
          for (String id : argCombo) {
            String argName = null;
            String[] bits = id.split(":");
            if (bits.length == 1) {
              argName = "Theme";
            } else {
              assert (bits.length == 2);
              id = bits[1];
              argName = bits[0];
            }
            assert argName != null;
            newEvent.addArg(id, argName);
          }

          rids.add(newEvent.id);
          newEvents.add(newEvent);
          if (verbose) {
            System.err.println("New split event: " + newEvent);
          }
        }

        replacements.put(event.id, rids);
        for (String rid : rids) {
          idToTokenIndex.put(rid, idToTokenIndex.get(event.anchor.id));
          idToSentenceIndex.put(rid, idToSentenceIndex.get(event.anchor.id));
        }
      } else {
        // no need to split
        newEvents.add(event);
      }
    }
    
    // replace arguments with the changed event ids
    replaceArgs(newEvents, replacements);
    
    return newEvents;
  }
  
  private static void createNewSplitEvents(Collection<String> crtEvent, 
      List<Collection<String>> argGroups,
      int argGroupIndex,
      Collection<Collection<String>> splitEventArgs) {
    if(argGroupIndex >= argGroups.size()){
      Collection<String> newEvent = new HashSet<String>(crtEvent);
      splitEventArgs.add(newEvent);
      return;
    }
    
    Collection<String> argGroup = argGroups.get(argGroupIndex);
    for(String arg: argGroup){
      crtEvent.add(arg);
      createNewSplitEvents(crtEvent, argGroups, argGroupIndex + 1, splitEventArgs);
      crtEvent.remove(arg);
    }
  }
  
  private static void printSplits(PrintStream os, 
      Event event, 
      HashMap<String, Collection<String>> pathToThemes, 
      Map<String, Integer> idToTokenIndex,
      Collection<Collection<String>> splitEventArgs) {
    os.println("=================================================================================");
    os.println("Sentence:");
    int aidx = getOrigTokenIndex(event.anchor.id, event.dep, idToTokenIndex); // anchor index
    SentenceInfo si = event.dep.sentences.get(event.anchor.sentence);
    
    for(int i = 0; i < si.originalSentence.length; i ++){
      if(i == aidx) os.print(" <<");
      else os.print(" ");
      os.print(si.originalSentence[i].word);
      if(i == aidx) os.print(">>");
    }
    os.println();
    os.println("Splitting event " + event.toString());
    os.println("Anchor:\n\t" + event.id + ": " + si.originalSentence[aidx].word);    
    os.println("Arguments:");
    for(String id: event.argIds){
      @SuppressWarnings("unused")
      int idx = getOrigTokenIndex(id, event.dep, idToTokenIndex);
      os.println("\t" + id);
    }
    os.println("Clusters:");
    for(String path: pathToThemes.keySet()){
      Collection<String> args = pathToThemes.get(path);
      os.print("\t" + path + ":");
      for(String id: args) {
        os.print(" " + id);
      }
      os.println();
    }
    os.println("New events:");
    for(Collection<String> nev: splitEventArgs) {
      os.print("\t");
      for(String na: nev){
        os.print(" " + na);
      }
      os.println();
    }
    os.println("=================================================================================");    
  }
  
  private static int addThemeWithPath(HashMap<String, Collection<String>> pathToThemes, String path, String id) {
    Collection<String> ids = pathToThemes.get(path);
    if(ids == null){
      ids = new ArrayList<String>();
      pathToThemes.put(path, ids);
    }
    ids.add(id);
    return ids.size();
  }
  
  private static Set<Set<String>> makeArgCombinations(Set<String> argNames,
      Set<Pair<String, String>> splits) {
    Set<Set<String>> combos = new HashSet<Set<String>>();
    combos.add(argNames);
    Set<Set<String>> combosUpdated = new HashSet<Set<String>>(combos);
    for (Pair<String, String> Pair : splits) {
      for (Set<String> combo : combos) {
        // if the current combo of args has two args which need to be split,
        if (combo.contains(Pair.first()) && combo.contains(Pair.second())) {
          // create two new sets
          combosUpdated.remove(combo);
          
          Set<String> combo1 = new HashSet<String>(combo);
          Set<String> combo2 = new HashSet<String>(combo);
          combo1.remove(Pair.first());
          combo2.remove(Pair.second());
          combosUpdated.add(combo1);
          combosUpdated.add(combo2);
        }
      }
      combos = new HashSet<Set<String>>(combosUpdated);
    }
    
    return combos;
  }
  
  static void printIdToSentence(String h, Map<String, Integer> idToSentenceIndex) { 
    System.err.println("ID TO SENTENCE " + h);
    for(String id: idToSentenceIndex.keySet()){
      System.err.println(id + ": " + idToSentenceIndex.get(id));
    }
  }  
  
  public String convert(String depAsString) throws Exception {
    DependencyReader reader = DependencyReader.createDependencyReader("BIONLP");
    reader.startReadingFromString(depAsString, true);
    DependencyInstance dep;
    dep = reader.getNext();
    if(dep == null) {
      throw new RuntimeException("Extracted null DependencyInstance from sentence: " + depAsString);
    }
    return convert(dep);
  }
  
  private String convert(DependencyInstance dep) throws Exception {
    HashMap<Integer, Entity> entities = new HashMap<Integer, Entity>();
    HashMap<Integer, Event> events = new HashMap<Integer, Event>();
    
    // maps an entity or event ID to a sentence token index 
    // if the ID is an event ID, maps it to the index of the anchor of the event
    HashMap<String, Integer> idToTokenIndex = new HashMap<String, Integer>();
    // maps an entity or event ID to a sentence index (for doc level parsing)
    HashMap<String, Integer> idToSentenceIndex = new HashMap<String, Integer>();
    
    // create entities and events for everything not mapped to root
    Set<String> zeroArgEvents = new java.util.HashSet<String>();
    Set<String> entityTypes = dataset.getEntityTypes();
    for (int i = 1; i < dep.length(); i++) {
      idToTokenIndex.put(dep.ids[i], i);
      idToSentenceIndex.put(dep.ids[i], dep.sentenceIndex[i]);
      if (entityTypes.contains(dep.postags[i]) && dep.heads[i] == 0)
        continue;
      Entity ent = makeEntity(dep, i);
      if (entityTypes.contains(dep.postags[i])) {
        entities.put(i, ent);
      } else {
        Event ev = new Event(CRT_EVENT_ID, dep.postags[i], ent, dep);
        idToTokenIndex.put(ev.id, i);
        idToSentenceIndex.put(ev.id, dep.sentenceIndex[i]);
        CRT_EVENT_ID++;
        events.put(i, ev);
        zeroArgEvents.add(ev.id);
      }
    }
    
    // create dependencies, i.e., add entities as event args
    for (int i = 1; i < dep.length(); i++) {
      if (dep.heads[i] == 0) {
        continue;
      }
      Event head = events.get(dep.heads[i]);
      // it may happen that the parser proposes a dependency where the head is
      // not an entity. this is not legal.
      if (head != null) {
        zeroArgEvents.remove(head.id);
        if (entityTypes.contains(dep.postags[i])) {
          Entity arg = entities.get(i);
          head.addArg(arg.id, dep.deprels[i]);
        } else {
          Event arg = events.get(i);
          assert (arg != null);
          head.addArg(arg.id, dep.deprels[i]);
        }
      }
    }
    
    Collection<Event> eventValues = events.values();
    
    // semantic post processing
    eventValues = postProcessing(eventValues, zeroArgEvents, dep.docId, idToTokenIndex, idToSentenceIndex);
    
    // save anchors and events
    StringWriter os = new StringWriter();
    Set<String> seenAnchors = new HashSet<String>();
    
    for (Event ev : eventValues) {
      Entity a = ev.anchor;
      if (!seenAnchors.contains(a.id)) {
        Integer start;
        Integer end;
        String text;
        if (OFFSETS_BASED_ON_EXTENT) {
          start = a.offsets.first();
          end = a.offsets.second();
          text = a.text;
        } else {
          start = a.headOffsets.first();
          end = a.headOffsets.second();
          text = a.headText;
        }
        
        os.append(String.format("%s\t%s %d %d\t%s\n", a.id, a.type, start, end, text));
        seenAnchors.add(a.id);
      }
    }
    for (Event ev : eventValues) {
      Set<String> seenTypes = new HashSet<String>();
      os.append(String.format("%s\t%s:%s", ev.id, ev.anchor.type, ev.anchor.id));
      for (int i = 0; i < ev.argIds.size(); i++) {
        String type = dataset.simplifyRelationName(ev.argNames.get(i));
        if (seenTypes.contains(type)) {
          type = getNextAvailableRelationName(type, seenTypes);
        }
        os.append(String.format(" %s:%s", type, ev.argIds.get(i)));
        seenTypes.add(type);
      }
      os.append("\n");
    }
    return os.toString();
  }

  private String getNextAvailableRelationName(String type, Set<String> seenTypes) {
    String newType;
    int index = 1;
    do {
      index++;
      newType = type + index;
    } while (seenTypes.contains(newType));
    
    return newType;
  }

  private Set<String> findEventsWithoutTheme(Collection<Event> events, String docid) {
    Set<String> ids = new HashSet<String>();
    boolean found = true;
    while (found) {
      found = false;
      for (Event e : events) {
        if (dataset.eventsRequiringTheme.contains(e.type)) {
          boolean themeFound = false;
          for (int i = 0; i < e.argIds.size(); i++) {
            String id = e.argIds.get(i);
            if (ids.contains(id))
              continue;
            String an = e.argNames.get(i);
            if (an.toLowerCase().contains("theme")) {
              themeFound = true;
              break;
            }
          }
          if (!themeFound && !ids.contains(e.id)) {
            ids.add(e.id);
            found = true;
          }
        }
      }
    }
    return ids;
  }

  private void convertCauses(Collection<Event> events) {
    for (Event e : events) {
      if (!dataset.eventsWithOnlyThemes.contains(e.type)) {
        continue;
      }
      List<String> newNames = new ArrayList<String>();
      for (String n : e.argNames) {
        if (n.equals("Cause")) {
          newNames.add("Theme");
          NEEDS_RECURSION = true;
        } else {
          newNames.add(n);
        }
      }
      assert (newNames.size() == e.argNames.size());
      e.argNames = newNames;
    }
  }
  
  /**
   * Converts "root" arguments to "Theme" (root is not an accepted label)
   * @param events
   */
  private void convertRoots(Collection<Event> events) {
    for (Event e : events) {
      List<String> newNames = new ArrayList<String>();
      for (String n : e.argNames) {
        if (n.equalsIgnoreCase("root")) {
          newNames.add("Theme");
          NEEDS_RECURSION = true;
        } else {
          newNames.add(n);
        }
      }
      assert (newNames.size() == e.argNames.size());
      e.argNames = newNames;
    }
  }
  
  /**
   * Removes all "root" arguments
   * @param events
   * @return Ids of the events left with zero arguments
   */
  private Set<String> removeRoots(Collection<Event> events) {
    Set<String> empties = new HashSet<String>();
    for (Event e : events) {
      List<String> newNames = new ArrayList<String>();
      List<String> newIds = new ArrayList<String>();
      for (int i = 0; i < e.argIds.size(); i ++) {
        String n = e.argNames.get(i);
        String id = e.argIds.get(i);
        if (n.equalsIgnoreCase("root")) {
          NEEDS_RECURSION = true;
        } else {
          newNames.add(n);
          newIds.add(id);
        }
      }
      e.argNames = newNames;
      e.argIds = newIds;
      if(e.argIds.size() == 0){
        empties.add(e.id);
      }
    }
    return empties;
  }
  
  private Collection<Event> removeZeroArgEvents(Collection<Event> events, Set<String> idsToRemove) {
    Collection<Event> newEvents = new ArrayList<Event>();
    Set<String> newRemovals = new HashSet<String>();
    for (Event e : events) {
      if (idsToRemove.contains(e.id) && !dataset.zeroArgEvents.contains(e.type)) {
        // System.err.println("REMOVING EVENT " + e.type);
        // will be removed
      } else {
        newEvents.add(e);
        // remove any args in the idsToRemove set
        for (String id : idsToRemove) {
          e.removeArg(id);
        }
        if (e.argIds.size() == 0 && !dataset.zeroArgEvents.contains(e.type)) {
          newRemovals.add(e.id);
        }
      }
    }
    if (newRemovals.size() == 0)
      return newEvents;
    return removeZeroArgEvents(newEvents, newRemovals);
  }
  
  public void test_makeArgCombinations() {
    Set<Pair<String,String>> splits = new HashSet<Pair<String,String>>();
    splits.add(new Pair<String,String>("b", "a"));
    splits.add(new Pair<String,String>("b", "c"));
    
    Set<Set<String>> combosx = makeArgCombinations(new HashSet<String>(Arrays.asList("a", "b", "c")), splits);
    for (Set<String> comobz : combosx) {
      System.out.println(comobz);
    }    
  }
}
