package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;

import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.stats.Counters;
import edu.stanford.nlp.util.Pair;

public class DependencyParser {

  public ParserOptions options;

  private DependencyPipe pipe;
  private DependencyDecoder decoder;
  private Parameters params;

  public DependencyParser(DependencyPipe pipe, ParserOptions options) {
    this.pipe = pipe;
    this.options = options;

    // Set up arrays
    params = new Parameters(pipe.dataAlphabet.size());
    decoder = options.secondOrder ? new DependencyDecoder2O(pipe)
        : new DependencyDecoder(pipe);
  }

  public void train(int[] instanceLengths, File trainForest)
      throws IOException {

    int i = 0;
    for (i = 0; i < options.numIters; i++) {

      System.out.print(" Iteration " + i);
      System.out.print("[");

      long start = System.currentTimeMillis();

      trainingIter(instanceLengths, trainForest, i + 1);

      long end = System.currentTimeMillis();
      System.out.println("|Time:" + (end - start) + "]");
    }

    params.averageParams(i * instanceLengths.length);
  }

  private void trainingIter(int[] instanceLengths, File trainForest, int iter) throws IOException {

    ObjectInputStream in = new ObjectInputStream(new FileInputStream(trainForest));

    int numInstances = instanceLengths.length;

    for (int i = 0; i < numInstances; i++) {
      if ((i + 1) % 500 == 0) {
        System.out.print((i + 1) + ",");
      }

      int length = instanceLengths[i];

      // Get production crap.
      FeatureVector[][][] fvs = new FeatureVector[length][length][2];
      double[][][] probs = new double[length][length][2];
      FeatureVector[][][][] nt_fvs = new FeatureVector[length][pipe.types.length][2][2];
      double[][][][] nt_probs = new double[length][pipe.types.length][2][2];
      FeatureVector[][][] fvs_trips = new FeatureVector[length][length][length];
      double[][][] probs_trips = new double[length][length][length];
      FeatureVector[][][] fvs_sibs = new FeatureVector[length][length][2];
      double[][][] probs_sibs = new double[length][length][2];

      DependencyInstance inst;

      if (options.secondOrder) {
        inst = ((DependencyPipe2O) pipe).readInstance(in, length, fvs, probs,
            fvs_trips, probs_trips, fvs_sibs, probs_sibs, nt_fvs, nt_probs,
            params);
      } else {
        inst = pipe.readInstance(in, length, fvs, probs, nt_fvs, nt_probs,
            params);
      }

      double upd = (double) (options.numIters * numInstances
          - (numInstances * (iter - 1) + (i + 1)) + 1);
      int K = options.trainK;
      Object[][] d = null;
      if (options.decodeType.equals("proj")) {
        if (options.secondOrder) {
          d = ((DependencyDecoder2O) decoder).decodeProjective(inst, fvs,
              probs, fvs_trips, probs_trips, fvs_sibs, probs_sibs, nt_fvs,
              nt_probs, K);
        } else {
          d = decoder.decodeProjective(inst, fvs, probs, nt_fvs, nt_probs, K);
        }
      }
      if (options.decodeType.equals("non-proj")) {
        if (options.secondOrder) {
          d = ((DependencyDecoder2O) decoder).decodeNonProjective(inst, fvs,
              probs, fvs_trips, probs_trips, fvs_sibs, probs_sibs, nt_fvs,
              nt_probs, K, params, false);
        } else {
          d = decoder.decodeNonProjective(inst, fvs, probs, nt_fvs, nt_probs, K, params);
        }
      }
      params.updateParamsMIRA(inst, d, upd);
    }

    System.out.print(numInstances);

    in.close();
  }

  // /////////////////////////////////////////////////////
  // Saving and loading models
  // /////////////////////////////////////////////////////
  public void saveModel(String file) throws IOException {
    ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(file));
    out.writeObject(params.parameters);
    out.writeObject(pipe.dataAlphabet);
    out.writeObject(pipe.typeAlphabet);
    out.close();
  }

  public void loadModel(String file) throws Exception {
    ObjectInputStream in = new ObjectInputStream(new FileInputStream(file));
    params.parameters = (double[]) in.readObject();
    pipe.dataAlphabet = (Alphabet) in.readObject();
    pipe.typeAlphabet = (Alphabet) in.readObject();
    in.close();
    pipe.closeAlphabets();
  }

  // ////////////////////////////////////////////////////
  // Get Best Parses ///////////////////////////////////
  // ////////////////////////////////////////////////////
  public void outputParses() throws IOException {
    String tFile = options.testfile;
    String file = options.outfile;

    long start = System.currentTimeMillis();

    pipe.initInputFile(tFile);
    pipe.initOutputFile(file);
    pipe.startReaderForStacking(options.stackingtestfiles);

    System.out.print("Processing Sentence: ");
    Pair<DependencyInstance,DependencyInstance[]> instances = pipe.nextInstance();
    DependencyInstance instance;
    DependencyInstance[] stackingInstances;
    int cnt = 0;
    while (instances != null) {
      instance = instances.first();
      stackingInstances = instances.second();
      
      cnt++;
      System.out.print(cnt + " ");
      DependencyInstance dependencyInstance = parseInstance(instance, stackingInstances);
      pipe.outputInstance(dependencyInstance, options);

      instances = pipe.nextInstance();
    }
    pipe.close();

    long end = System.currentTimeMillis();
    System.out.println("Took: " + (end - start));
  }
  
  public DependencyInstance parseInstance(DependencyInstance instance, DependencyInstance[] stackingInstances) {
    String[] forms = instance.forms;

    int length = forms.length;

    FeatureVector[][][] fvs = new FeatureVector[forms.length][forms.length][2];
    double[][][] probs = new double[forms.length][forms.length][2];
    FeatureVector[][][][] nt_fvs = new FeatureVector[forms.length][pipe.types.length][2][2];
    double[][][][] nt_probs = new double[forms.length][pipe.types.length][2][2];
    FeatureVector[][][] fvs_trips = new FeatureVector[length][length][length];
    double[][][] probs_trips = new double[length][length][length];
    FeatureVector[][][] fvs_sibs = new FeatureVector[length][length][2];
    double[][][] probs_sibs = new double[length][length][2];
    if (options.secondOrder)
      ((DependencyPipe2O) pipe).fillFeatureVectors(instance, fvs, probs,
          fvs_trips, probs_trips, fvs_sibs, probs_sibs, nt_fvs, nt_probs,
          params, stackingInstances);
    else
      pipe.fillFeatureVectors(instance, fvs, probs, nt_fvs, nt_probs, params, stackingInstances);

    int K = options.testK;
    Object[][] d = null;
    if (options.decodeType.equals("proj")) {
      if (options.secondOrder)
        d = ((DependencyDecoder2O) decoder).decodeProjective(instance, fvs,
            probs, fvs_trips, probs_trips, fvs_sibs, probs_sibs, nt_fvs,
            nt_probs, K);
      else
        d = decoder.decodeProjective(instance, fvs, probs, nt_fvs, nt_probs,
            K);
    }
    if (options.decodeType.equals("non-proj")) {
      if (options.secondOrder)
        d = ((DependencyDecoder2O) decoder).decodeNonProjective(instance,
            fvs, probs, fvs_trips, probs_trips, fvs_sibs, probs_sibs, nt_fvs,
            nt_probs, K, params, true);
      else
        d = decoder.decodeNonProjective(instance, fvs, probs, nt_fvs,
            nt_probs, K, params);
    }

    String[] pos = instance.cpostags;
    String[] formsNoRoot = new String[forms.length - 1];
    int numTokens = formsNoRoot.length;
    String[] posNoRoot = new String[numTokens];
    
    for (int j = 0; j < numTokens; j++) {
      formsNoRoot[j] = forms[j + 1];
      posNoRoot[j] = pos[j + 1];
    }
    
    // shift all indices down by one to remove the root
    int [] startInOriginalNoRoot = new int[instance.startInOriginal.length - 1];
    for(int i = 1; i < instance.startInOriginal.length; i ++) startInOriginalNoRoot[i - 1] = instance.startInOriginal[i];
    int [] endInOriginalNoRoot = new int[instance.endInOriginal.length - 1];
    for(int i = 1; i < instance.endInOriginal.length; i ++) endInOriginalNoRoot[i - 1] = instance.endInOriginal[i];
    int [] headInOriginalNoRoot = new int[instance.headInOriginal.length - 1];
    for(int i = 1; i < instance.headInOriginal.length; i ++) headInOriginalNoRoot[i - 1] = instance.headInOriginal[i];
    String [] idsNoRoot = new String[instance.ids.length - 1];
    for(int i = 1; i < instance.ids.length; i ++) idsNoRoot[i - 1] = instance.ids[i];
    int [] sentsNoRoot = new int[instance.ids.length - 1];
    for(int i = 1; i < instance.ids.length; i ++) sentsNoRoot[i - 1] = instance.sentenceIndex[i];
    double [] probsNoRoot = new double[instance.triggerProbs.length - 1];
    for(int i = 1; i < instance.triggerProbs.length; i ++) probsNoRoot[i - 1] = instance.triggerProbs[i];
    double [] firstProbsNoRoot = new double[instance.firstTriggerProbs.length - 1];
    for(int i = 1; i < instance.firstTriggerProbs.length; i ++) firstProbsNoRoot[i - 1] = instance.firstTriggerProbs[i];
    
    ParseInfo topParseInfo = getKBestParseInfo(0, d, numTokens);
    
    DependencyInstance dependencyInstance = new DependencyInstance(formsNoRoot,
        posNoRoot, topParseInfo.labels, topParseInfo.heads, instance.docId, instance.sentences,
        startInOriginalNoRoot, endInOriginalNoRoot, headInOriginalNoRoot,
        idsNoRoot, sentsNoRoot, probsNoRoot, firstProbsNoRoot);
    
    if (K > 1) {
      // shrink K as needed
      K = 0;
      for (int i = 0; i < d.length; i++) {
        if (d[i] != null && d[i][0] != null) {
          K++;
        }
      }
      
      int [][] nbestHeads = new int[K][];
      String[][] nbestLabels = new String[K][];
      double[] nbestScores = new double[K];
      int[] nbestNonProjectiveRearrangements = new int[K];

      // load the top K parses
      int currentOutputIndex = 0;
      for (int kbestIndex = 0; kbestIndex < d.length; kbestIndex++) {
        if (d[kbestIndex] == null || d[kbestIndex][0] == null) {
          continue;
        }
        ParseInfo parseInfo = getKBestParseInfo(kbestIndex, d, numTokens);
        nbestHeads[currentOutputIndex] = parseInfo.heads;
        nbestLabels[currentOutputIndex] = parseInfo.labels;
        nbestScores[currentOutputIndex] = parseInfo.score;
        nbestNonProjectiveRearrangements[currentOutputIndex] = parseInfo.nonProjectiveRearrangements;
        
        currentOutputIndex++;
      }
      
      // install them into our dependency instance
      dependencyInstance.headsKBest = nbestHeads;
      dependencyInstance.deprelsKBest = nbestLabels;
      dependencyInstance.parseScoresKBest = nbestScores;
      dependencyInstance.nonProjectiveRearrangementsKBest = nbestNonProjectiveRearrangements;
    }
    
    return dependencyInstance;
  }

  public static class ParseInfo {
    public ParseInfo(int[] heads, String[] labels, double score,
        int nonProjectiveRearrangements) {
      this.heads = heads;
      this.labels = labels;
      this.score = score;
      this.nonProjectiveRearrangements = nonProjectiveRearrangements;
    }
    
    int[] heads;
    String[] labels;
    double score;
    int nonProjectiveRearrangements;
  }
  
  private ParseInfo getKBestParseInfo(int k, Object[][] kbestParses, int length) {
    String[] res = ((String) kbestParses[k][1]).split(" ");
    
    String[] labels = new String[length];
    int[] heads = new int[length];
    
    for (int j = 0; j < length; j++) {
      String[] trip = res[j].split("[\\|:]");
      labels[j] = pipe.types[Integer.parseInt(trip[2])];
      heads[j] = Integer.parseInt(trip[0]);
    }

    double score;
    int rearrangements;
    if (kbestParses[k][3] == null) {
      rearrangements = 0;
    } else {
      rearrangements = (Integer)kbestParses[k][3];
    }
    if (kbestParses[k][2] == null) {
      score = -1;
    } else {
      score = (Double)kbestParses[k][2];
    }
    return new ParseInfo(heads, labels, score, rearrangements);
  }

  public static void printTypes(PrintStream os, Alphabet types) {
    Object [] keys = types.toArray();
    for(Object key: keys) {
      os.println("\tType:" + key + ", value:" + types.lookupAndAddIndex(key));
    }
  }
  
  static int countFeaturesSeenAtLeast(Counter<String> feats, int minCount) {
    int count = 0;
    for(String feat: feats.keySet()){
      double c = feats.getCount(feat);
      if(c >= minCount) count ++;
    }
    return count;
  }
  
  public static void trainFromOptions(ParserOptions options) throws Exception {
    DependencyPipe pipe = options.secondOrder ? new DependencyPipe2O(options)
      : new DependencyPipe(options);

    if (options.featureCountPruning) {
      pipe.featureCounts = Counters.loadIntCounter(options.featureCountsFilename, String.class);
      System.out.println("Read feature counts for " + pipe.featureCounts.size() + " features.");
      System.out.println("Found " + countFeaturesSeenAtLeast(pipe.featureCounts, options.minFeatureCount) 
          + " features seen at least " + options.minFeatureCount + " times.");
    }
    
    System.out.println("Generating train forest in " + options.trainforest.getAbsolutePath());
    int[] instanceLengths = pipe.createInstances(options.trainfile,
        options.trainforest, options.stackingtrainfiles);

    pipe.closeAlphabets();

    DependencyParser dp = new DependencyParser(pipe, options);

    int numFeats = pipe.dataAlphabet.size();
    int numTypes = pipe.typeAlphabet.size();
    System.out.print("Num Feats: " + numFeats);
    System.out.println(".\tNum Edge Labels: " + numTypes);
    printTypes(System.out, pipe.typeAlphabet);

    System.out.println("Training using forest in " + options.trainforest);
    dp.train(instanceLengths, options.trainforest);

    System.out.print("Saving model...");
    dp.saveModel(options.modelName);
    System.out.println("done.");
  }
  
  public static void testFromOptions(ParserOptions options) throws Exception {
    DependencyPipe pipe = options.secondOrder ? new DependencyPipe2O(options)
      : new DependencyPipe(options);
    
    if (options.featureCountPruning) {
      pipe.featureCounts = Counters.loadIntCounter(options.featureCountsFilename, String.class);
      System.out.println("Read feature counts for " + pipe.featureCounts.size() + " features.");
    }

    DependencyParser dp = new DependencyParser(pipe, options);

    System.out.print("\tLoading model...");
    dp.loadModel(options.modelName);
    System.out.println("done.");

    pipe.closeAlphabets();
    
    dp.outputParses();
  }

}
