package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Map.Entry;

import edu.stanford.nlp.ie.machinereading.RelationFeatureFactory;
import edu.stanford.nlp.ie.machinereading.msteventextractor.ParserOptions.conjoiningMethod;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.ling.IndexedWord;
import edu.stanford.nlp.ling.Label;
import edu.stanford.nlp.process.Morphology;
import edu.stanford.nlp.stats.Counter;
import edu.stanford.nlp.stats.Counters;
import edu.stanford.nlp.stats.IntCounter;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.Trees;
import edu.stanford.nlp.trees.semgraph.SemanticGraph;
import edu.stanford.nlp.trees.semgraph.SemanticGraphEdge;
import edu.stanford.nlp.util.CollectionUtils;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;

public class DependencyPipe {

  public Alphabet dataAlphabet;
  public Alphabet typeAlphabet;

  private DependencyReader depReader;
  private DependencyReader[] stackingReaders;
  private DependencyWriter depWriter;

  public String[] types;
  public int[] typesInt;

  public boolean labeled = false;
  private boolean isCONLL = true;

  private ParserOptions options;
  
  // only used if we're collecting or applying feature counts
  protected Counter<String> featureCounts = null;
  // only used if we're collecting per-edge feature information
  protected BufferedWriter edgeFeatureStatisticsWriter = null;
  
  protected static DataSet dataset;
  
  public DependencyPipe(ParserOptions options) throws IOException {
    this.options = options;
    dataset = options.dataset;

    if (!options.format.equals("CONLL"))
      isCONLL = false;

    dataAlphabet = new Alphabet();
    typeAlphabet = new Alphabet();
    
    if (options.collectFeatureCounts || options.featureCountPruning) {
      featureCounts = new IntCounter<String>();
    }
    
    if (options.edgeFeatureStatisticsFilename != null) {
      /*
       * Initialize edge feature statistics file
       */
      File outputFile = new File(options.edgeFeatureStatisticsFilename);
      try {
        edgeFeatureStatisticsWriter = new BufferedWriter(new FileWriter(outputFile));
      } catch (IOException e) {
        System.out.println("Couldn't create edge feature statistics file:");
        e.printStackTrace();
      }
    }

    depReader = DependencyReader.createDependencyReader(options);
  }

  public void initInputFile(String file) throws IOException {
    labeled = depReader.startReading(file);
  }

  public void initOutputFile(String file) throws IOException {
    depWriter = DependencyWriter
        .createDependencyWriter(options.format, labeled);
    depWriter.startWriting(file);
  }

  public void outputInstance(DependencyInstance instance, ParserOptions options) throws IOException {
    depWriter.write(instance, options);
  }

  public void close() throws IOException {
    if (null != depWriter) {
      depWriter.finishWriting();
    }
  }

  public String getType(int typeIndex) {
    return types[typeIndex];
  }
  
  protected final Pair<DependencyInstance, DependencyInstance[]> nextInstance() throws IOException {
    DependencyInstance instance = depReader.getNext();
    if (instance == null || instance.forms == null)
      return null;

    DependencyInstance[] stackingInstances = getNextStackingInstances();
    instance.setFeatureVector(createFeatureVector(instance, stackingInstances));

    String[] labs = instance.deprels;
    int[] heads = instance.heads;

    StringBuffer spans = new StringBuffer(heads.length * 5);
    for (int i = 1; i < heads.length; i++) {
      spans.append(heads[i]).append("|").append(i).append(":").append(
          typeAlphabet.lookupAndAddIndex(labs[i])).append(" ");
    }
    instance.actParseTree = spans.substring(0, spans.length() - 1);

    return new Pair<DependencyInstance, DependencyInstance[]>(instance, stackingInstances);
  }

  public int[] createInstances(String filename, File forestFilename, String[] stackingFilenames)
      throws IOException {

    createAlphabet(filename, stackingFilenames);

    System.out.println("Num Features: " + dataAlphabet.size());
    
    if (options.collectFeatureCounts) {
      Counters.saveCounter(featureCounts, options.featureCountsFilename);
      System.out.format("Saved collected feature counts for %s features\n.", featureCounts.size());
    }

    labeled = depReader.startReading(filename);
    startReaderForStacking(stackingFilenames);

    ArrayList<Integer> lengths = new ArrayList<Integer>();
    ObjectOutputStream forestOutputStream = new ObjectOutputStream(new FileOutputStream(forestFilename));

    DependencyInstance instance = depReader.getNext();
    DependencyInstance[] stackingInstances = getNextStackingInstances();
    
    int count = 0;

    System.out.println("Creating Feature Vector Instances: ");
    while (instance != null) {
      
      if (count % 250 == 0)
        System.out.println();
      if (count % 10 == 0)
        System.out.print(count + " ");

      instance.setFeatureVector(createFeatureVector(instance, stackingInstances));

      String[] labs = instance.deprels;
      int[] heads = instance.heads;

      StringBuffer spans = new StringBuffer(heads.length * 5);
      for (int i = 1; i < heads.length; i++) {
        spans.append(heads[i]).append("|").append(i).append(":").append(
            typeAlphabet.lookupAndAddIndex(labs[i])).append(" ");
      }
      instance.actParseTree = spans.substring(0, spans.length() - 1);

      lengths.add(instance.length());

      writeInstance(instance, forestOutputStream, stackingInstances);
      instance = null; // presumably this is here to help with GC?
      for (int index = 0; index < stackingFilenames.length; index++) {
        stackingInstances[index] = null;
      }
      
      instance = depReader.getNext();
      stackingInstances = getNextStackingInstances();

      count++;
    }

    System.out.println();
    closeAlphabets();
    forestOutputStream.close();

    int[] ret = new int[lengths.size()];
    for (int i = 0; i < lengths.size(); i++)
      ret[i] = lengths.get(i);
    return ret;
  }

  public void startReaderForStacking(String[] stackingFilenames) throws IOException {
    stackingReaders = new DependencyReader[stackingFilenames.length];
    for (int index = 0; index < stackingFilenames.length; index++) {
      String stackingFilename = stackingFilenames[index];
      DependencyReader stackingReader = DependencyReader.createDependencyReader(options);
      stackingReader.startReading(stackingFilename);
      stackingReaders[index] = stackingReader;
    }
  }

  public DependencyInstance[] getNextStackingInstances() throws IOException {
    DependencyInstance[] stackingInstances = new DependencyInstance[stackingReaders.length];
    for (int index = 0; index < stackingReaders.length; index++) {
      stackingInstances[index] = stackingReaders[index].getNext();
    }
    return stackingInstances;
  }

  private final void createAlphabet(String file, String[] stackingFilenames) throws IOException {
    System.out.print("Creating Alphabet ... ");

    labeled = depReader.startReading(file);
    startReaderForStacking(stackingFilenames);

    DependencyInstance instance = depReader.getNext();
    DependencyInstance[] stackingInstances = getNextStackingInstances();

    int count = 0;
    while (instance != null) {
      // System.err.println("createAlphabet: looking at instance: " + instance);
      String[] labs = instance.deprels;
      for (int i = 0; i < labs.length; i++)
        typeAlphabet.lookupAndAddIndex(labs[i]);

      createFeatureVector(instance, stackingInstances);

      instance = depReader.getNext();
      stackingInstances = getNextStackingInstances();
      
      if (count % 250 == 0)
        System.out.println();
      if (count % 10 == 0)
        System.out.print(count + " ");
      count++;
    }

    closeAlphabets();

    System.out.println("Done.");
  }

  public void closeAlphabets() {
    dataAlphabet.stopGrowth();
    typeAlphabet.stopGrowth();

    types = new String[typeAlphabet.size()];
    Object[] keys = typeAlphabet.toArray();
    for (int i = 0; i < keys.length; i++) {
      int indx = typeAlphabet.lookupAndAddIndex(keys[i]);
      types[indx] = (String) keys[i];
    }

    KBestParseForest.rootType = typeAlphabet.lookupAndAddIndex("<root-type>");
    
    Object[] featureNames = dataAlphabet.toArray();
    
    if (options.featureNamesFilename != null) {
      System.out.println("Writing feature names to " + options.featureNamesFilename);
      File outputFile = new File(options.featureNamesFilename);
      try {
        BufferedWriter bufferedOutput = new BufferedWriter(new FileWriter(
            outputFile));

        for (int i = 0; i < featureNames.length; i++) {
          bufferedOutput.write(i + "\t" + featureNames[i] + "\n");
        }
        bufferedOutput.close();
      } catch (IOException e) {
        System.out.println("Couldn't open feature names file.");
        e.printStackTrace();
      }
      
      System.out.println("Done writing feature names.");
    }
  }

  public final void add(String feat, FeatureVector fv) {
    add(feat, 1.0, fv);
  }
  
  public final void add(String feat, double value, FeatureVector fv) {    
    if (includeFeature(feat)) {
      int num = dataAlphabet.lookupAndAddIndex(feat);
      if (num >= 0) {
        fv.add(num, value);
      }
    }
    
    writeFeatureInformation(dataAlphabet.lookupIndex(feat), value);
  }
  
  public final void addWithSentenceInformation(String feat, FeatureVector fv, boolean sameSentence) {
    // conjoiningMethod.NONE doesn't conjoin features with distance
    // conjoiningMethod.DISTANCE_ONLY only includes features conjoined with distance
    // conjoiningMethod.BOTH only includes both types of features
    
    if (options.conjoinSentenceDistance != conjoiningMethod.DISTANCE_ONLY) {
      add(feat, fv);
    }
    
    if (options.conjoinSentenceDistance != conjoiningMethod.NONE) {
      add(feat + " || " + (sameSentence ? "S" : "D"), fv);
    }
  }

  /**
   * Returns true if feat is currently selected. 
   */
  public boolean includeFeature(String feat) {
    if (feat.startsWith("MST:") && !options.filterOriginalMSTFeatures) {
      return true;
    }
    
    boolean matchesPrefixRegex = options.featureNamePrefixes != null && options.featureNamePrefixes.matcher(feat).find();
    boolean okayToAdd = false;
    
    if (options.collectFeatureCounts) {
      featureCounts.incrementCount(feat);
    } 
    
    if (options.featureCountPruning) {
      int count = (int) featureCounts.getCount(feat);
      okayToAdd = (count >= options.minFeatureCount) && matchesPrefixRegex;
    } else if (matchesPrefixRegex) {
      okayToAdd = true;
    }
    
    return okayToAdd;
  }

  /*
   * Original MST code for creating feature vectors (with calls to our extensions)
   */
  
  public FeatureVector createFeatureVector(DependencyInstance instance, DependencyInstance[] stackingInstances) {

    final int instanceLength = instance.length();

    String[] labs = instance.deprels;
    int[] heads = instance.heads;

    FeatureVector fv = new FeatureVector();
    for (int i = 0; i < instanceLength; i++) {
      if (heads[i] == -1)
        continue;
      int small = i < heads[i] ? i : heads[i];
      int large = i > heads[i] ? i : heads[i];
      boolean attR = i < heads[i] ? false : true;
      
      printEdgeId(instance, heads[i], i, -1, labs[i], true);
      
      addCoreFeatures(instance, small, large, attR, fv, stackingInstances);
      if (labeled) {
        addLabeledFeatures(instance, i, labs[i], attR, true, fv, stackingInstances);
        addLabeledFeatures(instance, heads[i], labs[i], attR, false, fv, stackingInstances);
      }
    }
    
    addExtendedFeatures(instance, fv);

    return fv;
  }

  public boolean inSameSentence(DependencyInstance instance, int index1, int index2) {
    int childOrigSent = instance.sentenceIndex[index1];
    int parentOrigSent = instance.sentenceIndex[index2];
    return childOrigSent == parentOrigSent;    
  }
  
  public void addCoreFeatures(DependencyInstance instance, int small,
      int large, boolean attR, FeatureVector fv, DependencyInstance[] stackingInstances) {
    boolean sameSentence = inSameSentence(instance, small, large);
    
    String[] forms = instance.forms;
    String[] pos = instance.postags;
    String[] posA = instance.cpostags;

    String att = attR ? "RA" : "LA";

    int dist = Math.abs(large - small);
    String distBool = "0";
    if (dist > 10)
      distBool = "10";
    else if (dist > 5)
      distBool = "5";
    else
      distBool = Integer.toString(dist - 1);

    String attDist = "&" + att + "&" + distBool;

    addLinearFeatures("POS", pos, small, large, attDist, fv, sameSentence);
    addLinearFeatures("CPOS", posA, small, large, attDist, fv, sameSentence);

    // ////////////////////////////////////////////////////////////////////

    int headIndex = small;
    int childIndex = large;
    if (!attR) {
      headIndex = large;
      childIndex = small;
    }

    addTwoObsFeatures("HC", forms[headIndex], pos[headIndex],
        forms[childIndex], pos[childIndex], attDist, fv, sameSentence);

    if (isCONLL) {
      addTwoObsFeatures("HCA", forms[headIndex], posA[headIndex],
          forms[childIndex], posA[childIndex], attDist, fv, sameSentence);

      addTwoObsFeatures("HCC", instance.lemmas[headIndex], pos[headIndex],
          instance.lemmas[childIndex], pos[childIndex], attDist, fv, sameSentence);

      addTwoObsFeatures("HCD", instance.lemmas[headIndex], posA[headIndex],
          instance.lemmas[childIndex], posA[childIndex], attDist, fv, sameSentence);

      if (options.discourseMode) {
        // Note: The features invoked here are designed for
        // discourse parsing (as opposed to sentential
        // parsing). It is conceivable that they could help for
        // sentential parsing, but current testing indicates that
        // they hurt sentential parsing performance.

        addDiscourseFeatures(instance, small, large, headIndex, childIndex,
            attDist, fv);

      } else {
        // Add in features from the feature lists. It assumes
        // the feature lists can have different lengths for
        // each item. For example, nouns might have a
        // different number of morphological features than
        // verbs.

        for (int i = 0; i < instance.feats[headIndex].length; i++) {
          for (int j = 0; j < instance.feats[childIndex].length; j++) {
            addTwoObsFeatures("FF" + i + "*" + j, instance.forms[headIndex],
                instance.feats[headIndex][i], instance.forms[childIndex],
                instance.feats[childIndex][j], attDist, fv, sameSentence);

            addTwoObsFeatures("LF" + i + "*" + j, instance.lemmas[headIndex],
                instance.feats[headIndex][i], instance.lemmas[childIndex],
                instance.feats[childIndex][j], attDist, fv, sameSentence);
          }
        }
      }
    } else {
      // We are using the old MST format. Pick up stem features
      // the way they used to be done. This is kept for
      // replicability of results for old versions.
      int hL = forms[headIndex].length();
      int cL = forms[childIndex].length();
      if (hL > 5 || cL > 5) {
        addOldMSTStemFeatures(instance.lemmas[headIndex], pos[headIndex],
            instance.lemmas[childIndex], pos[childIndex], attDist, hL, cL, fv, sameSentence);
      }
    }

    dataset.addDomainSpecificFeatures(this, instance, headIndex, childIndex, fv);
    addEventParsingCoreFeatures(instance, headIndex, childIndex, fv);
    addStackingCoreFeatures(instance, stackingInstances, headIndex, childIndex, fv);
  }

  private final void addLinearFeatures(String type, String[] obsVals,
      int first, int second, String attachDistance, FeatureVector fv, boolean sameSentence) {
    String pLeft = first > 0 ? obsVals[first - 1] : "STR";
    String pRight = second < obsVals.length - 1 ? obsVals[second + 1] : "END";
    String pLeftRight = first < second - 1 ? obsVals[first + 1] : "MID";
    String pRightLeft = second > first + 1 ? obsVals[second - 1] : "MID";

    // feature posR posMid posL
    StringBuilder featPos = new StringBuilder(type + "PC=" + obsVals[first]
        + " " + obsVals[second]);

    for (int i = first + 1; i < second; i++) {
      String allPos = "MST:" + featPos.toString() + ' ' + obsVals[i];
      addWithSentenceInformation(allPos, fv, sameSentence);
      addWithSentenceInformation(allPos + attachDistance, fv, sameSentence);
    }

    addCorePosFeatures(type + "PT", pLeft, obsVals[first], pLeftRight,
        pRightLeft, obsVals[second], pRight, attachDistance, fv, sameSentence);
  }

  private final void addCorePosFeatures(String prefix, String leftOf1,
      String one, String rightOf1, String leftOf2, String two, String rightOf2,
      String attachDistance, FeatureVector fv, boolean sameSentence) {
    
    // feature posL-1 posL posR posR+1

    addWithSentenceInformation("MST:" + prefix + "=" + leftOf1 + " " + one + " " + two + "*" + attachDistance,
        fv, sameSentence);

    StringBuilder feat = new StringBuilder("MST:" + prefix + "1=" + leftOf1 + " " + one
        + " " + two);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append(' ').append(rightOf2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2=" + leftOf1 + " " + two + " "
        + rightOf2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "3=" + leftOf1 + " " + one + " "
        + rightOf2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "4=" + one + " " + two + " " + rightOf2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    // ///////////////////////////////////////////////////////////
    prefix = "A" + prefix;

    // feature posL posL+1 posR-1 posR
    addWithSentenceInformation("MST:" + prefix + "1=" + one + " " + rightOf1 + " " + leftOf2 + "*"
        + attachDistance, fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "1=" + one + " " + rightOf1 + " "
        + leftOf2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append(' ').append(two);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2=" + one + " " + rightOf1 + " " + two);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "3=" + one + " " + leftOf2 + " " + two);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "4=" + rightOf1 + " " + leftOf2 + " "
        + two);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    // /////////////////////////////////////////////////////////////
    prefix = "B" + prefix;

    // // feature posL-1 posL posR-1 posR
    feat = new StringBuilder("MST:" + prefix + "1=" + leftOf1 + " " + one + " "
        + leftOf2 + " " + two);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    // // feature posL posL+1 posR posR+1
    feat = new StringBuilder("MST:" + prefix + "2=" + one + " " + rightOf1 + " " + two
        + " " + rightOf2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
  }

  /**
   * Add features for two items, each with two observations, e.g. head, head
   * pos, child, and child pos.
   * 
   * The use of StringBuilders is not yet as efficient as it could be, but this
   * is a start. (And it abstracts the logic so we can add other features more
   * easily based on other items and observations.)
   **/
  private final void addTwoObsFeatures(String prefix, String item1F1,
      String item1F2, String item2F1, String item2F2, String attachDistance,
      FeatureVector fv, boolean sameSentence) {

    StringBuilder feat = new StringBuilder("MST:" + prefix + "2FF1=" + item1F1);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF1=" + item1F1 + " " + item1F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF1=" + item1F1 + " " + item1F2 + " "
        + item2F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF1=" + item1F1 + " " + item1F2 + " "
        + item2F2 + " " + item2F1);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF2=" + item1F1 + " " + item2F1);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF3=" + item1F1 + " " + item2F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF4=" + item1F2 + " " + item2F1);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF4=" + item1F2 + " " + item2F1 + " "
        + item2F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF5=" + item1F2 + " " + item2F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF6=" + item2F1 + " " + item2F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF7=" + item1F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF8=" + item2F1);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

    feat = new StringBuilder("MST:" + prefix + "2FF9=" + item2F2);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);
    feat.append('*').append(attachDistance);
    addWithSentenceInformation(feat.toString(), fv, sameSentence);

  }

  public void addLabeledFeatures(DependencyInstance instance, int word,
      String label, boolean attR, boolean childFeatures, FeatureVector fv, DependencyInstance[] stackingInstances) {

    if (!labeled)
      return;

    String[] forms = instance.forms;
    String[] pos = instance.postags;

    String att = "";
    if (attR)
      att = "RA";
    else
      att = "LA";

    att += "&" + childFeatures;

    String w = forms[word];
    String wP = pos[word];

    String wPm1 = word > 0 ? pos[word - 1] : "STR";
    String wPp1 = word < pos.length - 1 ? pos[word + 1] : "END";

    // these don't use addWithSentenceInformation since labeled features only know one of the words on the edge
    add("MST:NTS1=" + label + "&" + att, fv);
    add("MST:ANTS1=" + label, fv);
    for (int i = 0; i < 2; i++) {
      String suff = i < 1 ? "&" + att : "";
      suff = "&" + label + suff;

      add("MST:NTH=" + w + " " + wP + suff, fv);
      add("MST:NTI=" + wP + suff, fv);
      add("MST:NTIA=" + wPm1 + " " + wP + suff, fv);
      add("MST:NTIB=" + wP + " " + wPp1 + suff, fv);
      add("MST:NTIC=" + wPm1 + " " + wP + " " + wPp1 + suff, fv);
      add("MST:NTJ=" + w + suff, fv); // this
    }
    
    dataset.addDomainSpecificLabeledFeatures(this, instance, word, label, attR, childFeatures, fv);
    addEventParsingLabeledFeatures(instance, word, label, attR, childFeatures, false, fv);
    addEventParsingLabeledFeatures(instance, word, label, attR, childFeatures, true, fv);
    
    addStackingLabeledFeatures(instance, stackingInstances, word, label, attR, childFeatures, fv);
  }

  private void addDiscourseFeatures(DependencyInstance instance, int small,
      int large, int headIndex, int childIndex, String attDist, FeatureVector fv) {
    boolean sameSentence = inSameSentence(instance, small, large);
    
    addLinearFeatures("FORM", instance.forms, small, large, attDist, fv, sameSentence);
    addLinearFeatures("LEMMA", instance.lemmas, small, large, attDist, fv, sameSentence);

    addTwoObsFeatures("HCB1", instance.forms[headIndex],
        instance.lemmas[headIndex], instance.forms[childIndex],
        instance.lemmas[childIndex], attDist, fv, sameSentence);

    addTwoObsFeatures("HCB2", instance.forms[headIndex],
        instance.lemmas[headIndex], instance.forms[childIndex],
        instance.postags[childIndex], attDist, fv, sameSentence);

    addTwoObsFeatures("HCB3", instance.forms[headIndex],
        instance.lemmas[headIndex], instance.forms[childIndex],
        instance.cpostags[childIndex], attDist, fv, sameSentence);

    addTwoObsFeatures("HC2", instance.forms[headIndex],
        instance.postags[headIndex], instance.forms[childIndex],
        instance.cpostags[childIndex], attDist, fv, sameSentence);

    addTwoObsFeatures("HCC2", instance.lemmas[headIndex],
        instance.postags[headIndex], instance.lemmas[childIndex],
        instance.cpostags[childIndex], attDist, fv, sameSentence);

    // // Use this if your extra feature lists all have the same length.
    for (int i = 0; i < instance.feats.length; i++) {
      addLinearFeatures("F" + i, instance.feats[i], small, large, attDist, fv, sameSentence);

      addTwoObsFeatures("FF" + i, instance.forms[headIndex],
          instance.feats[i][headIndex], instance.forms[childIndex],
          instance.feats[i][childIndex], attDist, fv, sameSentence);

      addTwoObsFeatures("LF" + i, instance.lemmas[headIndex],
          instance.feats[i][headIndex], instance.lemmas[childIndex],
          instance.feats[i][childIndex], attDist, fv, sameSentence);

      addTwoObsFeatures("PF" + i, instance.postags[headIndex],
          instance.feats[i][headIndex], instance.postags[childIndex],
          instance.feats[i][childIndex], attDist, fv, sameSentence);

      addTwoObsFeatures("CPF" + i, instance.cpostags[headIndex],
          instance.feats[i][headIndex], instance.cpostags[childIndex],
          instance.feats[i][childIndex], attDist, fv, sameSentence);

      for (int j = i + 1; j < instance.feats.length; j++) {
        addTwoObsFeatures("CPF" + i + "_" + j, instance.feats[i][headIndex],
            instance.feats[j][headIndex], instance.feats[i][childIndex],
            instance.feats[j][childIndex], attDist, fv, sameSentence);
      }

      for (int j = 0; j < instance.feats.length; j++) {
        addTwoObsFeatures("XFF" + i + "_" + j, instance.forms[headIndex],
            instance.feats[i][headIndex], instance.forms[childIndex],
            instance.feats[j][childIndex], attDist, fv, sameSentence);

        addTwoObsFeatures("XLF" + i + "_" + j, instance.lemmas[headIndex],
            instance.feats[i][headIndex], instance.lemmas[childIndex],
            instance.feats[j][childIndex], attDist, fv, sameSentence);

        addTwoObsFeatures("XPF" + i + "_" + j, instance.postags[headIndex],
            instance.feats[i][headIndex], instance.postags[childIndex],
            instance.feats[j][childIndex], attDist, fv, sameSentence);

        addTwoObsFeatures("XCF" + i + "_" + j, instance.cpostags[headIndex],
            instance.feats[i][headIndex], instance.cpostags[childIndex],
            instance.feats[j][childIndex], attDist, fv, sameSentence);
      }
    }

    // Test out relational features
    if (options.useRelationalFeatures) {
      for (int rf_index = 0; rf_index < instance.relFeats.length; rf_index++) {
        String headToChild = "H2C" + rf_index
            + instance.relFeats[rf_index].getFeature(headIndex, childIndex);

        addTwoObsFeatures("RFA1", instance.forms[headIndex],
            instance.lemmas[headIndex], instance.postags[childIndex],
            headToChild, attDist, fv, sameSentence);

        addTwoObsFeatures("RFA2", instance.postags[headIndex],
            instance.cpostags[headIndex], instance.forms[childIndex],
            headToChild, attDist, fv, sameSentence);

        addTwoObsFeatures("RFA3", instance.lemmas[headIndex],
            instance.postags[headIndex], instance.forms[childIndex],
            headToChild, attDist, fv, sameSentence);

        addTwoObsFeatures("RFB1", headToChild, instance.postags[headIndex],
            instance.forms[childIndex], instance.lemmas[childIndex], attDist,
            fv, sameSentence);

        addTwoObsFeatures("RFB2", headToChild, instance.forms[headIndex],
            instance.postags[childIndex], instance.cpostags[childIndex],
            attDist, fv, sameSentence);

        addTwoObsFeatures("RFB3", headToChild, instance.forms[headIndex],
            instance.lemmas[childIndex], instance.postags[childIndex], attDist,
            fv, sameSentence);
      }
    }
  }
  
  protected void addExtendedFeatures(DependencyInstance instance,
      FeatureVector fv) {
    // intentionally left blank -- this is extended in DependencyPipe20
  }
  
  public void fillFeatureVectors(DependencyInstance instance,
      FeatureVector[][][] fvs, double[][][] probs,
      FeatureVector[][][][] nt_fvs, double[][][][] nt_probs, Parameters params,
      DependencyInstance[] stackingInstances) {

    final int instanceLength = instance.length();

    // Get production crap.
    for (int w1 = 0; w1 < instanceLength; w1++) {
      for (int w2 = w1 + 1; w2 < instanceLength; w2++) {
        for (int ph = 0; ph < 2; ph++) {
          boolean attR = ph == 0 ? true : false;

          int childInt = attR ? w2 : w1;
          int parInt = attR ? w1 : w2;
          
          printEdgeId(instance, parInt, childInt, -1, "unknown", false);

          FeatureVector prodFV = new FeatureVector();
          addCoreFeatures(instance, w1, w2, attR, prodFV, stackingInstances);
          double prodProb = params.getScore(prodFV);
          fvs[w1][w2][ph] = prodFV;
          probs[w1][w2][ph] = prodProb;
        }
      }
    }

    if (labeled) {
      for (int w1 = 0; w1 < instanceLength; w1++) {
        for (int t = 0; t < types.length; t++) {
          String type = types[t];
          for (int ph = 0; ph < 2; ph++) {

            boolean attR = ph == 0 ? true : false;
            for (int ch = 0; ch < 2; ch++) {

              boolean child = ch == 0 ? true : false;

              printEdgeId(instance, child ? -1 : w1, child ? w1 : -1, -1, type, false);
              
              FeatureVector prodFV = new FeatureVector();
              addLabeledFeatures(instance, w1, type, attR, child, prodFV, stackingInstances);

              double nt_prob = params.getScore(prodFV);
              nt_fvs[w1][t][ph][ch] = prodFV;
              nt_probs[w1][t][ph][ch] = nt_prob;

            }
          }
        }
      }
    }
  }

  /**
   * Write an instance to an output stream for later reading. 
   * 
   **/
  protected void writeInstance(DependencyInstance instance,
      ObjectOutputStream out, DependencyInstance[] stackingInstances) {

    int instanceLength = instance.length();

    try {

      for (int w1 = 0; w1 < instanceLength; w1++) {
        for (int w2 = w1 + 1; w2 < instanceLength; w2++) {
          for (int ph = 0; ph < 2; ph++) {
            boolean attR = ph == 0 ? true : false;
            
            int childInt = attR ? w2 : w1;
            int parInt = attR ? w1 : w2;
            
            printEdgeId(instance, parInt, childInt, -1, "unknown", false);
            
            FeatureVector prodFV = new FeatureVector();
            addCoreFeatures(instance, w1, w2, attR, prodFV, stackingInstances);
            out.writeObject(prodFV.keys());
          }
        }
      }
      out.writeInt(-3);

      if (labeled) {
        for (int w1 = 0; w1 < instanceLength; w1++) {
          for (int t = 0; t < types.length; t++) {
            String type = types[t];
            for (int ph = 0; ph < 2; ph++) {
              boolean attR = ph == 0 ? true : false;
              for (int ch = 0; ch < 2; ch++) {
                boolean child = ch == 0 ? true : false;
                FeatureVector prodFV = new FeatureVector();
                addLabeledFeatures(instance, w1, type, attR, child, prodFV, stackingInstances);
                out.writeObject(prodFV.keys());
              }
            }
          }
        }
        out.writeInt(-3);
      }

      writeExtendedFeatures(instance, out);

      out.writeObject(instance.fv.keys());
      out.writeInt(-4);

      out.writeObject(instance);
      out.writeInt(-1);

      out.reset();

    } catch (IOException e) {
    }

  }

  /**
   * Override this method if you have extra features that need to be written to
   * disk. For the basic DependencyPipe, nothing happens.
   * 
   */
  protected void writeExtendedFeatures(DependencyInstance instance,
      ObjectOutputStream out) throws IOException {
  }

  /**
   * Read an instance from an input stream.
   * 
   **/
  public DependencyInstance readInstance(ObjectInputStream in, int length,
      FeatureVector[][][] fvs, double[][][] probs,
      FeatureVector[][][][] nt_fvs, double[][][][] nt_probs, Parameters params)
      throws IOException {

    try {

      // Get production crap.
      for (int w1 = 0; w1 < length; w1++) {
        for (int w2 = w1 + 1; w2 < length; w2++) {
          for (int ph = 0; ph < 2; ph++) {
            FeatureVector prodFV = new FeatureVector((int[]) in.readObject());
            double prodProb = params.getScore(prodFV);
            fvs[w1][w2][ph] = prodFV;
            probs[w1][w2][ph] = prodProb;
          }
        }
      }
      int last = in.readInt();
      if (last != -3) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      if (labeled) {
        for (int w1 = 0; w1 < length; w1++) {
          for (int t = 0; t < types.length; t++) {
            @SuppressWarnings("unused")
            String type = types[t];

            for (int ph = 0; ph < 2; ph++) {
              for (int ch = 0; ch < 2; ch++) {
                FeatureVector prodFV = new FeatureVector((int[]) in
                    .readObject());
                double nt_prob = params.getScore(prodFV);
                nt_fvs[w1][t][ph][ch] = prodFV;
                nt_probs[w1][t][ph][ch] = nt_prob;
              }
            }
          }
        }
        last = in.readInt();
        if (last != -3) {
          System.out.println("Error reading file.");
          System.exit(0);
        }
      }

      FeatureVector nfv = new FeatureVector((int[]) in.readObject());
      last = in.readInt();
      if (last != -4) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      DependencyInstance marshalledDI;
      marshalledDI = (DependencyInstance) in.readObject();
      marshalledDI.setFeatureVector(nfv);

      last = in.readInt();
      if (last != -1) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      return marshalledDI;

    } catch (ClassNotFoundException e) {
      System.out.println("Error reading file.");
      System.exit(0);
    }

    // this won't happen, but it takes care of compilation complaints
    return null;
  }

  /**
   * Get features for stems the old way. The only way this differs from calling
   * addTwoObsFeatures() is that it checks the lengths of the full lexical items
   * are greater than 5 before adding features.
   * 
   */
  private final void addOldMSTStemFeatures(String hLemma, String headP,
      String cLemma, String childP, String attDist, int hL, int cL,
      FeatureVector fv, boolean sameSentence) {

    String all = hLemma + " " + headP + " " + cLemma + " " + childP;
    String hPos = headP + " " + cLemma + " " + childP;
    String cPos = hLemma + " " + headP + " " + childP;
    String hP = headP + " " + cLemma;
    String cP = hLemma + " " + childP;
    @SuppressWarnings("unused")
    String oPos = headP + " " + childP;
    String oLex = hLemma + " " + cLemma;

    addWithSentenceInformation("MST:SA=" + all + attDist, fv, sameSentence); // this
    addWithSentenceInformation("MST:SF=" + oLex + attDist, fv, sameSentence); // this
    addWithSentenceInformation("MST:SAA=" + all, fv, sameSentence); // this
    addWithSentenceInformation("MST:SFF=" + oLex, fv, sameSentence); // this

    if (cL > 5) {
      addWithSentenceInformation("MST:SB=" + hPos + attDist, fv, sameSentence);
      addWithSentenceInformation("MST:SD=" + hP + attDist, fv, sameSentence);
      addWithSentenceInformation("MST:SK=" + cLemma + " " + childP + attDist, fv, sameSentence);
      addWithSentenceInformation("MST:SM=" + cLemma + attDist, fv, sameSentence); // this
      addWithSentenceInformation("MST:SBB=" + hPos, fv, sameSentence);
      addWithSentenceInformation("MST:SDD=" + hP, fv, sameSentence);
      addWithSentenceInformation("MST:SKK=" + cLemma + " " + childP, fv, sameSentence);
      addWithSentenceInformation("MST:SMM=" + cLemma, fv, sameSentence); // this
    }
    if (hL > 5) {
      addWithSentenceInformation("MST:SC=" + cPos + attDist, fv, sameSentence);
      addWithSentenceInformation("MST:SE=" + cP + attDist, fv, sameSentence);
      addWithSentenceInformation("MST:SH=" + hLemma + " " + headP + attDist, fv, sameSentence);
      addWithSentenceInformation("MST:SJ=" + hLemma + attDist, fv, sameSentence); // this

      addWithSentenceInformation("MST:SCC=" + cPos, fv, sameSentence);
      addWithSentenceInformation("MST:SEE=" + cP, fv, sameSentence);
      addWithSentenceInformation("MST:SHH=" + hLemma + " " + headP, fv, sameSentence);
      addWithSentenceInformation("MST:SJJ=" + hLemma, fv, sameSentence); // this
    }
  }
  
  /*
   * Features for event parsing which use two words (but not the relation type)
   */
  
  /*
   * Adds all the new "core" features specific to event parsing -- these are a
   * function of two words but not the relation type
   */
  public void addEventParsingCoreFeatures(DependencyInstance instance, int headIndex, int childIndex,
      FeatureVector fv) {
    assert(instance != null);
    assert(instance.triggerProbs != null);
    assert(instance.firstTriggerProbs != null);
    addEventParsingTriggerProbFeatures("head", instance, headIndex, null, fv);
    addEventParsingTriggerProbFeatures("child", instance, childIndex, null, fv);

    int childOrigHead = instance.headInOriginal[childIndex];
    int parentOrigHead = instance.headInOriginal[headIndex];
    int childOrigSent = instance.sentenceIndex[childIndex];
    int parentOrigSent = instance.sentenceIndex[headIndex];
    
    addEventParsingCoreGeneralizedTypeFeatures(instance, headIndex, childIndex, childOrigSent, parentOrigSent, fv);
    
    // in different sentences!
    if (childOrigSent != parentOrigSent && childOrigSent != -1 && parentOrigSent != -1) {
      addEventParsingDocumentLevelFeatures(instance, childOrigHead, parentOrigHead, childOrigSent, parentOrigSent, fv);
      // can't add the other features 
      return;
    }

    addEventParsingConstituencyPathFeatures(instance, childOrigHead, parentOrigHead, childOrigSent, parentOrigSent, fv);

    addEventParsingDepPathFeatures(instance, childOrigHead, parentOrigHead, childOrigSent, parentOrigSent, true, fv);
    addEventParsingDepPathFeatures(instance, childOrigHead, parentOrigHead, childOrigSent, parentOrigSent, false, fv);

    addEventParsingSurfaceFeatures(instance, headIndex, childIndex, childOrigSent, parentOrigSent, fv);      
  }
  
  private void addEventParsingDocumentLevelFeatures(
      DependencyInstance instance, int childOrigHead, int parentOrigHead,
      int childOrigSent, int parentOrigSent, FeatureVector fv) {
    
    if (parentOrigSent == -1 || childOrigSent == -1) {
      return; // currently don't handle the case where one of these is the root -- should add it for anchor features
    }
        
    int dist = childOrigSent - parentOrigSent;
    if (dist != 0) {
      add("events:doclevel:diffsent:headchild:bool", fv);
    }
    add("events:doclevel:diffsent:headchild:dist " + Integer.toString(dist), fv);
    add("events:doclevel:diffsent:headchild:absdist " + Integer.toString(Math.abs(dist)), fv);

    // special version of dependency path features for when they're not in the same sentence
    // TODO only uncollapsed graph for now
    SemanticGraph parentGraph = instance.sentences.get(parentOrigSent).originalGraph;
    SemanticGraph childGraph = instance.sentences.get(childOrigSent).originalGraph;
    
    IndexedWord parentWord = parentGraph.getNodeByIndexSafe(parentOrigHead + 1);   
    IndexedWord childWord = childGraph.getNodeByIndexSafe(childOrigHead + 1);
    if (parentWord == null) {
      System.out.format("parentWord is null: index %d in graph:\n%s\n\n", parentOrigHead, parentGraph);
      System.out.println("Tree: " + instance.sentences.get(parentOrigSent).originalTreeAsString);
    }
    if (childWord == null) {
      System.out.format("childWord is null: index %d in graph:\n%s\n\n", childOrigHead, childGraph);
      System.out.println("Tree: " + instance.sentences.get(childOrigSent).originalTreeAsString);
    }
    
    List<SemanticGraphEdge> parentEdges = getSemanticEdgesFromPathToRoot(parentWord, parentGraph);
    List<SemanticGraphEdge> childEdges = getSemanticEdgesFromPathToRoot(childWord, childGraph);
    
    List<String> parentSpecificRelations = RelationFeatureFactory.dependencyPathAsList(parentEdges, parentWord, false);
    List<String> childSpecificRelations = RelationFeatureFactory.dependencyPathAsList(childEdges, childWord, false);
    
    List<String> parentGeneralizedRelations = RelationFeatureFactory.dependencyPathAsList(parentEdges, parentWord, true);
    List<String> childGeneralizedRelations = RelationFeatureFactory.dependencyPathAsList(childEdges, childWord, true);
    
    addEventParsingDocumentLevelPathFeatures(parentWord, childWord, parentSpecificRelations, childSpecificRelations, "specificRelations", fv);
    addEventParsingDocumentLevelPathFeatures(parentWord, childWord, parentGeneralizedRelations, childGeneralizedRelations, "generalizedRelations", fv);
    
    // now we look at the words along the paths
    List<IndexedWord> verticesPath;
    try {
      verticesPath = parentGraph.getPathToRoot(parentWord);
    } catch (IllegalArgumentException iae) {
      verticesPath = new ArrayList<IndexedWord>();
    }
    int parentPathLength = verticesPath.size();

    CoreLabel sentenceSeparatorDummyToken = new CoreLabel();
    sentenceSeparatorDummyToken.setWord("<sentence seperator>");
    sentenceSeparatorDummyToken.setLemma("<sentence seperator>");
    sentenceSeparatorDummyToken.setValue("<sentence seperator>");
    sentenceSeparatorDummyToken.setTag("<sentence seperator>");
    verticesPath.add(new IndexedWord(sentenceSeparatorDummyToken));
    
    List<IndexedWord> childPath;
    try {
      childPath = childGraph.getPathToRoot(childWord);
    } catch (IllegalArgumentException iae) {
      childPath = new ArrayList<IndexedWord>();
    }
    int childPathLength = childPath.size();
    Collections.reverse(childPath);
    verticesPath.addAll(childPath);
    
    Pair<List<String>,List<String>> wordsAndStemmedWords = getWordsAndStemmedWords(verticesPath);
    List<String> wordsInPath = wordsAndStemmedWords.first();
    List<String> stemmedWordsInPath = wordsAndStemmedWords.second();
    
    addPathFeatures(wordsInPath, "events:doclevel:path:dependency:words", fv);
    addPathFeatures(stemmedWordsInPath, "events:doclevel:path:dependency:stemmedWords", fv);
    
    // lengths of the paths
    int parentBucketedLength2 = parentPathLength - (parentPathLength % 2);
    int parentBucketedLength5 = parentPathLength - (parentPathLength % 5);
    int childBucketedLength2 = childPathLength - (childPathLength % 2);
    int childBucketedLength5 = childPathLength - (childPathLength % 5);
    
    add("events:doclevel:path:dependency:lengths " + parentPathLength + ":" + childPathLength, fv);
    add("events:doclevel:path:dependency:bucketedLengths2 " + parentBucketedLength2 + ":" + childBucketedLength2, fv);
    add("events:doclevel:path:dependency:bucketedLengths5 " + parentBucketedLength5 + ":" + childBucketedLength5, fv);
  }

  private void addEventParsingDocumentLevelPathFeatures(IndexedWord parentWord, IndexedWord childWord, List<String> parentRelationsPath, List<String> childRelationsPath, 
      String relationsDescription, FeatureVector fv) {
    // add padding if necessary
    while (parentRelationsPath.size() < 5) {
        parentRelationsPath.add(0, "(null)");
    }
    while (childRelationsPath.size() < 5) {
      childRelationsPath.add(0, "(null)");
    }
    
    List<String> fullPath = new ArrayList<String>(parentRelationsPath);
    Collections.reverse(fullPath);
    fullPath.add("<sentence separator>");
    fullPath.addAll(childRelationsPath);
    
    List<String> lexicalizedFullPath = lexicalizeRelationsPath(parentWord, childWord, fullPath);
    // TODO could do typedFullPath, generalizedTypesFullPath as well
    
    addPathFeatures(fullPath, "events:doclevel:path:dependency:" + relationsDescription + ":fullPath", fv);
    addPathFeatures(lexicalizedFullPath, "events:doclevel:path:dependency:" + relationsDescription + ":lexicalizedFullPath", fv);
  }

  private void addEventParsingCoreGeneralizedTypeFeatures(DependencyInstance instance, int headIndex, int childIndex,
      int childOrigSent, int parentOrigSent, FeatureVector fv) {
    String headForm = instance.forms[headIndex];
    String headType = instance.postags[headIndex];
    Set<String> generalizedHeadTypes = getGeneralizedTypes(headType);

    String childForm = instance.forms[childIndex];
    String childType = instance.postags[childIndex];
    Set<String> generalizedChildTypes = getGeneralizedTypes(childType);
    
    int distance = Math.abs(headIndex - childIndex);
    int bucketedDistance2 = distance - (distance % 2);
    int bucketedDistance5 = distance - (distance % 5);
    
    boolean sameSentence = inSameSentence(instance, headIndex, childIndex);
    
    for (String generalizedHeadType : generalizedHeadTypes) {
      for (String generalizedChildType : generalizedChildTypes) {
        addWithSentenceInformation("events:types:unlabeled:typeType: " + generalizedHeadType + " " + generalizedChildType, fv, sameSentence);
        addWithSentenceInformation("events:types:unlabeled:distTypeType: " + generalizedHeadType + " " + generalizedChildType + " " + distance, fv, sameSentence);
        addWithSentenceInformation("events:types:unlabeled:bucketedDist2TypeType: " + generalizedHeadType + " " + generalizedChildType + " " + bucketedDistance2, fv, sameSentence);
        addWithSentenceInformation("events:types:unlabeled:bucketedDist5TypeType: " + generalizedHeadType + " " + generalizedChildType + " " + bucketedDistance5, fv, sameSentence);
        addWithSentenceInformation("events:types:unlabeled:dirTypeType: " + generalizedHeadType + " " + generalizedChildType + " " + (headIndex - childIndex > 0 ? "L" : "R"), fv, sameSentence);
        addWithSentenceInformation("events:types:unlabeled:formType: " + headForm + " " + generalizedChildType, fv, sameSentence);
        addWithSentenceInformation("events:types:unlabeled:typeForm: " + generalizedHeadType + " " + childForm, fv, sameSentence);
        
        if (childOrigSent != parentOrigSent && childOrigSent != -1 && parentOrigSent != -1) {
          add("events:types:unlabeled:diffSent:typeType: " + generalizedHeadType + " " + generalizedChildType, fv);
          if (Math.abs(childOrigSent - parentOrigSent) == 1) { // adjacent sentences
            add("events:types:unlabeled:adjSent:typeType: " + generalizedHeadType + " " + generalizedChildType, fv);  
          }
        }
      }
    }
  }

  public void addEventParsingTriggerProbFeatures(String prefix,
      DependencyInstance instance, int index,
      String label, FeatureVector fv) {
    // the root node has no probability
    if(index < 0){
      System.err.println("Found ROOT node (" + index + ") when adding prob features!");
      return;
    }
    double prob = instance.triggerProbs[index];
    double firstProb = instance.firstTriggerProbs[index];
    
    // TODO these features could use the word's form/POS as well
    
    String desc = "prob";
    String suffix = "";
    if (label != null) {
      desc = "relationAndProb";
      suffix = ":" + label;
    }
    desc += ":" + prefix;
    
    if(prob == -1) {
      add("trigger:" + desc + ":nonvalue" + suffix, fv);
    } else {
      addProbFeature("trigger:" + desc + ":value" + suffix, prob, fv);
      if(firstProb == 0){
        add("trigger:" + desc + ":firstchoice" + suffix, fv);
      } else {
        double beam = firstProb - prob;
        addProbFeature("trigger:" + desc + ":beam" + suffix, beam, fv);
      }
    }
  }

  public void addEventParsingConstituencyPathFeatures(DependencyInstance instance, 
      int childHeadIndex,
      int parentHeadIndex,
      int childOrigSent, 
      int parentOrigSent, 
      FeatureVector fv) {
    assert (childOrigSent == parentOrigSent || childOrigSent == -1 || parentOrigSent == -1);
    Tree tree = instance.sentences.get(childOrigSent == -1 ? parentOrigSent : childOrigSent).originalTree;
    List<Tree> leaves = tree.getLeaves();

    List<String> constituencyPath;
    List<String> constituencyPathUnlex = new ArrayList<String>();
    
    if (parentHeadIndex == -1) {
      // if parent is the root, we have a special set of features -- instead of
      // path between head and argument, we use the path from the root to the
      // node
      List<Tree> constituencyPathNodes = Trees.pathFromRoot(leaves.get(childHeadIndex), tree);
      constituencyPath = new ArrayList<String>();
      for (Tree node : constituencyPathNodes) {
        constituencyPath.add(node.label().value());
      }
      
      constituencyPathUnlex = constituencyPath.subList(0, constituencyPath.size() - 1);
    } else if (childHeadIndex == -1) {
      // we got an invalid configuration (this happens when writeInstances() creates all possible dependencies)
      constituencyPath = new ArrayList<String>();
      constituencyPath.add("<invalid-configuration>");
      constituencyPathUnlex.add("<invalid-configuration>");
    } else {
      // path between head and child in constituency tree
      constituencyPath = Trees.pathNodeToNode(leaves.get(parentHeadIndex),
          leaves.get(childHeadIndex), tree);

      if (constituencyPath.size() > 2) {
        int length = constituencyPath.size();
        constituencyPathUnlex = constituencyPath.subList(1, length - 1);
      }
    }
    
    add("events:path:constituency:includeEndpoints " + StringUtils.join(constituencyPath), fv);
    String constituencyPathUnlexString = "(null)";
    if (constituencyPathUnlex.size() > 0) {
      constituencyPathUnlexString = StringUtils.join(constituencyPathUnlex);
    }
    add("events:path:constituency:noEndpoints " + constituencyPathUnlexString, fv);

    int length = constituencyPathUnlex.size();
    int bucketedLength3 = length - (length % 3);
    int bucketedLength10 = length - (length % 10);
    add("events:path:constituency:length " + length, fv);
    add("events:path:constituency:bucketedLength:3 " + bucketedLength3, fv);
    add("events:path:constituency:bucketedLength:10 " + bucketedLength10, fv);
  }
  
  public void addEventParsingDepPathFeatures(DependencyInstance instance, 
      int childHeadIndex,
      int parentHeadIndex,
      int childOrigSent, 
      int parentOrigSent, 
      boolean collapse, FeatureVector fv) {
    assert (childOrigSent == parentOrigSent || childOrigSent == -1 || parentOrigSent == -1);
    String desc = (collapse ? "collapsed" : "uncollapsed") + ":";
    int origSent = childOrigSent == -1 ? parentOrigSent : childOrigSent;
    SemanticGraph graph = collapse ? 
        instance.sentences.get(origSent).originalGraphCollapsed : 
        instance.sentences.get(origSent).originalGraph;
    
    IndexedWord parentWord;
    IndexedWord childWord;

    boolean parentIsRoot = false;
    try {
      if (parentHeadIndex == -1) {
        // hopefully the first root is okay -- if we run into problems, we'll
        // have to switch to the getPathToRoot method
        parentWord = graph.getFirstRoot();
        parentIsRoot = true;
      } else {
        parentWord = graph.getNodeByIndex(parentHeadIndex + 1);
      }
      if (childHeadIndex == -1) {
        // this is an "invalid configuration" (the child is the root).
        return;
      } else {
        childWord = graph.getNodeByIndex(childHeadIndex + 1);
      }
    } catch (IllegalArgumentException iae) {
      // this happens when one of the words is a preposition which has been
      // folded into a relation name and the actual IndexedWord has been removed
      return;
    }
    
    // first we look only at the grammatical relation names along the path
    List<SemanticGraphEdge> edgePath = graph.getShortestPathEdges(childWord, parentWord);
    // sanity check: once in a blue moon this is null
    if (edgePath == null) {
      // this happens if the graph gets disconnected.  it's rare but not a show stopper.
      add("events:path:dependency:" + desc + "edge:NIL", fv);
      return;
    }
    
    List<String> specificRelations = RelationFeatureFactory.dependencyPathAsList(edgePath, childWord, false);
    List<String> generalizedRelations = RelationFeatureFactory.dependencyPathAsList(edgePath, childWord, true);
    
    List<String> lexicalizedSpecificRelations = lexicalizeRelationsPath(parentWord, childWord, specificRelations);
    List<String> lexicalizedGeneralizedRelations = lexicalizeRelationsPath(parentWord, childWord, generalizedRelations);
    
    List<String> typedSpecificRelations = new ArrayList<String>(specificRelations);
    typedSpecificRelations.add(parentWord.tag());
    typedSpecificRelations.add(0, childWord.tag());

    String edgePathDescription = StringUtils.join(edgePath); 
    if (edgePath.size() == 0) {
      edgePathDescription = "(null)";
    }
    if (specificRelations.size() == 0) {
      specificRelations.add("(null)");
    }
    if (generalizedRelations.size() == 0) {
      generalizedRelations.add("(null)");
    }
    if (typedSpecificRelations.size() == 0) {
      typedSpecificRelations.add("(null)");
    }
    
    // add the paths themselves
    add("events:path:dependency:" + desc + "edge:undirectedSpecificRelations "
        + edgePathDescription, fv);
    add("events:path:dependency:" + desc + "edge:specificRelations "
        + StringUtils.join(specificRelations), fv);
    add("events:path:dependency:" + desc + "edge:generalizedRelations "
        + StringUtils.join(generalizedRelations), fv);
    // and lexicalized versions
    add("events:path:dependency:" + desc + "edge:lexicalizedSpecificRelations "
        + StringUtils.join(lexicalizedSpecificRelations), fv);
    add("events:path:dependency:" + desc + "edge:lexicalizedGeneralizedRelations "
        + StringUtils.join(lexicalizedGeneralizedRelations), fv);
    // and typed versions
    add("events:path:dependency:" + desc + "edge:typedSpecificRelations "
        + StringUtils.join(typedSpecificRelations), fv);
    // and lexicalized versions bucketed by length
    String specificRelationsShort, generalizedRelationsShort, typedSpecificRelationsShort; // only includes the full path if it's short 
    int specificRelationsSize = specificRelations.size(); // (same size as generalizedRelations)
    if (specificRelationsSize > 2) {
      specificRelationsShort = "<too-long>";
      generalizedRelationsShort = "<too-long>";
      typedSpecificRelationsShort = "<too-long>";
    } else {
      specificRelationsShort = StringUtils.join(lexicalizedSpecificRelations);
      generalizedRelationsShort = StringUtils.join(lexicalizedGeneralizedRelations);
      typedSpecificRelationsShort = StringUtils.join(typedSpecificRelations);
    }
    add("events:path:dependency:" + desc + "edge:bucketedLexicalizedSpecificRelations:" + specificRelationsSize + " "
        + specificRelationsShort, fv);
    add("events:path:dependency:" + desc + "edge:bucketedLexicalizedGeneralizedRelations:" + specificRelationsSize + " "
        + generalizedRelationsShort, fv);
    add("events:path:dependency:" + desc + "edge:bucketedTypedSpecificRelations:" + specificRelationsSize + " "
        + typedSpecificRelationsShort, fv);
    
    // TODO include length with path
    // TODO include length with path + lexicalized h and m

    // add n-grams from the paths
    List<List<SemanticGraphEdge>> undirectedEdgeNGrams = CollectionUtils.getNGrams(edgePath, 2, 4);
    for (List<SemanticGraphEdge> undirectedEdgeNGram : undirectedEdgeNGrams) {
      add("events:path:dependency:" + desc + "ngrams:edge:undirectedSpecificRelations " + StringUtils.join(undirectedEdgeNGram), fv);
    }
    List<List<String>> specificRelationsNGrams = CollectionUtils.getNGrams(specificRelations, 2, 4);
    for (List<String> specificRelationNGram : specificRelationsNGrams) {
      add("events:path:dependency:" + desc + "ngrams:edge:specificRelations " + StringUtils.join(specificRelationNGram), fv);
    }
    List<List<String>> generalizedRelationsNGrams = CollectionUtils.getNGrams(generalizedRelations, 2, 4);
    for (List<String> generalizedRelationNGram : generalizedRelationsNGrams) {
      add("events:path:dependency:" + desc + "ngrams:edge:generalizedRelations " + StringUtils.join(generalizedRelationNGram), fv);
    }
    
    // TODO optionally lexicalize these?
    List<List<String>> prefixes = CollectionUtils.getPrefixesAndSuffixes(specificRelations, 1, 4, "<padding>", true, false);
    List<List<String>> suffixes = CollectionUtils.getPrefixesAndSuffixes(specificRelations, 1, 4, "<padding>", false, true);
    List<List<String>> prefixesAndSuffixes = CollectionUtils.getPrefixesAndSuffixes(specificRelations, 1, 3, "<padding>", true, true);
    for (List<String> prefix : prefixes) {
      add("events:path:dependency:" + desc + "prefixes:edge:specificRelations " + StringUtils.join(prefix), fv);
    }
    for (List<String> suffix : suffixes) {
      add("events:path:dependency:" + desc + "suffixes:edge:specificRelations " + StringUtils.join(suffix), fv);
    }
    for (List<String> seq : prefixesAndSuffixes) {
      add("events:path:dependency:" + desc + "endpoints:edge:specificRelations " + StringUtils.join(seq), fv);
    }
    
    // now we look at the words along the paths
    List<IndexedWord> verticesPath = graph.getShortestPathNodes(childWord, parentWord);
    Pair<List<String>,List<String>> wordsAndStemmedWords = getWordsAndStemmedWords(verticesPath);
    List<String> wordsInPath = wordsAndStemmedWords.first();
    List<String> stemmedWordsInPath = wordsAndStemmedWords.second();
        
    add("events:path:dependency:" + desc + "vertices:words " + StringUtils.join(wordsInPath), fv);
    add("events:path:dependency:" + desc + "vertices:lemmas " + StringUtils.join(stemmedWordsInPath), fv);

    List<List<String>> wordNGrams = CollectionUtils.getNGrams(wordsInPath, 2, 4);
    for (List<String> wordNGram : wordNGrams) {
      add("events:path:dependency:" + desc + "ngrams:vertices:words " + StringUtils.join(wordNGram), fv);
    }
    List<List<String>> stemmedWordNGrams = CollectionUtils.getNGrams(stemmedWordsInPath, 2, 4);
    for (List<String> stemmedWordNGram : stemmedWordNGrams) {
      add("events:path:dependency:" + desc + "ngrams:vertices:lemmas " + StringUtils.join(stemmedWordNGram), fv);
    }
    
    // path length features
    int length = stemmedWordsInPath.size();
    int bucketedLength3 = length - (length % 3);
    int bucketedLength10 = length - (length % 10);
    add("events:path:dependency:" + desc + "length " + length, fv);
    add("events:path:dependency:" + desc + "bucketedLength:3 " + bucketedLength3, fv);
    add("events:path:dependency:" + desc + "bucketedLength:10 " + bucketedLength10, fv);
    
    // lexicalized lengths
    String parentAndChild = parentWord.word() + " " + childWord.word();
    add("events:path:dependency:" + desc + "lexicalizedLength " + parentAndChild + " " + length, fv);
    add("events:path:dependency:" + desc + "lexicalizedBucketedLength:3 " + parentAndChild + " " + bucketedLength3, fv);
    add("events:path:dependency:" + desc + "lexicalizedBucketedLength:10 " + parentAndChild + " " + bucketedLength10, fv);
    
    // special version of these to emphasize that the parent was the root node
    // rather than another entity/trigger word 
    if (parentIsRoot) {
      add("events:path:dependency:" + desc + "anchor:edge:undirectedSpecificRelations "
          + edgePathDescription, fv);
      add("events:path:dependency:" + desc + "anchor:edge:specificRelations "
          + StringUtils.join(specificRelations), fv);
      add("events:path:dependency:" + desc + "anchor:edge:generalizedRelations "
          + StringUtils.join(generalizedRelations), fv);
      
      for (List<String> prefix : prefixes) {
        add("events:path:dependency:" + desc + "anchor:prefixes:edge:specificRelations " + StringUtils.join(prefix), fv);
      }
      for (List<String> suffix : suffixes) {
        add("events:path:dependency:" + desc + "anchor:suffixes:edge:specificRelations " + StringUtils.join(suffix), fv);
      }
      for (List<String> seq : prefixesAndSuffixes) {
        add("events:path:dependency:" + desc + "anchor:endpoints:edge:specificRelations " + StringUtils.join(seq), fv);
      }
      
      add("events:path:dependency:" + desc + "anchor:length " + length, fv);
      add("events:path:dependency:" + desc + "anchor:bucketedLength:3 " + bucketedLength3, fv);
      add("events:path:dependency:" + desc + "anchor:bucketedLength:10 " + bucketedLength10, fv);
      
      add("events:path:dependency:" + desc + "anchor:lexicalizedLength " + childWord.word() + " " + length, fv);
      add("events:path:dependency:" + desc + "anchor:lexicalizedBucketedLength:3 " + childWord.word() + " " + bucketedLength3, fv);
      add("events:path:dependency:" + desc + "anchor:lexicalizedBucketedLength:10 " + childWord.word() + " " + bucketedLength10, fv);
    }
    
    // TODO lexicalized versions of the "edge" dependency paths
    // TODO add in relation name to some of these paths
    // TODO remove links which are inside the same entity
  }
  
  public void addEventParsingSurfaceFeatures(DependencyInstance instance,
      int headIndex, 
      int childIndex,
      int childOrigSent, 
      int parentOrigSent, 
      FeatureVector fv) {
    assert (childOrigSent == parentOrigSent || childOrigSent == -1 || parentOrigSent == -1);
    boolean sameSentence = inSameSentence(instance, headIndex, childIndex);
    int leftBoundary, rightBoundary;
    if (headIndex < childIndex) {
      leftBoundary = instance.endInOriginal[headIndex];
      rightBoundary = instance.startInOriginal[childIndex];
    } else {
      leftBoundary = instance.endInOriginal[childIndex];
      rightBoundary = instance.startInOriginal[headIndex];
    }
    
    // these throw out the case where one of the nodes is the root -- need features for this case
    if (leftBoundary != -1 && rightBoundary != -1) {
      int distance = rightBoundary - leftBoundary;
      int bucketedDistance3 = distance - (distance % 3);
      int bucketedDistance10 = distance - (distance % 10);
      addWithSentenceInformation("events:surface:distanceBetweenEntities " + distance, fv, sameSentence);
      addWithSentenceInformation("events:surface:bucketedDistanceBetweenEntities:3 " + bucketedDistance3, fv, sameSentence);
      addWithSentenceInformation("events:surface:bucketedDistanceBetweenEntities:10 " + bucketedDistance10, fv, sameSentence);

      List<Tree> leavesAsNodes = instance.sentences.get(childOrigSent).originalTree.getLeaves();
      List<String> leaves = new ArrayList<String>();
      for (Tree node : leavesAsNodes) {
        leaves.add(node.label().value());
      }
      
      List<Label> tagsAsLabels = new ArrayList<Label>(instance.sentences.get(childOrigSent).originalTree.preTerminalYield());
      List<String> tags = new ArrayList<String>();
      for (Label tag : tagsAsLabels) {
        tags.add(tag.value());
      }
      
      // TODO make a version that uses heads instead of boundaries as well?
      for (int offset = -2; offset <= 1; offset++) {
        for (int length = 1; length <= 3; length++) {
          for (String seqType : new String[] { "words", "tags" }) {
            List<String> seq = seqType.equals("words") ? leaves : tags;
            
            int seqLength = seq.size();
            
            int leftStart = Math.max(leftBoundary + offset, 0);
            int leftEnd = Math.min(leftStart + length, seqLength);
            leftEnd = Math.min(leftEnd, seqLength);
            
            int rightStart = Math.max(leftEnd, rightBoundary - offset - length);
            int rightEnd = Math.min(rightBoundary - offset, seqLength);
            rightEnd = Math.max(rightStart, rightEnd);
            
            if (leftStart < leftEnd) {
              List<String> subseq = seq.subList(leftStart, leftEnd);
              if (rightEnd > leftEnd) {
                subseq.addAll(seq.subList(rightStart, rightEnd));
              }

              if (headIndex < childIndex) {
                Collections.reverse(subseq);
              }

              addWithSentenceInformation("events:surface:" + seqType + "BetweenBoundaries:" + offset
                  + ":" + length + " " + StringUtils.join(subseq), fv, sameSentence);
            }
          }
        }
      }
    }    
  }
  
  private void addStackingCoreFeatures(DependencyInstance instance, DependencyInstance[] stackingInstances, int headIndex, int childIndex,
      FeatureVector fv) {
    if (stackingInstances == null || stackingInstances.length == 0) {
      return;
    }
    
    int stackingInstanceIndex = 0;
    for (DependencyInstance stackingInstance : stackingInstances) {
      int stackingHeadIndex = stackingInstance.heads[childIndex];
      int stackingReversedHeadIndex = stackingInstance.heads[headIndex];
      String headPos = stackingInstance.cpostags[headIndex];
      String childPos = stackingInstance.cpostags[childIndex];
      String label = stackingInstance.deprels[childIndex];
      
      if (stackingHeadIndex == headIndex) {
        addStackingCoreFeaturesForEdge(stackingInstanceIndex, headPos, childPos, label, "sameHead", fv);
      } else if (stackingReversedHeadIndex == childIndex) { // same arc but wrong direction
        addStackingCoreFeaturesForEdge(stackingInstanceIndex, headPos, childPos, label, "reversedHead", fv);
      } else {
        addStackingCoreFeaturesForEdge(stackingInstanceIndex, headPos, childPos, label, "differentHead", fv);
      }
      
      stackingInstanceIndex++;
    }
  }

  private void addStackingCoreFeaturesForEdge(int stackingInstanceIndex, String headPos, String childPos, String label, String description,
      FeatureVector fv) {
    add("stacking:" + description + ":" + stackingInstanceIndex + ":count", fv);
    /* more detailed stacking features:
    add("stacking:" + description + ":" + stackingInstanceIndex + ":headPos " + headPos, fv);
    add("stacking:" + description + ":" + stackingInstanceIndex + ":childPos " + childPos, fv);
    add("stacking:" + description + ":" + stackingInstanceIndex + ":headAndChildPos " + headPos + " " + childPos, fv);
    add("stacking:" + description + ":" + stackingInstanceIndex + ":label " + label, fv);
    add("stacking:" + description + ":" + stackingInstanceIndex + ":headPosAndlabel " + headPos + " " + label, fv);
    add("stacking:" + description + ":" + stackingInstanceIndex + ":childPosAndlabel " + childPos + " " + label, fv);
    add("stacking:" + description + ":" + stackingInstanceIndex + ":headAndChildPosAndlabel " + headPos + " " + childPos + " " + label, fv);
    */
  }
  
  /*
   * Features for event parsing which use one word and the edge label (relation name)
   */

  private void addEventParsingLabeledFeatures(DependencyInstance instance,
      int word, String relationName, boolean attachingRight, boolean childFeatures,
      boolean collapse, FeatureVector fv) {
    
    if (word == 0) { // the MST root token -- no meaningful features to add
      return;
    }
    
    String form = instance.forms[word];
    String type = instance.postags[word];
    Set<String> generalizedTypes = getGeneralizedTypes(type);
    for (String generalizedType : generalizedTypes) {
      // XXX domain specific -- taking it out since it shouldn't have any effect
      /*
      if (generalizedType.equalsIgnoreCase("Protein")) {
        continue; // Protein's generalized type is their normal type so they're handled by basic features
      }
      */
      
      String suffix = " " + relationName + " " + (childFeatures ? "child" : "head");
      String dir = " " + (attachingRight ? "right" : "left");
      
      add("events:types:formType: " + form + " " + generalizedType + suffix, fv);
      add("events:types:dirFormType: " + form + " " + generalizedType + suffix + dir, fv);
      add("events:types:type: " + generalizedType + suffix, fv);
      add("events:types:dirType: " + generalizedType + suffix + dir, fv);
    }
    
    int headInOriginal = instance.headInOriginal[word];
    int sent = instance.sentenceIndex[word];
    SemanticGraph graph = collapse ? 
        instance.sentences.get(sent).originalGraphCollapsed : 
        instance.sentences.get(sent).originalGraph;
    String graphType = (collapse ? "collapsed" : "uncollapsed") + ":";

    IndexedWord node;
    if (headInOriginal == -1) {
      node = graph.getFirstRoot();
    } else {
      try {
        node = graph.getNodeByIndex(headInOriginal + 1);
      } catch (IllegalArgumentException iea) {
        // this word may no longer be in the graph (words like "of" turn into relations)
        return;
      }
    }
    
    List<Collection<IndexedWord>> nodeLists = new ArrayList<Collection<IndexedWord>>();
    nodeLists.add(graph.getParents(node));
    nodeLists.add(graph.getChildren(node));
    try {
      nodeLists.add(graph.getSiblings(node));
    } catch (NullPointerException npe) {
      // this happens when node is a root
      nodeLists.add(new ArrayList<IndexedWord>());
    }
    String[] listNames = new String[] {"parents", "children", "siblings"};
    
    for (int i = 0; i < listNames.length; i++) {
      String name = listNames[i];
      Collection<IndexedWord> list = nodeLists.get(i);
      for (IndexedWord relatedNode : list) {
        add("events:graph:" + graphType + name + ":node " + node
            + " " + relatedNode + " " + relationName + " " + childFeatures, fv);
      }
      add("events:graph:" + graphType + name + ":size " + node + " "
          + list.size() + " " + relationName + " " + childFeatures, fv);
    }
    
    addEventParsingTriggerProbFeatures(childFeatures ? "child" : "head", instance, word, relationName, fv);
  }
  
  private void addStackingLabeledFeatures(DependencyInstance instance, DependencyInstance[] stackingInstances, int word, String label,
      boolean attR, boolean childFeatures, FeatureVector fv) {
    if (word == 0) { // the MST root token -- no meaningful features to add
      return;
    }
    
    if (stackingInstances == null || stackingInstances.length == 0) {
      return;
    }
    
    int stackingInstanceIndex = 0;
    for (DependencyInstance stackingInstance : stackingInstances) {
      String stackingLabel = null;
      
      // TODO much of this commented out due to last minute experiments on stacking features
      
      if (childFeatures) { // this is the child -- this is the easy case
        stackingLabel = stackingInstance.deprels[word];
        /*
        String childPos = stackingInstance.cpostags[word];
        String headPos = stackingInstance.cpostags[stackingInstance.heads[word]];
        */
        
        if (stackingLabel.equals(label)) {
          add("stacking:sameLabel:" + stackingInstanceIndex + ":count", fv);
          /*
          add("stacking:sameLabel:" + stackingInstanceIndex + ":label " + label, fv);
          add("stacking:sameLabel:" + stackingInstanceIndex + ":labelAndChildPos " + label + " " + childPos, fv);
          add("stacking:sameLabel:" + stackingInstanceIndex + ":labelAndHeadPos " + label + " " + headPos, fv);
          add("stacking:sameLabel:" + stackingInstanceIndex + ":labelAndHeadAndChildPos " + label + " " + headPos + " " + childPos, fv);
          */
        } else {
          add("stacking:differentLabel:" + stackingInstanceIndex + ":count", fv);
          /*
          add("stacking:differentLabel:" + stackingInstanceIndex + ":label " + label, fv);
          add("stacking:differentLabel:" + stackingInstanceIndex + ":labelAndChildPos " + label + " " + childPos, fv);
          add("stacking:differentLabel:" + stackingInstanceIndex + ":labelAndHeadPos " + label + " " + headPos, fv);
          add("stacking:differentLabel:" + stackingInstanceIndex + ":labelAndHeadAndChildPos " + label + " " + headPos + " " + childPos, fv);
          */
        }
      }
      /*
      else {  // this is the parent -- we need to invert the heads mapping to find the child
                // as a result, we may end up with multiple candidates
        String headPos = stackingInstance.cpostags[word];
        String childPos = null;
        String matchingStringLabel = null;
        
        // first, look at all children to see if any have us as a parent
        for (int stackingChild = 1; stackingChild < stackingInstance.heads.length; stackingChild++) {
          int stackingHead = stackingInstance.heads[stackingChild];
          if (stackingHead == word) {
            stackingLabel = stackingInstance.deprels[stackingChild];
            // if we find something with the same label, we stop
            if (stackingLabel.equals(label)) {
              matchingStringLabel = stackingLabel;
              childPos = stackingInstance.cpostags[stackingChild];
              break;
            }
          }
        }
        
        if (matchingStringLabel == null) {
          // can't do child features here since we never found a match
          add("stacking:headDifferentLabel:" + stackingInstanceIndex + ":label " + label, fv);
          add("stacking:headDifferentLabel:" + stackingInstanceIndex + ":labelAndHeadPos " + label + " " + headPos, fv);
        } else {
          add("stacking:headSameLabel:" + stackingInstanceIndex + ":label " + label, fv);
          add("stacking:headSameLabel:" + stackingInstanceIndex + ":labelAndHeadPos " + label + " " + headPos, fv);
          add("stacking:headSameLabel:" + stackingInstanceIndex + ":labelAndChildPos " + label + " " + childPos, fv);
          add("stacking:headSameLabel:" + stackingInstanceIndex + ":labelAndHeadAndChildPos " + label + " " + headPos + " " + childPos, fv);
        }
      }
      */
            
      stackingInstanceIndex++;
    }
  }
  
  /*
   * Helper functions for creating/adding features
   */
  
  private void addPathFeatures(List<String> path, String featureNamePrefix, FeatureVector fv) {
    // add path itself
    add(featureNamePrefix + ":path: " + StringUtils.join(path), fv);
    
    // prefixes and suffixes
    List<List<String>> prefixes = CollectionUtils.getPrefixesAndSuffixes(path, 1, 4, "<padding>", true, false);
    List<List<String>> suffixes = CollectionUtils.getPrefixesAndSuffixes(path, 1, 4, "<padding>", false, true);
    List<List<String>> prefixesAndSuffixes = CollectionUtils.getPrefixesAndSuffixes(path, 1, 4, "<padding>", true, true);

    for (List<String> prefix : prefixes) {
      add(featureNamePrefix + ":prefixes: " + StringUtils.join(prefix), fv);
    }
    for (List<String> suffix : suffixes) {
      add(featureNamePrefix + ":suffixes: " + StringUtils.join(suffix), fv);
    }
    for (List<String> seq : prefixesAndSuffixes) {
      add(featureNamePrefix + ":endpoints: " + StringUtils.join(seq), fv);
    }
  }
  
  private void addProbFeature(String name, double value, FeatureVector fv) {
    assert (value >= 0 && value <= 1);
    // System.err.println(name + ":" + value);
    add(name, value, fv);

    for (int i = 0; i < 4; i++) {
      double start = ((double) i) / 4;
      double end = start + 0.25;
      if (value >= start && value < end) {
        add(name + ":bin:" + i, fv);
      }
    }
  }
  
  /**
   * Given an entity type ("Gene_expression") returns a set of more general types containing it ("simple" and "non-protein" in the case of "Gene_expression").
   */
  public Set<String> getGeneralizedTypes(String entityType) {
    Set<String> generalizedTypes = new HashSet<String>();
    
    for (Entry<String, Set<String>> entry : dataset.getTypeGeneralizations().entrySet()) {
      if (entry.getValue().contains(entityType)) {
        generalizedTypes.add(entry.getKey());
      }
    }
    
    return generalizedTypes;
  }

  private List<SemanticGraphEdge> getSemanticEdgesFromPathToRoot(IndexedWord word, SemanticGraph graph) {
    List<SemanticGraphEdge> edgePath;
    
    if (word != null) {
      edgePath = graph.getShortestPathEdges(graph.getFirstRoot(), word);
    } else {
      edgePath = new ArrayList<SemanticGraphEdge>();
    }
    
    return edgePath;
  }

  private Pair<List<String>, List<String>> getWordsAndStemmedWords(List<IndexedWord> parentVerticesPath) {
    List<String> wordsInPath = new ArrayList<String>();
    List<String> stemmedWordsInPath = new ArrayList<String>();
    if (parentVerticesPath != null) {
      for (IndexedWord node : parentVerticesPath) {
        wordsInPath.add(node.word());
        stemmedWordsInPath.add(Morphology.stemStatic(node.value(), node.tag()).value());
      }
    } else {
      wordsInPath.add("null");
      stemmedWordsInPath.add("null");
    }
    return new Pair<List<String>, List<String>>(wordsInPath, stemmedWordsInPath);
  }
  
  /**
   * Takes a path and surrounds it with parentWord and childWord
   */
  private List<String> lexicalizeRelationsPath(IndexedWord parentWord,
      IndexedWord childWord, List<String> path) {
    List<String> lexicalizedGeneralizedRelations = new ArrayList<String>(path);
    String parentWordString;
    if (parentWord == null) {
      parentWordString = "(null)";
    } else {
      parentWordString = parentWord.word();
    }
    String childWordString;
    if (childWord == null) {
      childWordString = "(null)";
    } else {
      childWordString = childWord.word();
    }
    
    lexicalizedGeneralizedRelations.add(parentWordString.toLowerCase());
    
    lexicalizedGeneralizedRelations.add(0, childWordString.toLowerCase());
    return lexicalizedGeneralizedRelations;
  }

  /*
   * Functions for dumping edge statistics
   */
  
  protected void printEdgeId(DependencyInstance instance, int parentIndex, int childIndex,
      int siblingChildIndex, String relationType, boolean knownGold) {
    if (edgeFeatureStatisticsWriter != null) {
      try {
        edgeFeatureStatisticsWriter.write("\n" + instance.docId + " " + parentIndex + " " + childIndex + " " + siblingChildIndex + " "
            + relationType + " " + (knownGold ? "+" : "?") + " ");
      } catch (IOException e) {
        System.out.println("Couldn't write to edge feature statistics file:");
        e.printStackTrace();
      }
    }
  }
  
  protected void writeFeatureInformation(int featureNum, double value) {
    if (edgeFeatureStatisticsWriter != null) {
      if (featureNum < 0) {
        return;
      }

      String valueDesc = value == 1 ? "" : "=" + value;
      try {
        edgeFeatureStatisticsWriter.write(featureNum + valueDesc + " ");
      } catch (IOException e) {
        System.out.println("Couldn't write to edge feature statistics file:");
        e.printStackTrace();
      }
    }
  }  
}