package edu.stanford.nlp.ie.machinereading.msteventextractor;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import edu.stanford.nlp.ie.machinereading.RelationFeatureFactory;
import edu.stanford.nlp.ling.IndexedWord;
import edu.stanford.nlp.process.Morphology;
import edu.stanford.nlp.trees.semgraph.SemanticGraph;
import edu.stanford.nlp.trees.semgraph.SemanticGraphEdge;
import edu.stanford.nlp.util.CollectionUtils;
import edu.stanford.nlp.util.StringUtils;

public class DependencyPipe2O extends DependencyPipe {

  public DependencyPipe2O(ParserOptions options) throws IOException {
    super(options);
  }

  protected void addExtendedFeatures(DependencyInstance instance,
      FeatureVector fv) {

    final int instanceLength = instance.length();
    int[] heads = instance.heads;

    // find all trip features
    for (int i = 0; i < instanceLength; i++) {
      if (heads[i] == -1 && i != 0)
        continue;
      // right children
      int prev = i;
      for (int j = i + 1; j < instanceLength; j++) {
        if (heads[j] == i) {
          addTripFeatures(instance, i, prev, j, fv);
          addSiblingFeatures(instance, prev, j, prev == i, fv);
          prev = j;
        }
      }
      prev = i;
      for (int j = i - 1; j >= 0; j--) {
        if (heads[j] == i) {
          addTripFeatures(instance, i, prev, j, fv);
          addSiblingFeatures(instance, prev, j, prev == i, fv);
          prev = j;
        }
      }
    }
  }

  public void fillFeatureVectors(DependencyInstance instance,
      FeatureVector[][][] fvs, double[][][] probs,
      FeatureVector[][][] fvs_trips, double[][][] probs_trips,
      FeatureVector[][][] fvs_sibs, double[][][] probs_sibs,
      FeatureVector[][][][] nt_fvs, double[][][][] nt_probs, Parameters params,
      DependencyInstance[] stackingInstances) {

    fillFeatureVectors(instance, fvs, probs, nt_fvs, nt_probs, params, stackingInstances);

    final int instanceLength = instance.length();

    for (int w1 = 0; w1 < instanceLength; w1++) {
      for (int w2 = w1; w2 < instanceLength; w2++) {
        for (int w3 = w2 + 1; w3 < instanceLength; w3++) {
          FeatureVector prodFV = new FeatureVector();
          addTripFeatures(instance, w1, w2, w3, prodFV);
          double prodProb = params.getScore(prodFV);
          fvs_trips[w1][w2][w3] = prodFV;
          probs_trips[w1][w2][w3] = prodProb;
        }
      }
      for (int w2 = w1; w2 >= 0; w2--) {
        for (int w3 = w2 - 1; w3 >= 0; w3--) {
          FeatureVector prodFV = new FeatureVector();
          addTripFeatures(instance, w1, w2, w3, prodFV);
          double prodProb = params.getScore(prodFV);
          fvs_trips[w1][w2][w3] = prodFV;
          probs_trips[w1][w2][w3] = prodProb;
        }
      }
    }

    for (int w1 = 0; w1 < instanceLength; w1++) {
      for (int w2 = 0; w2 < instanceLength; w2++) {
        for (int wh = 0; wh < 2; wh++) {
          if (w1 != w2) {
            FeatureVector prodFV = new FeatureVector();
            addSiblingFeatures(instance, w1, w2, wh == 0, prodFV);
            double prodProb = params.getScore(prodFV);
            fvs_sibs[w1][w2][wh] = prodFV;
            probs_sibs[w1][w2][wh] = prodProb;
          }
        }
      }
    }
  }

  private final void addSiblingFeatures(DependencyInstance instance, int ch1,
      int ch2, boolean isST, FeatureVector fv) {

    String[] forms = instance.forms;
    String[] pos = instance.postags;

    // ch1 is always the closes to par
    String dir = ch1 > ch2 ? "RA" : "LA";

    String ch1_pos = isST ? "STPOS" : pos[ch1];
    String ch2_pos = pos[ch2];
    String ch1_word = isST ? "STWRD" : forms[ch1];
    String ch2_word = forms[ch2];

    add("MST:CH_PAIR=" + ch1_pos + "_" + ch2_pos + "_" + dir, fv);
    add("MST:CH_WPAIR=" + ch1_word + "_" + ch2_word + "_" + dir, fv);
    add("MST:CH_WPAIRA=" + ch1_word + "_" + ch2_pos + "_" + dir, fv);
    add("MST:CH_WPAIRB=" + ch1_pos + "_" + ch2_word + "_" + dir, fv);
    add("MST:ACH_PAIR=" + ch1_pos + "_" + ch2_pos, fv);
    add("MST:ACH_WPAIR=" + ch1_word + "_" + ch2_word, fv);
    add("MST:ACH_WPAIRA=" + ch1_word + "_" + ch2_pos, fv);
    add("MST:ACH_WPAIRB=" + ch1_pos + "_" + ch2_word, fv);

    int dist = Math.max(ch1, ch2) - Math.min(ch1, ch2);
    String distBool = "0";
    if (dist > 1)
      distBool = "1";
    if (dist > 2)
      distBool = "2";
    if (dist > 3)
      distBool = "3";
    if (dist > 4)
      distBool = "4";
    if (dist > 5)
      distBool = "5";
    if (dist > 10)
      distBool = "10";
    add("MST:SIB_PAIR_DIST=" + distBool + "_" + dir, fv);
    add("MST:ASIB_PAIR_DIST=" + distBool, fv);
    add("MST:CH_PAIR_DIST=" + ch1_pos + "_" + ch2_pos + "_" + distBool + "_" + dir,
        fv);
    add("MST:ACH_PAIR_DIST=" + ch1_pos + "_" + ch2_pos + "_" + distBool, fv);
    
    Set<String> ch1GeneralizedTypes = getGeneralizedTypes(ch1_pos);
    // the bogus STPOS tag doesn't have a generalized type
    if (isST) {
      ch1GeneralizedTypes.add(ch1_pos);
    }
    Set<String> ch2GeneralizedTypes = getGeneralizedTypes(ch2_pos);
    for (String ch1GeneralizedType : ch1GeneralizedTypes) {
      for (String ch2GeneralizedType : ch2GeneralizedTypes) {
        add("events:types:siblings:typeType: " + ch1GeneralizedType + " " + ch2GeneralizedType, fv);
      }
    }
    
    int sentIndex1 = instance.sentenceIndex[ch1];
    int sentIndex2 = instance.sentenceIndex[ch2];
    // args in different sentences
    if(sentIndex1 != sentIndex2){
      // we keep signed distance, to mark which argument appears before
      // TODO: should this be Math.abs? If we don't know the order in which siblings are generated, sign may be meaningless
      int sentDist = sentIndex1 - sentIndex2;
      if (sentIndex1 != -1 && sentIndex2 != -1) {
        if (sentDist != 0) {
          add("events:doclevel:diffsent:siblings:bool", fv);
        }
        add("events:doclevel:diffsent:siblings:dist " + Integer.toString(sentDist), fv);
        add("events:doclevel:diffsent:siblings:absdist " + Integer.toString(Math.abs(sentDist)), fv);
        // can't add the other features
      }
    } else {
      addEventParsingSiblingDepPathFeatures(instance, ch1, ch2, false, fv);
      addEventParsingSiblingDepPathFeatures(instance, ch1, ch2, true, fv);
    }
  }
  
  private void addEventParsingSiblingDepPathFeatures(DependencyInstance instance, int wordIndex1,
      int wordIndex2, boolean collapse, FeatureVector fv) {
    
    if (wordIndex1 == 0 || wordIndex2 == 0) { // one of them is the MST root
      return;
    }
    
    int sentIndex1 = instance.sentenceIndex[wordIndex1];
    int sentIndex2 = instance.sentenceIndex[wordIndex2];
    assert (sentIndex1 == sentIndex2);
    int wordOrigIndex1 = instance.headInOriginal[wordIndex1];
    int wordOrigIndex2 = instance.headInOriginal[wordIndex2];

    String desc = (collapse ? "collapsed" : "uncollapsed") + ":";
    SemanticGraph graph = collapse ? instance.sentences.get(sentIndex1).originalGraphCollapsed : 
      instance.sentences.get(sentIndex1).originalGraph;
    
    IndexedWord word1;
    IndexedWord word2;
    try {
      word1 = graph.getNodeByIndex(wordOrigIndex1 + 1);
      word2 = graph.getNodeByIndex(wordOrigIndex2 + 1);
    } catch (IllegalArgumentException iae) {
      // this happens when one of the words is a preposition which has been
      // folded into a relation name and the actual IndexedWord has been removed
      return;
    }
    
    // first we look only at the grammatical relation names along the path
    List<SemanticGraphEdge> edgePath = graph.getShortestPathEdges(word2, word1);

    List<String> specificRelations = RelationFeatureFactory.dependencyPathAsList(edgePath, word2, false);
    List<String> generalizedRelations = RelationFeatureFactory.dependencyPathAsList(edgePath, word2, true);

    if (edgePath == null) {
      // happens when the graph is disconnected
      edgePath = new ArrayList<SemanticGraphEdge>();
      
      specificRelations = new ArrayList<String>();
      specificRelations.add("(null)");
      
      generalizedRelations = new ArrayList<String>();
      generalizedRelations.add("(null)");
    }
    
    List<String> lexicalizedSpecificRelations = new ArrayList<String>(specificRelations);
    lexicalizedSpecificRelations.add(word1.word());
    lexicalizedSpecificRelations.add(0, word2.word());
    
    List<String> lexicalizedGeneralizedRelations = new ArrayList<String>(generalizedRelations);
    lexicalizedGeneralizedRelations.add(word1.word());
    lexicalizedGeneralizedRelations.add(0, word2.word());
    
    List<String> typedSpecificRelations = new ArrayList<String>(specificRelations);
    typedSpecificRelations.add(word1.tag());
    typedSpecificRelations.add(0, word2.tag());
    
    // add the paths themselves
    add("events:path:sibling:dependency:" + desc + "edge:undirectedSpecificRelations "
        + StringUtils.join(edgePath), fv);
    add("events:path:sibling:dependency:" + desc + "edge:specificRelations "
        + StringUtils.join(specificRelations), fv);
    add("events:path:sibling:dependency:" + desc + "edge:generalizedRelations "
        + StringUtils.join(generalizedRelations), fv);
    // and lexicalized versions
    add("events:path:sibling:dependency:" + desc + "edge:lexicalizedSpecificRelations "
        + StringUtils.join(lexicalizedSpecificRelations), fv);
    add("events:path:sibling:dependency:" + desc + "edge:lexicalizedGeneralizedRelations "
        + StringUtils.join(lexicalizedGeneralizedRelations), fv);
    // and typed versions
    add("events:path:sibling:dependency:" + desc + "edge:typedSpecificRelations "
        + StringUtils.join(typedSpecificRelations), fv);
    // and lexicalized versions bucketed by length
    String specificRelationsShort, generalizedRelationsShort, typedSpecificRelationsShort; // only includes the full path if it's short 
    int specificRelationsSize = specificRelations.size(); // (same size as generalizedRelations)
    if (specificRelationsSize > 2) {
      specificRelationsShort = "<too-long>";
      generalizedRelationsShort = "<too-long>";
      typedSpecificRelationsShort = "<too-long>";
    } else {
      specificRelationsShort = StringUtils.join(lexicalizedSpecificRelations);
      generalizedRelationsShort = StringUtils.join(lexicalizedGeneralizedRelations);
      typedSpecificRelationsShort = StringUtils.join(typedSpecificRelations);
    }
    add("events:path:sibling:dependency:" + desc + "edge:bucketedLexicalizedSpecificRelations:" + specificRelationsSize + " "
        + specificRelationsShort, fv);
    add("events:path:sibling:dependency:" + desc + "edge:bucketedLexicalizedGeneralizedRelations:" + specificRelationsSize + " "
        + generalizedRelationsShort, fv);
    add("events:path:sibling:dependency:" + desc + "edge:bucketedTypedSpecificRelations:" + specificRelationsSize + " "
        + typedSpecificRelationsShort, fv);
    
    // TODO include length with path
    // TODO include length with path + lexicalized h and m

    // add n-grams from the paths
    List<List<SemanticGraphEdge>> undirectedEdgeNGrams = CollectionUtils.getNGrams(edgePath, 2, 4);
    for (List<SemanticGraphEdge> undirectedEdgeNGram : undirectedEdgeNGrams) {
      add("events:path:sibling:dependency:" + desc + "ngrams:edge:undirectedSpecificRelations " + StringUtils.join(undirectedEdgeNGram), fv);
    }
    List<List<String>> specificRelationsNGrams = CollectionUtils.getNGrams(specificRelations, 2, 4);
    for (List<String> specificRelationNGram : specificRelationsNGrams) {
      add("events:path:sibling:dependency:" + desc + "ngrams:edge:specificRelations " + StringUtils.join(specificRelationNGram), fv);
    }
    List<List<String>> generalizedRelationsNGrams = CollectionUtils.getNGrams(generalizedRelations, 2, 4);
    for (List<String> generalizedRelationNGram : generalizedRelationsNGrams) {
      add("events:path:sibling:dependency:" + desc + "ngrams:edge:generalizedRelations " + StringUtils.join(generalizedRelationNGram), fv);
    }
    
    // TODO optionally lexicalize these?
    List<List<String>> prefixes = CollectionUtils.getPrefixesAndSuffixes(specificRelations, 1, 4, "<padding>", true, false);
    List<List<String>> suffixes = CollectionUtils.getPrefixesAndSuffixes(specificRelations, 1, 4, "<padding>", false, true);
    List<List<String>> prefixesAndSuffixes = CollectionUtils.getPrefixesAndSuffixes(specificRelations, 1, 3, "<padding>", true, true);
    for (List<String> prefix : prefixes) {
      add("events:path:sibling:dependency:" + desc + "prefixes:edge:specificRelations " + StringUtils.join(prefix), fv);
    }
    for (List<String> suffix : suffixes) {
      add("events:path:sibling:dependency:" + desc + "suffixes:edge:specificRelations " + StringUtils.join(suffix), fv);
    }
    for (List<String> seq : prefixesAndSuffixes) {
      add("events:path:sibling:dependency:" + desc + "endpoints:edge:specificRelations " + StringUtils.join(seq), fv);
    }
    
    // now we look at the words along the paths
    List<IndexedWord> verticesPath = graph.getShortestPathNodes(word2, word1);
    List<String> wordsInPath = new ArrayList<String>();
    List<String> stemmedWordsInPath = new ArrayList<String>();
    
    if (verticesPath != null) {
      for (IndexedWord node : verticesPath) {
        wordsInPath.add(node.word());
        stemmedWordsInPath.add(Morphology.stemStatic(node.value(), node.tag()).value());
      }
    } else {
      wordsInPath.add("null");
      stemmedWordsInPath.add("null");
    }
    
    add("events:path:sibling:dependency:" + desc + "vertices:words " + StringUtils.join(wordsInPath), fv);
    add("events:path:sibling:dependency:" + desc + "vertices:lemmas " + StringUtils.join(stemmedWordsInPath), fv);

    List<List<String>> wordNGrams = CollectionUtils.getNGrams(wordsInPath, 2, 4);
    for (List<String> wordNGram : wordNGrams) {
      add("events:path:sibling:dependency:" + desc + "ngrams:vertices:words " + StringUtils.join(wordNGram), fv);
    }
    List<List<String>> stemmedWordNGrams = CollectionUtils.getNGrams(stemmedWordsInPath, 2, 4);
    for (List<String> stemmedWordNGram : stemmedWordNGrams) {
      add("events:path:sibling:dependency:" + desc + "ngrams:vertices:lemmas " + StringUtils.join(stemmedWordNGram), fv);
    }
    
    // path length features
    int length = stemmedWordsInPath.size();
    int bucketedLength2 = length - (length % 2);
    int bucketedLength5 = length - (length % 5);
    add("events:path:sibling:dependency:" + desc + "length " + length, fv);
    add("events:path:sibling:dependency:" + desc + "bucketedLength:2 " + bucketedLength2, fv);
    add("events:path:sibling:dependency:" + desc + "bucketedLength:5 " + bucketedLength5, fv);
    
    // lexicalized lengths
    String bothWords = word1.word() + " " + word2.word();
    add("events:path:sibling:dependency:" + desc + "lexicalizedLength " + bothWords + " " + length, fv);
    add("events:path:sibling:dependency:" + desc + "lexicalizedBucketedLength:2 " + bothWords + " " + bucketedLength2, fv);
    add("events:path:sibling:dependency:" + desc + "lexicalizedBucketedLength:5 " + bothWords + " " + bucketedLength5, fv);
  }

  private final void addTripFeatures(DependencyInstance instance, int par,
      int ch1, int ch2, FeatureVector fv) {

    String[] pos = instance.postags;

    // ch1 is always the closest to par
    String dir = par > ch2 ? "RA" : "LA";

    String par_pos = pos[par];
    String ch1_pos = ch1 == par ? "STPOS" : pos[ch1];
    String ch2_pos = pos[ch2];

    String pTrip = par_pos + "_" + ch1_pos + "_" + ch2_pos;
    add("MST:POS_TRIP=" + pTrip + "_" + dir, fv);
    add("MST:APOS_TRIP=" + pTrip, fv);

    /* removed for now since they seem to be hurting
    String[] cpos = instance.cpostags;     
    String par_cpos = cpos[par];
    String ch1_cpos = ch1 == par ? "STPOS" : cpos[ch1];
    String ch2_cpos = cpos[ch2];
    
    Set<String> parGeneralizedTypes = getGeneralizedTypes(par_cpos);
    Set<String> ch1GeneralizedTypes = getGeneralizedTypes(ch1_cpos);
    Set<String> ch2GeneralizedTypes = getGeneralizedTypes(ch2_cpos);
    for (String parGeneralizedType : parGeneralizedTypes) {
      for (String ch1GeneralizedType : ch1GeneralizedTypes) {
        for (String ch2GeneralizedType : ch2GeneralizedTypes) {
          add("events:types:triples:typeTypeType: " + parGeneralizedType + " " + ch1GeneralizedType + " " + ch2GeneralizedType, fv);
        }
      }
    }
    
    if (!ch1_cpos.equals("STPOS")) {
      String[] lemmas = instance.lemmas;
      // these fire when we have a parent that has two real children (i.e. not sentence boundaries)
      // note: taking some of these out since they don't seem to be helping
      // add("events:types:triples:parentForm: " + instance.forms[par], fv);
      add("events:types:triples:parentType: " + par_cpos, fv);
      // add("events:types:triples:parentLemma: " + lemmas[par], fv);
      
      // these are similar to the above but fire when the children are the same type (which may
      // help distinguish some events, e.g. Binding)
      if (ch1_cpos.equalsIgnoreCase(ch2_cpos)) {
        add("events:types:triples:sameChildTypes:parentLemma: " + lemmas[par], fv);
        add("events:types:triples:sameChildTypes:parentLemma:type " + lemmas[par] + " " + ch1_cpos, fv);
        add("events:types:triples:sameChildTypes:parentPos: " + par_cpos, fv);
        add("events:types:triples:sameChildTypes:parentPos:type " + par_cpos + " " + ch1_cpos, fv);
      }
    }
    */
  }

  /**
   * Write out the second order features.
   * 
   **/
  protected void writeExtendedFeatures(DependencyInstance instance,
      ObjectOutputStream out) throws IOException {

    final int instanceLength = instance.length();

    for (int w1 = 0; w1 < instanceLength; w1++) {
      for (int w2 = w1; w2 < instanceLength; w2++) {
        for (int w3 = w2 + 1; w3 < instanceLength; w3++) {
          printEdgeId(instance, w1, w2, w3, "unknown", false);

          FeatureVector prodFV = new FeatureVector();
          addTripFeatures(instance, w1, w2, w3, prodFV);
          out.writeObject(prodFV.keys());
        }
      }
      for (int w2 = w1; w2 >= 0; w2--) {
        for (int w3 = w2 - 1; w3 >= 0; w3--) {
          printEdgeId(instance, w1, w2, w3, "unknown", false);
          
          FeatureVector prodFV = new FeatureVector();
          addTripFeatures(instance, w1, w2, w3, prodFV);
          out.writeObject(prodFV.keys());
        }
      }
    }

    out.writeInt(-3);

    for (int w1 = 0; w1 < instanceLength; w1++) {
      for (int w2 = 0; w2 < instanceLength; w2++) {
        for (int wh = 0; wh < 2; wh++) {
          if (w1 != w2) {
            printEdgeId(instance, -1, w1, w2, "unknown", false);
            
            FeatureVector prodFV = new FeatureVector();
            addSiblingFeatures(instance, w1, w2, wh == 0, prodFV);
            out.writeObject(prodFV.keys());
          }
        }
      }
    }

    out.writeInt(-3);
  }

  public DependencyInstance readInstance(ObjectInputStream in, int length,
      FeatureVector[][][] fvs, double[][][] probs,
      FeatureVector[][][] fvs_trips, double[][][] probs_trips,
      FeatureVector[][][] fvs_sibs, double[][][] probs_sibs,
      FeatureVector[][][][] nt_fvs, double[][][][] nt_probs, Parameters params)
      throws IOException {

    try {
      // Get production crap.
      for (int w1 = 0; w1 < length; w1++) {
        for (int w2 = w1 + 1; w2 < length; w2++) {
          for (int ph = 0; ph < 2; ph++) {
            FeatureVector prodFV = new FeatureVector((int[]) in.readObject());
            double prodProb = params.getScore(prodFV);
            fvs[w1][w2][ph] = prodFV;
            probs[w1][w2][ph] = prodProb;
          }
        }
      }
      int last = in.readInt();
      if (last != -3) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      if (labeled) {
        for (int w1 = 0; w1 < length; w1++) {
          for (int t = 0; t < types.length; t++) {
            @SuppressWarnings("unused")
            String type = types[t];
            for (int ph = 0; ph < 2; ph++) {
              for (int ch = 0; ch < 2; ch++) {
                FeatureVector prodFV = new FeatureVector((int[]) in
                    .readObject());
                double nt_prob = params.getScore(prodFV);
                nt_fvs[w1][t][ph][ch] = prodFV;
                nt_probs[w1][t][ph][ch] = nt_prob;
              }
            }
          }
        }
        last = in.readInt();
        if (last != -3) {
          System.out.println("Error reading file.");
          System.exit(0);
        }
      }

      for (int w1 = 0; w1 < length; w1++) {
        for (int w2 = w1; w2 < length; w2++) {
          for (int w3 = w2 + 1; w3 < length; w3++) {
            FeatureVector prodFV = new FeatureVector((int[]) in.readObject());
            double prodProb = params.getScore(prodFV);
            fvs_trips[w1][w2][w3] = prodFV;
            probs_trips[w1][w2][w3] = prodProb;
          }
        }
        for (int w2 = w1; w2 >= 0; w2--) {
          for (int w3 = w2 - 1; w3 >= 0; w3--) {
            FeatureVector prodFV = new FeatureVector((int[]) in.readObject());
            double prodProb = params.getScore(prodFV);
            fvs_trips[w1][w2][w3] = prodFV;
            probs_trips[w1][w2][w3] = prodProb;
          }
        }
      }
      last = in.readInt();
      if (last != -3) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      for (int w1 = 0; w1 < length; w1++) {
        for (int w2 = 0; w2 < length; w2++) {
          for (int wh = 0; wh < 2; wh++) {
            if (w1 != w2) {
              FeatureVector prodFV = new FeatureVector((int[]) in.readObject());
              double prodProb = params.getScore(prodFV);
              fvs_sibs[w1][w2][wh] = prodFV;
              probs_sibs[w1][w2][wh] = prodProb;
            }
          }
        }
      }
      last = in.readInt();
      if (last != -3) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      FeatureVector nfv = new FeatureVector((int[]) in.readObject());
      last = in.readInt();
      if (last != -4) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      DependencyInstance marshalledDI;
      marshalledDI = (DependencyInstance) in.readObject();
      marshalledDI.setFeatureVector(nfv);
      last = in.readInt();
      if (last != -1) {
        System.out.println("Error reading file.");
        System.exit(0);
      }

      return marshalledDI;

    } catch (ClassNotFoundException e) {
      System.out.println("Error reading file.");
      System.exit(0);
    }

    // this won't happen, but it takes care of compilation complaints
    return null;

  }

}
